// ------------------------------------------------------------------
// objectstream.cpp
//
// This file contains functions for reading and writing a brep or
// simplicial complex from/to a stream.
// ------------------------------------------------------------------
// Author: Stephen A. Vavasis
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------


#ifdef _MSC_VER
#if _MSC_VER == 1200
#pragma warning (disable: 4786)
#pragma warning (disable: 4788)
#endif
#endif


#include <iomanip>
#include <cctype>
#include "qerr.h"
#include "qbrep_constr.h"
#include "qsimpcomp.h"

// exported functions:

namespace QMG {
  Object_Type_Code read_header_from_istream(istream& istr, int& line_count);
  void init_faceselect_order(vector<unsigned int>& order, int gdim);
  string get_element_typename(int gdim, int order);
}

namespace {
  using namespace QMG;

  
  // ------------------------------------------------------------------
  // class Ostream_with_wraparound
  // Nonglobal class Ostream_with_wraparound is like an ostream,
  // except it checks whether line length has exceeded 72 and if so
  // inserts a newline.

  class Ostream_with_wraparound {
  private:
    ostream& real_stream_;    
    int linelength_;
    bool break_allowed_;
  public:
    ostream& real_stream() { return real_stream_;}
    void putstring(const ostringstream& ostr_tmp);
    explicit Ostream_with_wraparound(ostream& os) : real_stream_(os), linelength_(0),
    break_allowed_(true) { }
    void forbid_break() {break_allowed_ = false;}
    void enable_break() {break_allowed_ = true;}
  };

  void Ostream_with_wraparound::putstring(const ostringstream& ostr_tmp) {
    const string& outstr = ostr_tmp.str();
    int len = outstr.length();
    int plen = 0;
    bool hasnewline = false;
    for (int j = 0; j < len; ++j) {
      if (outstr[j] == '\n') {
        plen = 0;
        hasnewline = true;
      }
      else if (outstr[j] == '\t') {
        ++plen;
        if (plen % 8)
          plen += (8 - (plen % 8));
      }
      else {
        ++plen;
      }
    }
    if (hasnewline || !break_allowed_) {
      real_stream_ << outstr;
      linelength_ = plen;
    }
    else {
      if (linelength_ + plen > 72) {
        real_stream_ << "\n    ";
        linelength_ = 4;
      }
      real_stream_ << outstr;
      linelength_ += plen;
    }
  }

  template<class T> Ostream_with_wraparound& operator<<(Ostream_with_wraparound& os,
    const T& t) {
    ostringstream ostr_tmp;
    ostr_tmp << t;
    os.putstring(ostr_tmp);
    return os;
  }


  Ostream_with_wraparound& nobreak_begin(Ostream_with_wraparound& os) {
    os.forbid_break();
    return os;
  }

  Ostream_with_wraparound& nobreak_end(Ostream_with_wraparound& os) {
    os.enable_break();
    return os;
  }

  Ostream_with_wraparound& operator<<(Ostream_with_wraparound& os,
    Ostream_with_wraparound& (*f)(Ostream_with_wraparound&)) {
    return f(os);
  }

  Ostream_with_wraparound& operator<<(Ostream_with_wraparound& os,
    Real d) {
    ostringstream ostr_tmp;
    ostr_tmp << std::setprecision(17) << d;
    os.putstring(ostr_tmp);
    return os;
  }
 
   
    

  // ------------------------------------------------------------------
  // Nonglobal class Istream_skip_comments reads strings, ints and doubles
  // from an istream.  It throws an exception if it doesn't like the input.
  // It also skips comments, which are delimited by a # sign.

  const char COMMENT_CHAR = '#';
  const char LEFT_PAREN = '(';
  const char RIGHT_PAREN = ')';

  class Istream_skip_comments {
  private:
    istream& real_stream_;
    int linecount_;
    Error_Message err_;
    void increment_line_();
    void skip_space_();
    void skip_comments_();
  public:
    explicit Istream_skip_comments(istream& is, int init_linecount = -1) : 
    real_stream_(is), linecount_(init_linecount) {
      err_.set_string("");
      increment_line_(); 
    }
    void get_left_paren();
    bool get_left_paren2();
    bool get_string(string& s);
    bool get_int(int& i);
    bool get_uint(unsigned int& i);
    bool get_double(double& d);
    int linecount() const {return linecount_;}
    void go_to_eof();
  };

  void Istream_skip_comments::increment_line_() {
    if (linecount_ >= 0) {
      ++linecount_;
      ostringstream os;
      os << "Error occurred while reading line " << linecount_;
      err_.set_string(os.str());
    }
  }

  void Istream_skip_comments::skip_space_() {
    char c;
    for (;;) {
      if (!real_stream_.get(c)) return;
      if (c == '\n') increment_line_();
      if (!isspace(c)) {
        real_stream_.putback(c);
        return;
      }
    }
  }

  void Istream_skip_comments::skip_comments_() {
    char c;
    for (;;) {
      skip_space_();
      if (!real_stream_.get(c)) return;
      if (c != COMMENT_CHAR) {
        real_stream_.putback(c);
        return;
      }
      for (;;) {
        if (!real_stream_.get(c) || c == '\n')
          break;
      }
      increment_line_();
    }
  }

  void Istream_skip_comments::get_left_paren() {
    skip_comments_();
    char c;
    if (!real_stream_.get(c))
      throw_error("Unexpected eof while scanning for open-paren");
    if (c != LEFT_PAREN)
      throw_error("Expected open-paren");
  }

  bool Istream_skip_comments::get_left_paren2() {
    skip_comments_();
    char c;
    if (!real_stream_.get(c))
      throw_error("Unexpected eof while scanning for open-paren");
    if (c == LEFT_PAREN)
      return true;
    else if (c == RIGHT_PAREN)
      return false;
    else
      throw_error("Expected open or close paren");
    return false; // never reached; needed for MSVC      
  }

  // A string is terminate either by whitespace, provided there
  // is no pending open paren.
  // If there is an open paren, then wait for the close paren.

  bool Istream_skip_comments::get_string(string& s) {
    skip_comments_();
    char c;
    if (!real_stream_.get(c))
      throw_error("Unexpected eof while scanning for string");
    if (c == RIGHT_PAREN)
      return false;
    real_stream_.putback(c);
    int nest_level = 0;
    for (;;) {
      if (!real_stream_.get(c))
        throw_error("Unexpected eof in the middle of reading string");
      if (c == COMMENT_CHAR) {
        real_stream_.putback(c);
        skip_comments_();
        continue;
      }
      if (c == '\n') increment_line_();
      if (isspace(c) && nest_level == 0)
        return true;
      if (c == LEFT_PAREN) ++nest_level;
      if (c == RIGHT_PAREN) {
        --nest_level;
        if (nest_level == -1) {
          real_stream_.putback(c);
          return true;
        }
        if (nest_level == 0) {
          s.append(1, c);
          return true;
        }
      }
      s.append(1, c);
    }
  }

  bool Istream_skip_comments::get_int(int& i) {
    skip_comments_();
    char c;
    if (!real_stream_.get(c))
      throw_error("Unexpected eof in the middle of reading an int");
    if (c == RIGHT_PAREN)
      return false;
    real_stream_.putback(c);
    if (!(real_stream_ >> i)) 
      throw_error("Error while reading integer");
    return true;
  }
  
  bool Istream_skip_comments::get_uint(unsigned int& i) {
    skip_comments_();
    char c;
    if (!real_stream_.get(c))
      throw_error("Unexpected eof in the middle of reading an int");
    if (c == RIGHT_PAREN)
      return false;
    real_stream_.putback(c);
    if (!(real_stream_ >> i)) 
      throw_error("Error while reading unsigned integer");
    return true;
  }

  bool Istream_skip_comments::get_double(double& d) {
   skip_comments_();
    char c;
    if (!real_stream_.get(c))
      throw_error("Unexpected eof in the middle of reading an int");
    if (c == RIGHT_PAREN)
      return false;
    real_stream_.putback(c);
    if (!(real_stream_ >> d)) 
      throw_error("Error while reading double");
    return true;
  }

  void Istream_skip_comments::go_to_eof() {
    skip_comments_();
    if (!real_stream_.eof()) 
      throw_error("Symbols in the file beyond eof");
  }


  // This is a function object used for a map whose key is a string.
  // We want to compare strings disregarding case.

  class Nocase {
  public:
    bool operator() (const string& x, const string& y) const {
      return compare_nocase(x,y) < 0;
    }
  };

  
  // ------------------------------------------------------------------
  // Class Face_lookup_map_type
  // This is a map from strings to fspecs.
  // It generates errors upon duplication, and so on.  It is used
  // for keeping track of face names when a brep is read in.



  class Face_lookup_map_type {
  private:
    typedef map <string, Brep::Face_Spec, Nocase> Actual_Map_type;
    Actual_Map_type actual_map_;
  public:
    Face_lookup_map_type() { }
    ~Face_lookup_map_type() { }
    void insert(const string& facename, const Brep::Face_Spec& fspec) {
      const char* tmp = facename.c_str();
      string tmp2 = string(tmp);
      Actual_Map_type::value_type new_rec(tmp2, fspec);
      pair<Actual_Map_type::iterator, bool> rval = 
        actual_map_.insert(new_rec);

      if (!rval.second) {
        string errmsg = string("Duplicate face name ") + facename
          + " occurs in brep";
        throw_error(errmsg);
      }
      return;
    }

    Brep::Face_Spec find(const string& facename) const {
      Actual_Map_type::const_iterator it =
        actual_map_.find(facename);
      if (it == actual_map_.end()) {
        string errmsg = string("Face ") + facename  + 
          " listed as a subface but is not found";
        throw_error(errmsg);
      }
      return it -> second;
    }
  };

  class Bkpt_obj {
  private:
    int j;
  public:
    Bkpt_obj(int i) : j(i) { }
    ~Bkpt_obj();
  };

  Bkpt_obj::~Bkpt_obj() {
    int k = j + 1;
  }
}



// ------------------------------------------------------------------
// operator<< for breps.
// This routine outputs a brep to a stream in Ascii form.


QMG::ostream& operator<<(QMG::ostream& ostr, const QMG::Brep& g) {

  // output header info

  Ostream_with_wraparound os(ostr);
  os << g.io_header_code() << " " << g.gdim() << " " 
    << g.embedded_dim() << "\n(";

  // output brep prop vals
  
  // Output prop-val pairs
  for (Brep::Loop_over_propvals_of_brep pvloop0(g);
  pvloop0.notdone();
  ++pvloop0) {
    os << " "<< pvloop0.prop() << " " << pvloop0.val() << "\n";
  }

  os << ")\n(\n";
  {
    // Output control points
    for (Brep::ControlPointIndex j = 0; j < g.num_control_points(); ++j) {
      for (int d = 0; d < g.embedded_dim(); ++d) {
        os << g.control_point(j,d) << " ";
      }
      os << nobreak_begin << "# cp " << j << "\n" << nobreak_end;
    }
  }
  os << ") \n";
  for (int dim = 0; dim <= g.gdim(); ++dim) {
    // Output topological faces
    os << nobreak_begin << "(   # begin topological ";  
    const char* facetype;
    if (dim == 0)
      facetype = "vertices";
    else if (dim == 1)
      facetype = "edges";
    else if (dim == 2 && g.embedded_dim() == 3)
      facetype =  "surfaces";
    else 
      facetype = "regions";
    os << facetype << "\n" << nobreak_end;
    for (Brep::Face_Spec_Loop_Over_Faces_Of_Dim fspec(g, dim);
    fspec.notdone();
    ++fspec) {
      // Output prop-val pairs
      os << g.face_name(fspec) << "  (";
      for (Brep::Loop_over_propvals_of_face pvloop(g, fspec);
      pvloop.notdone();
      ++pvloop) {
        os << " "<< pvloop.prop() << " " << pvloop.val() << "\n";
      }
      os << ")  (";
      // Output bounding faces
      {
        for (Brep::Face_Spec_Loop_Over_Face_Subfaces subfspec(g, fspec);
        subfspec.notdone();
        ++subfspec) {
          if (!subfspec.is_internal_boundary()) {
            os << nobreak_begin;
            if (subfspec.orientation() == 0)
              os << "+";
            else if (subfspec.orientation() == 1)
              os << "-";
            os << g.face_name(subfspec) << nobreak_end << " ";
          }
        }
      }
      // Output internal boundary faces
      os << ") (";
      {
        for (Brep::Face_Spec_Loop_Over_Face_Subfaces subfspec(g, fspec);
        subfspec.notdone();
        ++subfspec) {
          if (subfspec.is_internal_boundary()) {
            os << g.face_name(subfspec) << " ";
          }
        }
      }
      // Output patches
      os << ")  (";
      for (Brep::Loop_over_patches_of_face patch(g, fspec);
      patch.notdone();
      ++patch) {
        if (patch.index() > 0)
          os << "\n   ";
        os << " (";
        if (fspec.fdim() == 0) {
          os << "vertex " << patch.control_point(0) << " ";
        }
        else if (fspec.fdim() == 1) {
          os << "bezier_curve " << patch.degree1() << " ";
          for (int j = 0; j < patch.degree1() + 1; ++j) {
            os << patch.control_point(j) << " ";
          }
        }
        else if (fspec.fdim() == 2) {
          if (patch.ptype() == BEZIER_QUAD) {
            os << "bezier_quad " << patch.degree1() << " " << patch.degree2() << " ";
            int k = 0;
            for (int j = 0; j < patch.degree2() + 1; ++j) {
              for (int i = 0; i < patch.degree1() + 1; ++i) {
                os << patch.control_point(k++) << " ";
              }
            }
          }
          else {
            os << "bezier_triangle " << patch.degree1() << " ";
            int ncp = (patch.degree1() + 1) * (patch.degree1() + 2) / 2;
            for (int j = 0; j < ncp; ++j) {
              os << patch.control_point(j) << " ";
            }
          }
        }
        os << ")";
      }
 
      os << " )\n";
    }
    os << nobreak_begin << " ) # end topological " << facetype << "\n" << nobreak_end;
  }
  return ostr;
}



// ------------------------------------------------------------------
// operator<< for meshes.
// This routine outputs a mesh to a stream in Ascii form.


QMG::ostream& operator<<(QMG::ostream& os, const QMG::SimpComplex& sc) {
  int di = sc.embedded_dim();
  int gdim = sc.gdim();
  os << sc.io_header_code() << " " << gdim << " " << di << "\n(";
  int npv = sc.num_prop_val();
  for (int pvi = 0; pvi < npv; ++pvi) {
    os << sc.prop(pvi) << " " << sc.val(pvi);
  }
  os << ")\n( # begin list of mesh nodes\n";
  for (SimpComplex::VertexOrdinalIndex vnumo = 0; vnumo < sc.num_nodes(); ++vnumo) {
    os <<  sc.ordinal_to_global(vnumo) << " ";
    for (int ii = 0; ii < di; ++ii)  {
      os <<  std::setprecision(17) << sc.real_coord_o(vnumo, ii) << " ";
    }
    os << '\n';
  }
  os << ") # end list of mesh nodes\n";
  for (int fdim = 0; fdim <= gdim; ++fdim) {
    os << "\n( # begin mesh info for topological entities of dimension " << fdim << '\n';
    for (Brep::FaceIndex faceind = 0; faceind < sc.brep_levelsize(fdim); ++faceind) {
      Brep::Face_Spec fspec(fdim, faceind);
      os << "(";
      if (fdim > 0 && fdim < di)
        os << " # mesh vertices lying on top. entity " << fspec << '\n';
      int nnf = sc.num_node_on_face(fspec);
      for (int seqno1 = 0; seqno1 < nnf; ++seqno1) {
        os << sc.node_on_face(fspec, seqno1) << " ";
        if (fdim > 0) {
          os << sc.patchind_on_face(fspec, seqno1) << " ";
          for (int l = 0; l < fdim; ++l) {
            os << std::setprecision(17) << sc.param_coord_on_face(fspec, seqno1, l)
              << " ";
          }
          os << '\n';
        }
      }
      os << ") (";
      if (fdim > 0) {
        os << " # mesh entities lying on top. entity " << fspec << '\n';
        int nmf = sc.num_meshface_on_face(fspec);
        for (int seqno2 = 0; seqno2 < nmf; ++seqno2) {
          for (int l = 0; l < fdim + 1; ++l) {
            os << sc.node_of_meshface_on_face(fspec, seqno2, l) << " ";
          }
          os << '\n';
        }
      }
      os << ") # " << fspec << '\n';
    }
    os << ") # end mesh info for topological entities of dimension " << fdim << '\n';
  }
  return os;
}   


// ------------------------------------------------------------------
// Read a brep from an istream.
// Second argument is true if the header code has already been read
// by another reader.  Third argument indicates the number of lines of
// the file already read.

QMG::Brep_Under_Construction
QMG::Brep_Under_Construction::read_from_istream(istream& istr, 
                                           bool code_already_read,
                                           int& line_count) {

  // current_op keeps track of the current operation of the parser;
  // used for issuing intelligent error messages.

  
  Error_Message errinfo;
  errinfo.set_string("Error occurred while reading header info");
  Istream_skip_comments is(istr, line_count);
  if (!code_already_read) {
    string code;
    is.get_string(code);
    if (compare_nocase(code, Brep::io_header_code()))
      throw_error("File begins with incorrect code");
  }
  int gdim, embedded_dim;
  if (!is.get_int(gdim)) throw_error("Misplaced right paren");
  if (!is.get_int(embedded_dim)) throw_error("Misplaced right paren");
  errinfo.set_string("Error occurred while brep's prop-val pairs");
  
  Brep_Under_Construction newbrep(gdim, embedded_dim);


  Propval_inserter pvi(newbrep);
  is.get_left_paren();
  for (;;) {
    string prop, val;
    if (!is.get_string(prop)) break;
    bool s2 = is.get_string(val);
    if (!s2) 
      throw_error("Unmatched property (property-values must come in pairs)");
    pvi.insert_propval(prop,val);
  }
  
  errinfo.set_string("Error occurred while reading control points");
  is.get_left_paren();
  Control_point_inserter cpi(newbrep);
  vector<Real> coord(embedded_dim, 0.0);
  for (;;) {
    bool double_read;
    for (int j = 0; j < embedded_dim; ++j) {
      double_read = is.get_double(coord[j]);
      if (!double_read) {
        if (j > 0) {
          string errmsg = "number of doubles in control point list";
          errmsg += " must be divisible by embedded_dim";
          throw_error(errmsg);
        }
        break;
      }
    }
    if (!double_read) break;
    cpi.insert_control_point(coord);
  }
  // This map is the associative map FaceName -> Fspec since
  // the ascii format refers to topological faces via their
  // symbolic names.
  
  Face_lookup_map_type facename_lookup;
 

  for (int dim = 0; dim <= gdim; ++dim) {
    const char* facetype = "unknown";
    if (dim == 0)
      facetype = "vertex";
    else if (dim == 1)
      facetype = "edge";
    else if (dim == 2)
      facetype = "surfaces";
    else if (dim == 3)
      facetype = "region";
    {
      ostringstream ost1;
      ost1 << "Error occurred while reading " << facetype << " list";
      errinfo.set_string(ost1.str());
    }
    Top_face_inserter tfi(newbrep, dim);
    is.get_left_paren();
    for (;;) {
      string facename;
      if (!is.get_string(facename)) break;
      Brep::Face_Spec fspec = tfi.insert_top_face(facename);
      facename_lookup.insert(facename, fspec);
      {
        ostringstream ost1;
        ost1 << "Error occurred while reading property-value list of topological " <<
          facetype << " \"" << facename << "\"";
        errinfo.set_string(ost1.str());
      }
      
      // Insert property value pairs.
      
      Top_face_propval_inserter tpvi(newbrep, fspec);
      is.get_left_paren();
      for (;;) {
        string prop, val;
        if (!is.get_string(prop)) break;
        bool s2 = is.get_string(val);
        if (!s2) 
          throw_error("Unmatched property (property-values must come in pairs)");
        tpvi.insert_propval(prop,val);
      }
      
      // Insert bounding faces.  Note: bounding faces may be prefixed
      // by + or - to indicate orientation.
      
      {
        ostringstream ost1;
        ost1 << "Error occurred while reading boundaries of topological " <<
          facetype << " \"" << facename << "\"";
        errinfo.set_string(ost1.str());
      }
      Top_face_bound_inserter tbi(newbrep, fspec);
      is.get_left_paren();
      for (;;) {
        string subfacename;
        if (!is.get_string(subfacename)) break;
        int ori = 2;
        if (subfacename[0] == '+') {
          ori = 0;
          subfacename = subfacename.substr(1, subfacename.length() - 1);
        }
        else if (subfacename[0] == '-') {
          ori = 1;
          subfacename = subfacename.substr(1, subfacename.length() - 1);
        }
        
        Brep::Face_Spec subfspec = facename_lookup.find(subfacename);
        tbi.insert_boundary(subfspec, ori);
      }
      
      
      // Insert internal bounding faces.
      {
        ostringstream ost1;
        ost1 << "Error occurred while reading internal boundaries of topological "
          << facetype << " \"" << facename << "\"";
        errinfo.set_string(ost1.str());
      }
      Top_face_ib_inserter ibi(newbrep, fspec);
      is.get_left_paren();
      for (;;) {
        string subfacename;
        if (!is.get_string(subfacename)) break;
        Brep::Face_Spec subspec = facename_lookup.find(subfacename);
        if (subspec.fdim() == dim - 1) {
          tbi.insert_boundary(subspec, 2);
          tbi.insert_boundary(subspec, 2);
        }
        else {
          ibi.insert_ib(subspec);
        }
      }
      // Insert patches
      {
        ostringstream ost1;
        ost1 << "Error occurred while reading patches of topological " << facetype
          << " \"" << facename << "\"";
        errinfo.set_string(ost1.str());
      }
      Top_face_patch_inserter pti(newbrep, fspec);
      is.get_left_paren();
      vector<Brep::ControlPointIndex> cp;
      int patchcount = 0;
      for (;;) {
        string patchstr;
        if (!is.get_string(patchstr)) break;
        ++patchcount;
        istringstream is2a(patchstr);
        Istream_skip_comments is2(is2a);
        is2.get_left_paren();
        string patchcode;
        bool string_read2 = is2.get_string(patchcode);
        if (!string_read2)
          throw_error("Missing patch type name");
        PatchType ptype = BEZIER_TRIANGLE;
        if (dim == 0) {
          if (compare_nocase(patchcode, "vertex") != 0)
            throw_error("Patch type must be 'vertex' for topological vertices");
        }
        else if (dim == 1) {
          if (compare_nocase(patchcode, "bezier_curve") != 0)
            throw_error("Patch type must be 'bezier_curve' for topological edges");
        }
        else if (dim == 2) {
          if (compare_nocase(patchcode, "bezier_quad") == 0)
            ptype = BEZIER_QUAD;
          else if (compare_nocase(patchcode, "bezier_triangle") == 0)
            ptype = BEZIER_TRIANGLE;
          else
            throw_error("Patch type must be either bezier_quad or bezier_triangle\n"
            "for topological surface");
        }
        else {
          throw_error("No patches allowed here");
        }
        int degree1, degree2;
        if (dim >= 1) {
          bool int_read = is2.get_int(degree1);
          if (!int_read) throw_error("Order (integer) missing");
        }
        if (dim == 2 && ptype == BEZIER_QUAD) {
          bool int_read = is2.get_int(degree2);
          if (!int_read) throw_error("Order (integer) missing (2)");
        }
        cp.resize(0);
        for (;;) {
          Brep::ControlPointIndex cpi;
          if (!is2.get_int(cpi)) break;
          cp.push_back(cpi);
        }
        pti.insert_patch(degree1, degree2, ptype, cp);
      }
      if (dim == 0 && patchcount != 1) {
        throw_error("Topological vertex must have exactly one patch");
      }
    }
  }
  line_count = is.linecount();
  return newbrep;
}


// ------------------------------------------------------------------
// Read a mesh from an istream.
// Second argument is true if the header code has already been read
// by another reader.  Third argument indicates the number of lines of
// the file already read.

QMG::SimpComplex_Under_Construction
QMG::SimpComplex_Under_Construction::
read_from_istream(istream& istr, 
                  bool code_already_read,
                  int& line_count) {

  // errinfo keeps track of the current operation of the parser;
  // used for issuing intelligent error messages.
  
  Error_Message errinfo;
  errinfo.set_string("Error occurred while reading header info");
  Istream_skip_comments is(istr, line_count);
  if (!code_already_read) {
    string code;
    is.get_string(code);
    if (compare_nocase(code, SimpComplex::io_header_code()))
      throw_error("File begins with incorrect code");
  }
  int gdim, embedded_dim;
  if (!is.get_int(gdim)) throw_error("Misplaced right paren");
  if (!is.get_int(embedded_dim)) throw_error("Misplaced right paren");


  errinfo.set_string("Error occurred while mesh's prop-val pairs");
  vector<int> levsize(gdim + 1, 0);
  
  SimpComplex_Under_Construction newsc(gdim, embedded_dim, levsize);

  is.get_left_paren();
  for (;;) {
    string prop, val;
    if (!is.get_string(prop)) break;
    bool s2 = is.get_string(val);
    if (!s2) 
      throw_error("Unmatched property (property-values must come in pairs)");
    newsc.add_prop_val(prop,val);
  }

  errinfo.set_string("Error occurred while reading node coordinates");
  is.get_left_paren();
  int vnum = 0;
  vector<Real> realcoord(embedded_dim);
  for (;;) {
    SimpComplex::VertexGlobalIndex vnum;
    if (!is.get_int(vnum))
      break;
    for (int j = 0; j < embedded_dim; ++j) {
      if (!is.get_double(realcoord[j]))
        throw_error("Missing or unbalanced number of doubles (real node coords)");
    }
    newsc.add_vertex(realcoord,vnum);
    
  }
  vector<Real> paramcoord;
  vector<SimpComplex::VertexGlobalIndex> vlist;
  for (int fdim = 0; fdim <= gdim; ++fdim) {
    is.get_left_paren();
    Brep::FaceIndex faceind = 0;
    paramcoord.resize(fdim);
    vlist.resize(fdim + 1);
    for (;;) {
      Brep::Face_Spec fspec(fdim,faceind);
      {
        ostringstream ost1;
        ost1 << "Error occurred while reading nodes incident on top. entity" << fspec;    
        errinfo.set_string(ost1.str());
      }
      if (!is.get_left_paren2()) break;
      newsc.add_fspec(fspec);
      for (;;) {
        SimpComplex::VertexGlobalIndex vnum;
        if (!is.get_int(vnum))
          break;
        Brep::PatchIndex patchind = 0;
        if (fdim > 0) {
          if (!is.get_int(patchind))
            throw_error("Missing integer (curve/patch index)");
        }
        for (int j = 0; j < fdim; ++j) {
          if (!is.get_double(paramcoord[j]))
            throw_error("Missing double (param coord of node on top. entity)");
        }
        newsc.add_vertex_bdryinc(vnum, fspec, patchind, paramcoord);
      }
      {
        ostringstream ost1;
        ost1 << "Error occurred while reading mesh entities lying on top. entity" << fspec;    
        errinfo.set_string(ost1.str());
      }
      is.get_left_paren();
      for (;;) {
        bool done = false;
        for (int j = 0; j < fdim + 1; ++j) {
          if (!is.get_int(vlist[j])) {
            if (j == 0) {
              done = true;
              break;
            }
            else {
              throw_error("Wrong number of integers in list of mesh entities");
            }
          }
        }
        if (done) break;
        newsc.add_simplex_face(vlist, fspec);
      }
      ++faceind;
    }
  }
  line_count = is.linecount();
  return newsc;
}


// Reads just the header code from an istream, and determines whether
// the file contains a mesh or a brep.  It keeps track of the number
// of lines read (for the purpose of useful error messages).

QMG::Object_Type_Code
QMG::read_header_from_istream(istream& istr, int& line_count) {
  
  Error_Message errinfo;
  errinfo.set_string("Error occurred while reading file type");
  Istream_skip_comments is(istr, line_count); 
  string code;
  is.get_string(code);
  line_count = is.linecount();
  if (compare_nocase(code, Brep::io_header_code()) == 0)
    return BREP;
  else if (compare_nocase(code, SimpComplex::io_header_code()) == 0)
    return SIMPCOMPLEX;
  else
    return UNKNOWN;
}

