// ------------------------------------------------------------------
// align.cpp
//
// This file contains the routines for aligning boxes in an orbit.
// Aligning means finding the close point on the brep entity to the
// face, and then making sure the close point is consistent between
// neighboring boxes.  Then vertices of the simplicial complex 
// are generated.
// ------------------------------------------------------------------
// Author: Stephen A. Vavasis
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------


#ifdef _MSC_VER
#if _MSC_VER == 1200
#include "qmatvec.h"
#endif
#endif

#include "qboxstack.h"
#include "qlog.h"
#include "qmg_gui.h"



// Routines defined in this file:

namespace QMG {
  namespace MG {
    using namespace QMG;

    void alignInitStaticData(int di); 

    void align_orbit(ActiveBoxVec& alignvec,
      int closefacedim,
      Real maxwarpdist,
      ActiveBoxVec& idlevec,
      ActiveBoxVec& smallboxvec,
      ActiveBoxVec& small_idle_boxvec,
      const Brep::Face_Spec& alignfspec,
      BoxFaceDag& dag,
      SimpComplex_Under_Construction& sc,
      vector<Brep::Face_Spec>& vtx_owner,
      const vector<Real>& asp_degrade,
      Logstream& logstr,
      Meshgen_gui& gui,
      const PatchTable& patchtable,
      IncidenceTable& inc_table,
      ActiveBoxVec::Create_Subbox_Workspace& wkspa);

    void align_orbit_lastphase(ActiveBoxVec& alignvec,
      ActiveBoxVec& smallboxvec,
      const Brep::Face_Spec& alignfspec,
      BoxFaceDag& dag,
      SimpComplex_Under_Construction& sc,
      vector<Brep::Face_Spec>& vtx_owner,
      Logstream& logstr,
      Meshgen_gui& gui);
  }
}



namespace {
  using namespace QMG;
  using namespace QMG::MG;

  ostream* debug_ostream;

  // ------------------------------------------------------------------
  // struct PriorityQueueKey
  // Priority for alignment: align highest-priority close faces first.
  // Priority queue also used so that neighboring boxes find each other.

  struct PriorityQueueKey {
    Real dist;
    Mask flatdim;
    IncidenceTable::Index incind;
    ActiveBoxVec::Index boxind;
    Base3 rindex;
    int my_idx;
    static int base_idx;
    
    // High priority means closer distance or lower incidence index.
    
    bool operator<(const PriorityQueueKey& other) const;
  };
  
  bool PriorityQueueKey::operator<(const PriorityQueueKey& other) const {
    if (this -> dist > other.dist)
      return true;
    if (this -> dist < other.dist)
      return false;
    if ((UInt32)(this -> flatdim) > (UInt32) (other.flatdim))
      return true;
    if ((UInt32)(this -> flatdim) < (UInt32) (other.flatdim))
      return false;
    int cmpind = IncidenceTable::compare_index(this -> incind, other.incind);
    if (cmpind > 0)
      return true;
    if (cmpind < 0)
      return false;
    if (this ->  boxind > other.boxind)
      return true;
    if (this -> boxind < other.boxind)
      return false;
    if ((UInt32) (this -> rindex) > (UInt32) (other.rindex))
      return true;
    if ((UInt32) (this -> rindex) < (UInt32) (other.rindex))
      return false;
    if (my_idx < other.my_idx)
      return true;
    return false;
  }
  

  // ------------------------------------------------------------------
  // struct PriorityQueueKey_lastphase
  // In the last phase, the priority queue is different because there is
  // no close face.  Instead, priority queue is used solely to match neighboring
  // boxes.


  struct PriorityQueueKey_lastphase {
    IntCoord position;
    ActiveBoxVec::Index boxind;
    Base3 rindex;
    bool operator<(const PriorityQueueKey_lastphase& other) const;
    int my_idx;
    static int base_idx;
    static int di;
  };


  int PriorityQueueKey_lastphase::di;
  int PriorityQueueKey::base_idx;
  int PriorityQueueKey_lastphase::base_idx;
  
  bool PriorityQueueKey_lastphase::operator<(const PriorityQueueKey_lastphase& other) const {
    int sgndif = IntCoord::compare(position, other.position);
    if (sgndif > 0)
      return true;
    if (sgndif < 0)
      return false;
    if (boxind > other.boxind)
      return true;
    if (boxind < other.boxind)
      return false;
    if (static_cast<UInt32>(rindex) > static_cast<UInt32>(other.rindex))
      return true;
    if (static_cast<UInt32>(rindex) < static_cast<UInt32>(other.rindex))
      return false;
    if (my_idx < other.my_idx) return true;
    return false;
  };

  // ------------------------------------------------------------------
  // l1_dist_to_boxface
  // Figures out the distance from a box face to the warp point in the l1
  // norm.


  Real l1_dist_to_boxface(const ActiveBox& b, 
    Base3 rindex,
    IncidenceTable::Index inc,
    const IncidenceTable& inc_tab) {

    Real dist = 0.0;
    Point p = b.face_real_lowerleft(rindex);
    Real rwidth = b.real_width();
    const Point& q = inc_tab.real_coord(inc);
    int reldim = 0;
    for (int i = 0; i < b.embedded_dim(); ++i) {
      bool b_is_flat = b.flatdim(i);
      bool face_is_flat = b_is_flat || rindex[reldim] < 2;
      if (face_is_flat)
        dist += fabs(p[i] - q[i]);
      else {
        if (q[i] < p[i] || q[i] > p[i] + rwidth) {
          return BIG_REAL;
        }
      }
      if (!b_is_flat) ++reldim;
    }
    return dist;
  }

  // Abstract base class for incidence priority queues with
  // single virtual void method.

  class IncidencePriorityQueue_ABC {
  public:    
    virtual void insert_box_closefaces(const ActiveBox& b) = 0;
    IncidencePriorityQueue_ABC() { }
    virtual ~IncidencePriorityQueue_ABC() { }
  private:
    // no copying
    IncidencePriorityQueue_ABC(const IncidencePriorityQueue_ABC&) { }
  };



  // Incidence priority queue.
  // Main data structure to align boxes in the orbit to incidences.  
  // Use this priority
  // queue to match up incidences from different boxes with each other, and to
  // favor the highest-priority incidences.  The highest priority
  // incidences are the ones with the smallest warp distance.


  class IncidencePriorityQueue : public IncidencePriorityQueue_ABC {
  private:
    priority_queue<PriorityQueueKey> pq_;
    int closefacedim_;
    Real maxwarpdist_act_;
    Brep::Face_Spec alignfspec_;
    const PatchTable& patchtable_;
    const IncidenceTable& inc_table_;
    Logstream& logstr_;
  public:
    IncidencePriorityQueue(int closefacedim,
      Real maxwarpdist_act,
      Brep::Face_Spec alignfspec,
      const PatchTable& patchtable,
      const IncidenceTable& inc_table,
      Logstream& logstr) :
         closefacedim_(closefacedim),
      maxwarpdist_act_(maxwarpdist_act),
      alignfspec_(alignfspec),
      patchtable_(patchtable),
      inc_table_(inc_table),
      logstr_(logstr)
    { }
    void insert_box_closefaces(const ActiveBox& b);
    PriorityQueueKey top() const {return pq_.top();}
    void pop() {pq_.pop();}
    int size() const {return pq_.size();}
  };


  // ------------------------------------------------------------------
  // insert_box_closeface_into_queue
  // Check which faces of a box are close to the input face.
  // Insert the close ones into the priority queue.

  void IncidencePriorityQueue::insert_box_closefaces(const ActiveBox& b) {
    
    StatVector<Real, MAXDIM_3EXP> dists;
    StatVector<IncidenceTable::Index, MAXDIM_3EXP> closefaces;
    int bdim = b.dim();
    int bdimpower3 = Base3::power3(bdim);
    int di = b.embedded_dim();
    if (logstr_.verbosity() >= 4)
      logstr_.str() << "For box " << b << " close faces are ";
    {
      for (int ri = 0; ri < bdimpower3; ++ri) {
        dists[ri] = BIG_REAL;
      }
    }
    for (ActiveBox::Loop_over_incidences incit(b, patchtable_, alignfspec_);
    incit.notdone();
    ++incit) {
      IncidenceTable::Index inc = incit.index();
      if (inc_table_.flatdim(inc).dim() != di - alignfspec_.fdim()) continue;
      Real bestdist = BIG_REAL;
      Base3 bestri = 0;
      {
        for (int ri = 0; ri < bdimpower3; ++ri) {
          Base3 rindex = ri;
          if (rindex.dim() == closefacedim_) {
            Real dist = l1_dist_to_boxface(b, rindex, inc, inc_table_);
            if (logstr_.verbosity() >= 7) {
              logstr_.str() << "   face = ";
              rindex.output(logstr_.str(), bdim);
              logstr_.str() << " incidence = incd#" << inc_table_.seqno(inc)
                << "# distance = " << dist << " mwd = " << maxwarpdist_act_
                << " dists[ri] = " << dists[ri] << '\n';
            }
            if (dist < bestdist && dist <= maxwarpdist_act_) {
              bestdist = dist;
              bestri = ri;
            }
          }
        }
      }
      if (bestdist < dists[bestri] && bestdist <= maxwarpdist_act_) {
        dists[bestri] = bestdist;
        closefaces[bestri] = inc;
      }
    }
    
    {
      PriorityQueueKey key;
      key.boxind = b.my_index();
      for (int ri = 0; ri < bdimpower3; ++ri) {
        Base3 rindex = ri;
        if (rindex.dim() == closefacedim_) {
          Real dist = dists[ri];
          if (dist <= maxwarpdist_act_) {
            key.dist = dist;
            key.rindex = rindex;
            
            // Compute the flatdim mask of the face of b indexed by rindex.
            
            key.flatdim = b.flatdim();
            int actual_dim = -1;
            for (int j = 0; j < bdim; ++j) {
              while (b.flatdim(++actual_dim));
              if (rindex[j] < 2)
                key.flatdim.set_bit(actual_dim);
            }
            key.incind = closefaces[ri];
            key.my_idx = ++key.base_idx;
            if (logstr_.verbosity() >= 4) {
              logstr_.str() << '{';
              key.rindex.output(logstr_.str(), bdim);
              logstr_.str() << ',' << key.dist 
                << ", incd#" << inc_table_.seqno(key.incind) << "#} ";
            }
            pq_.push(key);
          }
        }
      }
      if (logstr_.verbosity() >= 4)
        logstr_.str() << '\n';
    }
  }




  // Incidence priority queue for last phase.
  // Main data structure to align boxes in the orbit to incidences.  
  // Use this priority
  // queue to match up incidences from different boxes with each other, and to
  // favor the highest-priority incidences.  The highest priority
  // incidences are the ones with the smallest warp distance.


  class IncidencePriorityQueue_LastPhase : public IncidencePriorityQueue_ABC {
  private:
    priority_queue<PriorityQueueKey_lastphase> pq_;
    Logstream& logstr_;
  public:
    explicit IncidencePriorityQueue_LastPhase(Logstream& logstr) :
      logstr_(logstr)
    { }
    void insert_box_closefaces(const ActiveBox& b);
    PriorityQueueKey_lastphase top() const {return pq_.top();}
    int size() const {return pq_.size();}
    void pop() {pq_.pop();}
  };

  
  // ------------------------------------------------------------------
  // insert_box_closefaces_into_queue_lastphase
  // Insert a box's closefaces into the priority queue during last-phase alignment.
  // In this case, all faces are close.  Insert the vertices only.

  void IncidencePriorityQueue_LastPhase::insert_box_closefaces(const ActiveBox& b) {
    
    PriorityQueueKey_lastphase key;
    key.boxind = b.my_index();
    for (ActiveBox::Loop_over_vertices vertit(b);
    vertit.notdone();
    ++vertit) {
      key.position = vertit.intcoord();
      key.rindex = vertit.rindex();
      key.my_idx = ++key.base_idx;
      pq_.push(key);
    }
  }



  // ------------------------------------------------------------------
  // compute_l1_closepoint
  // Compute the closest point (in the l1 norm) among nearby incidences
  // to a box face.

  Point compute_l1_closepoint(const Point& warp_point, 
    const ActiveBox& b, 
    Base3 warp_rindex) {
    Point returnval;
    Real rwidth = b.real_width();
    Point lowerleft = b.face_real_lowerleft(warp_rindex);
    int reldim = 0;
    for (int i = 0; i < b.embedded_dim(); ++i) {
      bool is_flat = b.flatdim(i);
      Real lb = lowerleft[i];
      Real ub = lowerleft[i] + ((is_flat || warp_rindex[reldim] < 2)? 0.0 : rwidth);
      Real wp = warp_point[i];
      if (wp < lb)
        returnval[i] = lb;
      else if (wp > ub)
        returnval[i] = ub;
      else
        returnval[i] = wp;
      if (!is_flat) ++reldim;
    }
    return returnval;
  }



  // ------------------------------------------------------------------
  // warp_ok
  // Check if a proposed warp is OK.  "Warp" means select a close point
  // that is not on the box face.  We check whether it is OK by testing
  // whether the volume of the simplices induced with previously protected
  // boxes diminishes too much.

  bool warp_ok(Real dist,
    const Point& warp_point,
    const ActiveBox& b,
    Base3 warp_rindex,
    const BoxFaceDag& dag,
    BoxFaceDag::Index linkparent,
    const SimpComplex_Under_Construction& sc,
    Real asp_degrade,
    Logstream& logstr,
    double scfac,
    double tol,
    // workspace
    vector<SimpComplex::VertexOrdinalIndex>& volist) {
    if (logstr.verbosity() >= 4) {
      logstr.str() << "In warp_ok for box " << b << " linkparent is bfdi#" 
        << b.link_parent() << "# ";
    }
    if (dist == 0.0) {
      if (logstr.verbosity() >= 4) {
        logstr.str() << " dist=0 return true\n";
      }
      return true;
    }

    int di = sc.embedded_dim();
    int bdim = b.dim();
    int bdimpower3 = Base3::power3(bdim);
    Real realwidth = b.real_width();
    Point unwarped_point = compute_l1_closepoint(warp_point, b, warp_rindex);

#ifdef DEBUGGING
    Real computed_dist = Point::l1dist(warp_point, unwarped_point); 
    if (fabs(computed_dist - dist) > 1e3 * scfac * tol) {
      throw_error("Problem computing unwarped point");
    }
#endif

    Point currentpoint;
    Real matrix1_space[2 * MAXDIM * MAXDIM];
    Matrix matrix1(di, di, matrix1_space);
    Matrix matrix2(di, di, &matrix1_space[MAXDIM * MAXDIM]);



    for (int ri = 0; ri < bdimpower3; ++ri) {
      Base3 face_rindex = ri;
      if (bdim > 0) {
        if (face_rindex.dim() != bdim - 1 ||
          Base3::merge(face_rindex, warp_rindex) == face_rindex)
          continue;
      }
      currentpoint = b.face_real_lowerleft(face_rindex);

      // Row 0 gets overwritten below if bdim = 0.

      for (int jj = 0; jj < di; ++jj) {
        matrix1(0,jj) = currentpoint[jj] - warp_point[jj];
        matrix2(0,jj) = currentpoint[jj] - unwarped_point[jj];
      }
      int matrixrow = 1;
      {
        int actual_dim = -1;
        for (int ii = 0; ii < bdim; ++ii) {
          while (b.flatdim(++actual_dim));
          if (face_rindex[ii] < 2) continue;
          currentpoint[actual_dim] += realwidth;
          for (int jj = 0; jj < di; ++jj) {
            matrix1(matrixrow, jj) = currentpoint[jj] - warp_point[jj];
            matrix2(matrixrow, jj) = currentpoint[jj] - unwarped_point[jj];
          }
          ++matrixrow;
        }
      }
#ifdef DEBUGGING
      if (bdim > 0 && matrixrow != bdim)
        throw_error("Inconsistent matrixrow in align");
#endif

      if (linkparent < 0) {
#ifdef DEBUGGING
        if (bdim != di)
          throw_error("Inconsistent dimensions in align");
#endif
        Real dt1 = matrix1.determinant();
        Real dt2 = matrix2.determinant();
        if (logstr.verbosity() >= 4) {
          logstr.str() << " rind = ";
          warp_rindex.output(logstr.str(), bdim);
          logstr.str() << " facerindex = ";
          face_rindex.output(logstr.str(), bdim);
          if (logstr.verbosity() >= 7) {
            logstr.str() << '\n' << matrix1 << '\n' << matrix2;
          }
          logstr.str() << " dets = " << dt1 << ',' << dt2 << " asp_degrade = " << asp_degrade;
        }

        if (dt1 * dt2 <= 0.0 || fabs(dt1) < asp_degrade * fabs(dt2)) {
          if (logstr.verbosity() >= 4)
            logstr.str() << " return false 1 \n";
          return false;
        }
      }
      else {
        
        int scrap1, scrap2;
        dag.trace_path_to_root(linkparent, volist, scrap1, scrap2);
        
#ifdef DEBUGGING
        if (volist.size() != di - bdim) 
          throw_error("Inconsistent dimensions in align 2");
#endif

        for (int ii = 0; ii < di - bdim; ++ii) {
          SimpComplex::VertexOrdinalIndex thisvert = volist[ii];
          for (int jj = 0; jj < di; ++jj) {
            Real parentcoord = sc.real_coord_o(thisvert, jj);
            matrix1(ii + bdim, jj) = parentcoord - warp_point[jj];
            matrix2(ii + bdim, jj) = parentcoord - unwarped_point[jj];
          }
        }
        Real dt1 = matrix1.determinant();
        Real dt2 = matrix2.determinant();
        if (logstr.verbosity() >= 4) {
          logstr.str() << " rind = ";
          warp_rindex.output(logstr.str(), bdim);
          logstr.str() << " face_rindex = ";
          warp_rindex.output(logstr.str(), bdim);
          logstr.str() << " anc vtx = ";
          for (int l1 = 0; l1 < di - bdim; ++l1)
            logstr.str() << "ord#" << volist[l1] << "# ";
          if (logstr.verbosity() >= 7) {
            logstr.str() << '\n' << matrix1 << '\n' << matrix2;
          }
          logstr.str() << " dets = " << dt1 << ',' << dt2 << " asp_degrade = " << asp_degrade;
        }            
        if (dt1 * dt2 <= 0.0 || fabs(dt1) < asp_degrade * fabs(dt2)) {
          if (logstr.verbosity() >= 4) 
            logstr.str() << " return false\n";
          return false;
        }
      }
    }
    if (logstr.verbosity() >= 4) {
      logstr.str() << " return true\n";
    }
    return true;
  }

  
  // ------------------------------------------------------------------
  // launch_get_subbox_weight
  // When a facet of a box is launched as a new box, this routine computes 
  // the weight of its faces based on the weight of the parent.

  Weight_t launch_get_subbox_weight(const ActiveBox& ab,
    int parentbdim,
    Base3 closeface,
    Base3 rindex,
    int reldim,
    int facetval) {
    
    int newind = 0;
    Base3 parent_rindex = 0;      
    int rindex_dim = 0;
    int overlap_count = 0;
    for (int parentind = 0; parentind < parentbdim; ++parentind) {
      int thisdigit = (parentind == reldim)? facetval : rindex[newind++];
      parent_rindex.set_digit(parentind, thisdigit);
      if (thisdigit < 2) {
        if (thisdigit == closeface[parentind])
          ++overlap_count;
      }
      else {
        ++rindex_dim;
      }
    }

    Weight_t newwt = ab.weight(parent_rindex) / (parentbdim - rindex_dim - overlap_count);
    return newwt;
  }

  
  // ------------------------------------------------------------------
  // class ActiveBox_For_Launch
  // This is an active box whose facets are about to be launched as
  // new active boxes.  This class contains extra information
  // about computing weights, etc.
  // Specialized postprocessing inserts new box close faces into priority queue.
    

    
  class ActiveBox_For_Launch : public ActiveBox_As_Parent {
  private:
    int bdim_;
    Base3 closeface_;
    int reldim_, facetval_;
    mutable Brep::Face_Spec chambers[MAXDIM][2];
    mutable bool is_front_[MAXDIM][2];
    Brep::Face_Spec alignfspec_;
    IncidencePriorityQueue_ABC& ipq_;
  public:
    ActiveBox_For_Launch(const ActiveBox& ab, 
      Base3 closeface,
      const Brep::Face_Spec& alignfspec,
      IncidencePriorityQueue_ABC& ipq) :
    ActiveBox_As_Parent(ab), 
    bdim_(ab.dim()),
    closeface_(closeface),
    alignfspec_(alignfspec),
    ipq_(ipq) {
      for (int i = 0; i < bdim_; ++i) {
        for (int j = 0; j < 2; ++j)
          chambers[i][j] = Brep::Face_Spec(-2,-2);
      }
    }
    void set_reldim_facetval(int r, int f) {
      reldim_ = r; facetval_ = f;
    }
    Weight_t get_subbox_weight(Base3 ri) const {
      return launch_get_subbox_weight(*this, bdim_, closeface_, ri, reldim_, facetval_);
    }
    pair<Brep::Face_Spec,bool> chamber_containing_subbox(const ActiveBox& testbox,
      ActiveBoxVec::Create_Subbox_Workspace& wkspa) const;
    virtual void specialized_postprocess_subbox(const ActiveBox& newb,
      int lowest_patch_dim) const {
      if (lowest_patch_dim <= alignfspec_.fdim())
        ipq_.insert_box_closefaces(newb);
    }
  };
  
  
  pair<Brep::Face_Spec, bool> 
    ActiveBox_For_Launch::
    chamber_containing_subbox(const ActiveBox& testbox,
    ActiveBoxVec::Create_Subbox_Workspace& wkspa)  const {
    bool found = false;
    for (int i = 0; i < bdim_; ++i) {
      if (i == reldim_) continue;
      
      for (int j = 0; j < 2; ++j) {
        if (chambers[i][j].faceind() > -2) {
          found = true;
          chambers[reldim_][facetval_] = chambers[i][j];
          is_front_[reldim_][facetval_] = is_front_[i][j];
          break;
        }
      }
      if (found) break;
    }
    if (!found) {
      pair<Brep::Face_Spec,bool> rval = 
        ActiveBox_As_Parent::chamber_containing_subbox(testbox, wkspa);
      chambers[reldim_][facetval_] = rval.first;
      is_front_[reldim_][facetval_] = rval.second;
    }
    return pair<Brep::Face_Spec,bool>(chambers[reldim_][facetval_], 
      is_front_[reldim_][facetval_]);
  }

  
  // ------------------------------------------------------------------
  // compute_facet_address
  // Computes a box-address for a facet of an active box; used to
  // launch the facets as new boxes.


  void compute_facet_address(BoxAddress& bxa, 
    const ActiveBox& b,
    Base3 subfaceindex) {
    bxa = b.boxaddress();
    int di = b.embedded_dim();
    int reldim = -1;
    for (int i = 0; i < di; ++i) {
      if (!b.flatdim(i)) {
        ++reldim;
        if (subfaceindex[reldim] < 2) {
          bxa.flatdim().set_bit(i);
          if (subfaceindex[reldim] == 1) {
            bxa.lowerleft().increment(i, b.iwidth());
          }
        }
      }
    }
  }



  // ------------------------------------------------------------------
  // protect_box_and_launch_subfaces
  // This routine protects a box and launches its facets as new boxes.



  void protect_box_and_launch_subfaces(const ActiveBox& b,
    SimpComplex::VertexOrdinalIndex vertnum_o,
    Base3 closeface_rindex,
    const Brep::Face_Spec& apex,
    ActiveBoxVec& orbit,
    ActiveBoxVec& idle_vec,
    BoxFaceDag& dag,
    IncidencePriorityQueue& ipq,
    ActiveBoxVec::Create_Subbox_Workspace& wkspa) {

    BoxFaceDag::Index newind = dag.add_protected_box(b, vertnum_o);
    if (wkspa.logstr.verbosity() >= 2) { ////
      wkspa.logstr.str() << "New protected box " << b << " face ";
      closeface_rindex.output(wkspa.logstr.str(), b.dim());
      wkspa.logstr.str() << " at bfdi#" << newind << "#  parent is bfdi#"
        << b.link_parent() << "#\n";
    }
    int bwidth = b.iwidth();
    int bdim = b.dim();
    int actualdim = -1;
    Base3 fullindex = Base3::power3(bdim) - 1;
    BoxAddress bxa;
    

    ActiveBox_For_Launch abl(b, closeface_rindex, apex, ipq);
    
    for (int reldim = 0; reldim < bdim; ++reldim) {
      while (b.flatdim(++actualdim));
      for (int facetval = 0; facetval < 2; ++facetval) {
        Base3 subface_rindex = fullindex.modify_digit(reldim, facetval);
        if (Base3::merge(subface_rindex, closeface_rindex) != subface_rindex) {
          compute_facet_address(bxa, b, subface_rindex);
          abl.set_reldim_facetval(reldim, facetval);
          ActiveBoxVec::create_subbox(abl, bxa, bwidth, apex.fdim(),
            newind, facetval, actualdim, orbit, idle_vec, wkspa);
        }
      }
    }
  }

  
  // ------------------------------------------------------------------
  // protect_box_and_launch_subfaces_lastphase
  // Same as the above, except we are in the last phase so
  // the create_subbox routine is simplified.


  void protect_box_and_launch_subfaces_lastphase(const ActiveBox& b,
    SimpComplex::VertexOrdinalIndex vertnum_o,
    Base3 closepoint_rindex,
    const Brep::Face_Spec& apex,
    ActiveBoxVec& orbit,
    BoxFaceDag& dag,
    IncidencePriorityQueue_LastPhase& ipq,
    Logstream& logstr) {
    
    BoxFaceDag::Index newind = dag.add_protected_box(b,vertnum_o);
    if (logstr.verbosity() >= 2) { ////
      logstr.str() << "New protected interior box ";
      if (logstr.verbosity() >= 4)
        b.full_dump(logstr.str());
      else
        logstr.str() << b;
      logstr.str() << " closeface = ";
      closepoint_rindex.output(logstr.str(),b.dim()); 
      logstr.str() << " at bfdi#" << newind << "#  parent is bfdi#"
        << b.link_parent() << "#\n";
    }
    int bwidth = b.iwidth();
    int bdim = b.dim();
    int actualdim = -1;
    Base3 fullindex = Base3::power3(bdim) - 1;
    BoxAddress bxa;

    
    ActiveBox_For_Launch abl(b, closepoint_rindex, apex, ipq);
    
    for (int reldim = 0; reldim < bdim; ++reldim) {
      while (b.flatdim(++actualdim));
      for (int facetval = 0; facetval < 2; ++facetval) {
        Base3 subface_rindex = fullindex.modify_digit(reldim, facetval);
        if (Base3::merge(subface_rindex, closepoint_rindex) != subface_rindex) {
          compute_facet_address(bxa, b, subface_rindex);          
          abl.set_reldim_facetval(reldim, facetval);
          ActiveBoxVec::create_subbox_completely_interior(abl, bxa, 
            bwidth, orbit,  newind, facetval, actualdim, logstr);
        }
      }
    }
  }
  
  // ------------------------------------------------------------------
  // struct DeduplicationMapKey
  // Used as a key for a map to deduplicate a box stack.  Boxes are
  // duplicates if they have the same address and the same highest
  // incidence number.

    struct DeduplicationMapKey {
    BoxAddress bxa;
    PatchTable::Index highest_inc;
    bool operator<(const DeduplicationMapKey& other) const;
  };

  bool DeduplicationMapKey::operator<(const DeduplicationMapKey& other) const {
    if (highest_inc < other.highest_inc)
      return true;
    if (highest_inc > other.highest_inc)
      return false;
    if ((UInt32) (bxa.flatdim()) < (UInt32) (other.bxa.flatdim()))
      return true;
    if ((UInt32) (bxa.flatdim()) > (UInt32) (other.bxa.flatdim()))
      return false;
    int sgndiff = IntCoord::compare(bxa.lowerleft(), other.bxa.lowerleft());
    if (sgndiff < 0)
      return true;
    return false;
  }

  
  typedef map<DeduplicationMapKey, ActiveBoxVec::Index> DeduplicationMap;
  

}

void QMG::MG::alignInitStaticData(int di) {
  PriorityQueueKey_lastphase::di = di;
}

#ifdef NSF_DEMO
namespace QMG {
  namespace Nsf_Demo {
    using namespace QMG;
    using namespace QMG::MG;
    extern void show_alignment(const ActiveBoxVec& vec,
      const vector<ActiveBoxVec::Index>& boxindvec,
      const Point& alignpoint);
  }
}
#endif


// ------------------------------------------------------------------
// align_orbit
// This routine aligns an orbit of a brep.  First it puts all the
// closefaces in a priority queue.  Then boxes whose closefaces
// are completely covered are protected, their subfaces launched,
// and a new vertex appears in the simplicial complex.




void QMG::MG::align_orbit(ActiveBoxVec& alignvec,
                          int closefacedim,
                          Real maxwarpdist,
                          ActiveBoxVec& idlevec,
                          ActiveBoxVec& smallboxvec,
                          ActiveBoxVec& small_idle_boxvec,
                          const Brep::Face_Spec& alignfspec,
                          BoxFaceDag& dag,
                          SimpComplex_Under_Construction& sc,
                          vector<Brep::Face_Spec>& vtx_owner,
                          const vector<Real>& asp_degrade,
                          Logstream& logstr,
                          Meshgen_gui& gui,
                          const PatchTable& patchtable, 
                          IncidenceTable& inc_table,
                          ActiveBoxVec::Create_Subbox_Workspace& wkspa) {

  debug_ostream = &logstr.str();

  PriorityQueueKey::base_idx = 0;
  

  Real bxwidth = BoxAddress::multiplier() / ( (Real)(1 << alignvec.iwidth()) );
  Real maxwarpdist_act = maxwarpdist * bxwidth;
  
  IncidencePriorityQueue ipq(closefacedim, maxwarpdist_act, alignfspec,
    patchtable, inc_table, logstr);

  // Put box closefaces into the queue.

  {
    for (ActiveBoxVec::Loop_over_boxes_nonconst loop(alignvec);
    loop.notdone();
    ++loop) {
      NonConstActiveBox b = loop.getbox_nonconst();
#ifdef DEBUGGING
      if (b.completely_interior()) {
        throw_error("Completely interior box in wrong orbit?");
        return;
      }
#endif
      ipq.insert_box_closefaces(b);
      b.assign_close_face(false);
      gui.check_update();
    }
  }
  
  vector<PriorityQueueKey> boxes_to_align(0);
#ifdef NSF_DEMO
  vector<ActiveBoxVec::Index> boxind_vec;
#endif

  bool too_far_detected = false;
  int di = patchtable.embedded_dim();
  int phase = alignfspec.fdim();
  vector<SimpComplex::VertexOrdinalIndex> volist(di + 1);
  vector<Real> warp_point_re(di);
  vector<Real> warp_point_pa(alignfspec.fdim());
  Point warp_point_pa0;

  // Loop over the priority queue, either until the distance gets 
  // so big that an unacceptable warp is found, or the queue is empty.

  while (!too_far_detected && ipq.size() > 0) {
    
    gui.check_update();
    
    
    // Copy all the highest priority box faces with the same incidence index
    // to stack boxes_to_align.

    boxes_to_align.resize(0);
    Mask flatdim1 = ipq.top().flatdim;
    IncidenceTable::Index incind1 = ipq.top().incind;
    do {
      if (!(alignvec.iskilled(ipq.top().boxind)))
        boxes_to_align.push_back(ipq.top());
      ipq.pop();
    } while (ipq.size() > 0 && ipq.top().flatdim == flatdim1 &&
      ipq.top().incind == incind1);
    int basize = boxes_to_align.size();
    if (basize == 0)
      continue;

    if (logstr.verbosity() >= 2) { ////
      logstr.str() << "Attempting to align boxes\n";
      for (int jj = 0; jj < basize; ++jj) {
        ActiveBoxVec::Index boxind = boxes_to_align[jj].boxind;
        const ActiveBox& b1 = alignvec.getbox(boxind);
        logstr.str() << "    " << b1 << " ";
        boxes_to_align[jj].rindex.output(logstr.str(), b1.dim());
      }
      logstr.str() << " to incidence incd#" << inc_table.seqno(incind1) << "#\n";
    }

    
    // Test all of these boxes for OK aspect ratio.
    
    Real dist1 = boxes_to_align[0].dist;
    if (dist1 == BIG_REAL) {
      if (logstr.verbosity() >= 2) { ////
        logstr.str() << " distance = " << dist1 << " too far\n";
      }
      too_far_detected = true;
      break;
    }


    // Check if we are trying to align a too low-dimensional box face.

    {
      for (int jj = 0; jj < basize; ++jj) {
        ActiveBoxVec::Index boxind = boxes_to_align[jj].boxind;
        const ActiveBox& b1 = alignvec.getbox(boxind);
        if (b1.dim() + phase < di) {
          too_far_detected = true;
#ifdef DEBUGGING
          if (dist1 <= maxwarpdist_act * .05 && logstr.verbosity() >= 1) {
            logstr.str() << "Warning: warping anomaly may lead to poor aspect ratio.\n"
              << "Suggest you re-run QMG with smaller curvature setting.";
          }
#endif
          break;
        }
      }
    }

    if (too_far_detected)
      break;

    Point warp_point = inc_table.real_coord(incind1);
    if (dist1 > 0.0) {
      for (int jj = 0; jj < basize; ++jj) {
        ActiveBox b = alignvec.getbox(boxes_to_align[jj].boxind);
        too_far_detected = !warp_ok(dist1, warp_point, b,
          boxes_to_align[jj].rindex, dag, b.link_parent(), sc, 
          asp_degrade[b.dim()], logstr, wkspa.scfac, wkspa.tol, volist);
        if (too_far_detected) {
          if (logstr.verbosity() >= 2) { ////
            logstr.str() << " distance = " << dist1 << " too far detected on " << jj << "th box\n";
          }
          break;
        }
      }
    }
    
    if (too_far_detected)
      break;
      
    // Check if we can align these boxes by adding up their weights.
    
    Weight_t sumweight = 0.0;
    for (int kk = 0; kk < basize; ++kk) {
      sumweight += alignvec.getbox(boxes_to_align[kk].boxind).weight(boxes_to_align[kk].rindex);
    }
    
    if (logstr.verbosity() >= 2) ////
      logstr.str() << "weightsum = " << sumweight << '\n';
    if (sumweight > 1.00001)
      throw_error("Excess weight sum");

    // If the weight sum is 1, protect all the boxes, launch subfaces,
    // and put a new vertex in the simplicial complex.

    if (sumweight > 0.99999) {
      PatchTable::Index patchind = inc_table.patchind(incind1);
      warp_point_pa0 = inc_table.param_coord(incind1);
      for (int ll = 0; ll < di; ++ll)
        warp_point_re[ll] = warp_point[ll];
      for (int mm = 0; mm < alignfspec.fdim(); ++mm)
        warp_point_pa[mm] = warp_point_pa0[mm];
      pair<SimpComplex::VertexOrdinalIndex, SimpComplex::VertexGlobalIndex> rval = 
        sc.add_vertex(warp_point_re);
      SimpComplex::VertexOrdinalIndex vertnum_o = rval.first;
      SimpComplex::VertexGlobalIndex vertnum = rval.second;
      sc.add_vertex_bdryinc(vertnum, alignfspec,
        patchtable.orig_index(patchind), warp_point_pa);
      if (vtx_owner.size() <= vertnum_o) {
        vtx_owner.resize(vertnum_o + 1, alignfspec);
      }
      else 
        vtx_owner[vertnum_o] = alignfspec;
      if (logstr.verbosity() >= 2) { ////
        logstr.str() << "Alignment successful; new vertex vtx#" << vertnum 
          << "#   ord#" << vertnum_o << " at coords " << warp_point << " on patch " << patchind << '\n';
      }
#ifdef NSF_DEMO
      boxind_vec.resize(basize);
      for (int jj2 = 0; jj2 < basize; ++jj2) {
        boxind_vec[jj2] = boxes_to_align[jj2].boxind;
      }
      QMG::Nsf_Demo::show_alignment(alignvec, boxind_vec, warp_point);
#endif
      for (int jj = 0; jj < basize; ++jj) {
        ActiveBoxVec::Index boxind = boxes_to_align[jj].boxind;
        ActiveBox b = alignvec.getbox(boxind);
        Real maxwarpdist_act = maxwarpdist * b.real_width();
        Base3 rindex = boxes_to_align[jj].rindex;
        protect_box_and_launch_subfaces(b, vertnum_o, rindex,
          alignfspec, alignvec, idlevec, dag, ipq, wkspa);
        alignvec.killbox(boxind, logstr);
      }

    } 
    else {
      // If the weight sum is < 1, then mark the boxes to indicate that they
      // had a close face.  This forces them to be split.

      for (int jj = 0; jj < basize; ++jj) {
        alignvec.getbox_nonconst(boxes_to_align[jj].boxind).assign_close_face(true);
        // check_containment(alignvec.getbox(boxes_to_align[jj].boxind), idlevec, logstr.str());
      }
    }
  }
  
  
  // All the boxes that have a close subface and weren't aligned must be split.
  // Must split if this is the last pass.

  bool last_pass = closefacedim == di - phase;
  {
    for (ActiveBoxVec::Loop_over_boxes_nonconst loop(alignvec);
    loop.notdone();
    ++loop) {
      gui.check_update();
      ActiveBox b = loop.getbox();
      if (b.has_close_face() || last_pass) {
        if (b.iwidth() >= MAXLEV) 
          throw_error("Oversplitting for alignment (1)");
        ActiveBoxVec::split_box_and_push_children(phase, alignvec, 
          smallboxvec, small_idle_boxvec, b,  inc_table,  wkspa);
        alignvec.killbox(loop.index(), logstr);
      }
    }
  }
}



// ------------------------------------------------------------------
// align_orbit_lastphase
// This aligns boxes during the last phase.  It's quite similar
// to align, except there are some simplifications because
// there is no warping.



void QMG::MG::align_orbit_lastphase(ActiveBoxVec& alignvec,
                                    ActiveBoxVec& smallboxvec,
                                    const Brep::Face_Spec& alignfspec,
                                    BoxFaceDag& dag,
                                    SimpComplex_Under_Construction& sc,
                                    vector<Brep::Face_Spec>& vtx_owner,
                                    Logstream& logstr,
                                    Meshgen_gui& gui) {
  
  
  PriorityQueueKey_lastphase::base_idx = 0;
  
  IncidencePriorityQueue_LastPhase ipq(logstr);
  int di = alignfspec.fdim();

  // Put all the vertices of the boxes into the priority queue.

  {
    for (ActiveBoxVec::Loop_over_boxes loop(alignvec);
    loop.notdone();
    ++loop) {
      gui.check_update();
      ActiveBox b = loop.getbox();
#ifdef DEBUGGING
      if (!b.completely_interior()) {
        throw_error("Not completely interior box in wrong orbit");
        return;
      }
#endif
      ipq.insert_box_closefaces(b);
    }
  }
  
  vector<PriorityQueueKey_lastphase> boxes_to_align(0);
#ifdef NSF_DEMO
  vector<ActiveBoxVec::Index> boxind_vec;
#endif

  BoxAddress bxa;
  vector<Real> new_point_re(di);
  vector<Real> new_point_pa(0);

  while (ipq.size() > 0) {
    
    // Copy all the highest priority box faces with the same incidence index
    // to stack boxes_to_align.

    gui.check_update();
    
    boxes_to_align.resize(0);
    bxa.lowerleft() = ipq.top().position;
    do {
      if (!alignvec.iskilled(ipq.top().boxind))
        boxes_to_align.push_back(ipq.top());
      ipq.pop();
    } while (ipq.size() > 0 && ipq.top().position == bxa.lowerleft());
    int basize = boxes_to_align.size();
    if (basize == 0)
      continue;

    if (logstr.verbosity() >= 2) { ////
      logstr.str() << "Attempting to (last phase) align boxes\n";
      for (int jj = 0; jj < basize; ++jj) {
        ActiveBoxVec::Index boxind = boxes_to_align[jj].boxind;
        const ActiveBox& b1 = alignvec.getbox(boxind);
        logstr.str() << "    " << b1 << " ";
        boxes_to_align[jj].rindex.output(logstr.str(), b1.dim());
        logstr.str() << '\n';
      }
    }
    
    // Check if we can align these boxes by adding up their weights.
    
    Weight_t sumweight = 0.0;
    for (int jj = 0; jj < basize; ++jj) {
      sumweight += alignvec.getbox(boxes_to_align[jj].boxind).weight(boxes_to_align[jj].rindex);
    }
        
    if (logstr.verbosity() >= 2) ////
      logstr.str() << "weightsum = " << sumweight << '\n';
    if (sumweight > 1.00001)
      throw_error("Excess weight sum last phase");
    if (sumweight > 0.99999) {
      Point pt = bxa.real_lowerleft();
      for (int ll = 0; ll < di; ++ll)
        new_point_re[ll] = pt[ll];

      pair<SimpComplex::VertexOrdinalIndex, SimpComplex::VertexGlobalIndex> rval = 
        sc.add_vertex(new_point_re);

      SimpComplex::VertexOrdinalIndex vertnum_o = rval.first;
      SimpComplex::VertexGlobalIndex vertnum = rval.second;
      if (vtx_owner.size() <= vertnum_o) {
        vtx_owner.resize(vertnum_o + 1, alignfspec);
      }
      else 
        vtx_owner[vertnum_o] = alignfspec;
      if (logstr.verbosity() >= 2) { ////
        logstr.str() << "Alignment successful; new vertex vtx#" << vertnum 
          << "# ord #" << vertnum_o << " at coords " << pt << '\n';
      }
#ifdef NSF_DEMO
      boxind_vec.resize(basize);
      for (int jj2 = 0; jj2 < basize; ++jj2) {
        boxind_vec[jj2] = boxes_to_align[jj2].boxind;
      }
      QMG::Nsf_Demo::show_alignment(alignvec, boxind_vec, pt);
#endif
      for (int jj = 0; jj < basize; ++jj) {
        int boxind = boxes_to_align[jj].boxind;
        ActiveBox b = alignvec.getbox(boxind);
        protect_box_and_launch_subfaces_lastphase(b, vertnum_o,
          boxes_to_align[jj].rindex, alignfspec, alignvec, dag, ipq, logstr);
        alignvec.killbox(boxind, logstr);
      }
    } 
  }
  
  // All the boxes that have a close subface and weren't aligned must be split.
  
  {
    for (ActiveBoxVec::Loop_over_boxes loop(alignvec);
    loop.notdone();
    ++loop) {
      gui.check_update();
      ActiveBox b = loop.getbox();
      if (b.iwidth() >= MAXLEV)
        throw_error("Oversplitting for alignment (2)");
      ActiveBoxVec::split_completely_interior_box_and_push_children(b,
        smallboxvec, logstr);
      alignvec.killbox(loop.index(), logstr);
    }
  }
}

