// ------------------------------------------------------------------
// double
//
// This file contains the driver function for QMG routine
// to (1) make patch orientations consistent.
//    (2) make internal boundaries doubled.
// The function takes either a brep, or a (brep,mesh) pair as
// input.
// ------------------------------------------------------------------
// Author: Stephen A. Vavasis
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------


#ifdef _MSC_VER
#if _MSC_VER == 1200
#include "qmatvec.h"
#endif
#endif



#ifdef _MSC_VER
#include "qnamesp.h"
static void unused_func() {
  using QMG::pair;
  pair<int,int> b;
  pair<int,int> a(3,2);
  b = a;
}
#endif


#include "qbrep_constr.h"
#include "qmatvec.h"
#include "qancestor.h"
#include "qsimpcomp.h"
#include "qpatchtab.h"
#include "qunionfind.h"
#include "qlog.h"
#include "qerr.h"


namespace QMG {
  namespace MG {
    using namespace QMG;

    // see qsmall.cpp
    extern void small_initialize_static_data(int di);


    // Functions defined in this file.
    Brep_Under_Construction
      double_brep(const Brep& b,
                  const vector<Brep::Face_Spec>& faces_to_double,
                  double tol,
                  // other return variables:
                  Brep::Face_Labeling<Brep::FaceIndex>& oldfspec_to_new_map,
                  Brep::Face_Labeling<int>& face_num_copies,
                  map<PatchInBrepAddress, bool>& patch_is_flipped);
    SimpComplex_Under_Construction
      double_simpcomp(const Brep& b,
                      const Brep& newb,
                      const SimpComplex& sc,
                      const Brep::Face_Labeling<Brep::FaceIndex>& oldfspec_to_new_map,
                      const Brep::Face_Labeling<int>& num_copies,
                      const map<PatchInBrepAddress,bool>& patch_is_flipped,
                      double tol
                      );

  }
#ifdef DEBUGGING
  namespace Local_to_PatchMath_CPP {
    extern bool dump_everything;
    extern ostream* debug_str;
  }
#endif
}



namespace {
  using namespace QMG;
  using namespace QMG::MG;


  struct NewBrepFaceInfo {
    vector<Brep::FaceIndex> childlist;
    vector<int> child_ori;
    vector<Brep::Face_Spec> iblist;
    NewBrepFaceInfo() { }
  private:
    NewBrepFaceInfo(const NewBrepFaceInfo&) { }
    void operator=(const NewBrepFaceInfo&) { }
  };

  class NewBrepType {
    int di_;
    vector<NewBrepFaceInfo*>* nbrf_;
  public:
    NewBrepFaceInfo& get_face(const Brep::Face_Spec& fspec) {
      return *(nbrf_[fspec.fdim()][fspec.faceind()]);
    }
    Brep::Face_Spec add_face(int fdim) {
      nbrf_[fdim].push_back(new NewBrepFaceInfo());
      return Brep::Face_Spec(fdim, nbrf_[fdim].size() - 1);
    }
    NewBrepType(int di) : 
      di_(di), 
      nbrf_(new vector<NewBrepFaceInfo*>[di + 1]) { }
    ~NewBrepType();
  };

  NewBrepType::~NewBrepType() {
    for (int fdim = di_; fdim >= 0; --fdim) {
      int sz = nbrf_[fdim].size();
      for (int faceind = sz - 1; faceind >= 0; --faceind) {
        delete nbrf_[fdim][faceind];
      }
    }
    delete[] nbrf_;
  }
  
  struct AnglePatchIndPair {
    Real angle;
    PatchTable::Index patchind;
    int which_copy;
    bool is_forward;
    Brep::Face_Spec region_above;
    Brep::Face_Spec region_below;
    AnglePatchIndPair() { }
    AnglePatchIndPair(const AnglePatchIndPair& o) :
      angle(o.angle), patchind(o.patchind),
      which_copy(o.which_copy), is_forward(o.is_forward),
      region_above(o.region_above), region_below(o.region_below) { }
    AnglePatchIndPair(Real a, PatchTable::Index p, int w, bool i,
                      const Brep::Face_Spec& above, 
                      const Brep::Face_Spec& below) :
      angle(a), patchind(p), which_copy(w), is_forward(i),
      region_above(above), region_below(below) { }
    bool operator<(const AnglePatchIndPair& o) const;
    static Real tol;
  };

  Real AnglePatchIndPair::tol;

  bool AnglePatchIndPair::operator<(const AnglePatchIndPair& o) const {
    Real diff = angle - o.angle;
    if (fabs(diff) < tol) {
      bool rb = region_below.faceind() >= 0;
#ifdef DEBUGGING
      bool ra = region_above.faceind() >= 0;
      bool orb = o.region_below.faceind() >= 0;
      bool ora = o.region_above.faceind() >= 0;
      if ((ra && rb) || (!ra && !rb) || (ora && orb) ||
          (!ora && !orb) || (ra && ora) || (!ra && !ora))
        throw_error("Inconsistent region above/below for same-angle patches");
#endif
      return rb;
    }
    return diff < 0;
  }

 

  struct EdgeLookupKey {
    Brep::Face_Spec fspec1;
    Brep::Face_Spec fspec2;
    Brep::Face_Spec fspec3;
    bool operator<(const EdgeLookupKey& o) const;
    EdgeLookupKey() { }
    EdgeLookupKey(const Brep::Face_Spec& f1, 
                  const Brep::Face_Spec& f2,
                  const Brep::Face_Spec& f3) :
      fspec1(f1), fspec2(f2), fspec3(f3) { }
    EdgeLookupKey(const EdgeLookupKey& o) :
      fspec1(o.fspec1), fspec2(o.fspec2), fspec3(o.fspec3) { }
  };

  bool EdgeLookupKey::operator<(const EdgeLookupKey& o) const {
    if (fspec1 < o.fspec1) return true;
    if (fspec1 > o.fspec1) return false;
    if (fspec2 < o.fspec2) return true;
    if (fspec2 > o.fspec2) return false;
    return fspec3 < o.fspec3;
  }

 

  struct SegHitRecord {
    Real dist;
    int pos;
    bool interior_on_far_side;
    bool interior_on_near_side;
    SegHitRecord(Real d1,
                 int p1,
                 bool far1,
                 bool near1) : 
      dist(d1), pos(p1), interior_on_far_side(far1),
      interior_on_near_side(near1) { }
    SegHitRecord(const SegHitRecord& o) :
      dist(o.dist), pos(o.pos), interior_on_far_side(o.interior_on_far_side),
      interior_on_near_side(o.interior_on_near_side) { }
    SegHitRecord() { }
    bool operator<(const SegHitRecord& o) const;
    static double tol;
  };

  double SegHitRecord::tol;

  bool SegHitRecord::operator<(const SegHitRecord& o) const {
    Real diff = dist - o.dist;
    if (fabs(diff) < tol) {
#ifdef DEBUGGING
      if ((interior_on_near_side && interior_on_far_side) ||
          (!interior_on_near_side && !interior_on_far_side) ||
          (o.interior_on_near_side && o.interior_on_far_side) ||
          (!o.interior_on_near_side && !o.interior_on_far_side) ||
          (interior_on_near_side && o.interior_on_near_side) ||
          (!interior_on_near_side && !o.interior_on_near_side)) {
        throw_error("Inconsistent sides for same-angle patches");
      }
#endif
      return interior_on_near_side;
    }
    return diff < 0.0;
  }

  class OnErrorFlushLog : public QMG::Error_Task {
  private:
    Logstream& log_;
    // Override base case function
    virtual void on_error_do(const string& errmsg) { 
      log_.str() << errmsg << '\n';
      QMG::Error_Message::put_error_messages_into_stream(log_.str());
      log_.str() << std::flush;
    }
  public:
    OnErrorFlushLog(Logstream& logs) : log_(logs) { }
    ~OnErrorFlushLog() { }
  };

  const Real PI_OVER_2 = 1.57079632679490;

  struct FacetMapKey {
    SimpComplex::VertexGlobalIndex verts[MAXDIM];
    static int di;
    void sort();
    bool operator<(const FacetMapKey& other) const;
  };

  int FacetMapKey::di;

  ostream& operator<<(ostream& os, const FacetMapKey& key) {
    for (int i = 0; i < key.di; ++i)
      os << key.verts[i] << ' ';
    return os;
  }



  void FacetMapKey::sort() {
    for (int i = 0; i < di - 1; ++i) {
      for (int j = i + 1; j < di; ++j) {
        if (verts[j] < verts[i]) {
          SimpComplex::VertexGlobalIndex tmp = verts[i];
          verts[i] = verts[j];
          verts[j] = tmp;
        }
      }
    }
  }

  bool FacetMapKey::operator<(const FacetMapKey& o) const {
    for (int i = 0; i < di; ++i) {
      if (verts[i] < o.verts[i]) return true;
      if (verts[i] > o.verts[i]) return false;
    }
    return false;
  }


  struct FacetMapRec1 {
    Brep::FaceIndex owner_faceind;
    int owner_seqno;
    int which_facet : 4;
  };

  /*

  ostream& operator<<(ostream& os, const FacetMapRec1& rec) {
    os << "di:" << rec.owner_faceind << "#" << rec.owner_seqno
      << " fa " << rec.which_facet << " srchd?" << rec.is_searched
      << " isondbled?" << rec.is_on_doubled_ib;
    return os;
  }
  */

  struct FacetMapRec2 {
    Brep::Face_Spec fspec;
    int seqno;
    bool operator<(const FacetMapRec2& o) const;
  };

  bool FacetMapRec2::operator<(const FacetMapRec2& o) const {
    if (fspec < o.fspec) return true;
    if (fspec > o.fspec) return false;
    return seqno < o.seqno;
  }

  ostream& operator<<(ostream& os, const FacetMapRec2& rec) {
    ::operator<<(os, rec.fspec) << '#' << rec.seqno;
    return os;
  }


  struct FspecSimpMaskKey {
    Brep::Face_Spec low_dim_fspec;
    Brep::FaceIndex simpfaceind;
    int simpseqno;
    unsigned int mask;
    bool operator<(const FspecSimpMaskKey& o) const;
  };

  bool FspecSimpMaskKey::operator<(const FspecSimpMaskKey& o) const {
    if (low_dim_fspec < o.low_dim_fspec) return true;
    if (low_dim_fspec > o.low_dim_fspec) return false;
    if (simpfaceind < o.simpfaceind) return true;
    if (simpfaceind > o.simpfaceind) return false;
    if (simpseqno < o.simpseqno) return true;
    if (simpseqno > o.simpseqno) return false;
    return mask < o.mask;
  }

  struct SimpVertKey {
    Brep::Face_Spec fspec;
    int seqno;
    SimpComplex::VertexGlobalIndex vertex;
    bool operator<(const SimpVertKey& o) const;
  };

  bool SimpVertKey::operator<(const SimpVertKey& o) const {
    if (fspec < o.fspec) return true;
    if (fspec > o.fspec) return false;
    if (seqno < o.seqno) return true;
    if (seqno > o.seqno) return false;
    return vertex < o.vertex;
  }

  
  
  struct FspecCopyNum {
    Brep::Face_Spec fspec;
    int copynum;
    bool operator<(const FspecCopyNum& o) const;
  };

  bool FspecCopyNum::operator<(const FspecCopyNum& o) const {
    if (fspec < o.fspec) return true;
    if (fspec > o.fspec) return false;
    return copynum < o.copynum;
  }




  // Routine to determine which side of a surface has a certain
  // simplex.

  int determine_side(const Brep& b,
                     const Brep::Face_Spec& owner,
                     const SimpComplex& sc,
                     Brep::Face_Spec regionspec,
                     int seqno,
                     int which_facet,
                     const map<PatchInBrepAddress,bool>& patch_is_flipped,
                     double scfac,
                     double tol) {
    int di = b.embedded_dim();
    Point startpoint, basepoint, endpoint, tang, coef;
    PatchMath::Workspace workspace;
    vector<PatchMath::IntersectRecord> outputstack;

    {
      SimpComplex::VertexGlobalIndex vnum = 
        sc.node_of_meshface_on_face(regionspec, seqno, which_facet);
      SimpComplex::VertexOrdinalIndex vnumo = sc.global_to_ordinal(vnum);
      for (int j = 0; j < di; ++j)
        startpoint[j] = sc.real_coord_o(vnumo, j);
    }
    int side;
    bool degen = true;
    int trycount = 0;
    while (degen &&  ++trycount < 400) {
      degen = false;

      // Select a point in the base facet of the simplex
      // via interpolation with randomized coeffs.

      Real t = 1.0;
      {
        for (int j = 0; j < di - 1; ++j) {
          coef[j] = 1.0/((double)(2*di)) + 1.0/((double)(2*di)) * random_real();
          t -= coef[j];
        }
      }
      coef[di - 1] = t;
      {
        for (int i = 0; i < di; ++i)
          basepoint[i] = 0.0;
      }
      {
        int k = 0;
        for (int j = 0; j < di + 1; ++j) {
          if (j == which_facet) continue;
          SimpComplex::VertexGlobalIndex vnum = 
            sc.node_of_meshface_on_face(regionspec, seqno, j);
          SimpComplex::VertexOrdinalIndex vnumo = sc.global_to_ordinal(vnum);
          for (int i = 0; i < di; ++i)
            basepoint[i] += coef[k] * sc.real_coord_o(vnumo,i);
          ++k;
        }
      }
      // endpoint extrapolates from startpoint to basepoint by
      // a factor of 5.

      {
        for (int j = 0; j < di; ++j) {
          tang[j] = startpoint[j] - basepoint[j];
          endpoint[j] = startpoint[j] - 5.0 * tang[j];
        }
      }

      // Loop over patches of the face; for each patch find
      // the point where the segment from startpoint to
      // endpoint crosses.  Save the first hit.

      Real first_hit_dist = BIG_REAL;

      for (Brep::Loop_over_patches_of_face ptloop(b, owner);
           ptloop.notdone();
           ++ptloop) {
        outputstack.resize(0, PatchMath::IntersectRecord());
        ptloop.patchmath().intersect_seg(outputstack, workspace,
                                         startpoint, endpoint, scfac, tol, true);
        for (int j = 0; j < outputstack.size(); ++j) {
          if (outputstack[j].degen) {
            degen = true;
            break;
          }
          if (outputstack[j].dist < first_hit_dist) {
            first_hit_dist = outputstack[j].dist;
            Real ip = Point::inner_product(ptloop.patchmath()
                                           .normal(outputstack[j].paramcoord), tang);
            typedef map<PatchInBrepAddress,bool> MPIBB;
            MPIBB::const_iterator it = 
              patch_is_flipped.find(PatchInBrepAddress(owner, ptloop.index()));
            if (it == patch_is_flipped.end()) 
              throw_error("record missing from patch_is_flipped");
            if (it -> second)
              ip = -ip;
            if (ip < 0)
              side = 0;
            else
              side = 1;
          }
        }
        if (degen) break;
      }
      if (first_hit_dist > 10)
        degen = true;
    }
    if (degen)
      throw_error("Unrecoverable degeneracy in 'determine_side'");
    return side;
  }

  
}


// This routine doubles the internal boundary faces of a brep.
// The faces to double are given by the second argument.

QMG::MG::Brep_Under_Construction
QMG::MG::double_brep(const Brep& b,
                     const vector<Brep::Face_Spec>& faces_to_double,
                     double tol,
                     // other return variables:
                     Brep::Face_Labeling<Brep::FaceIndex>& oldfspec_to_new_map,
                     Brep::Face_Labeling<int>& face_num_copies,
                     map<PatchInBrepAddress, bool>& patch_is_flipped) {
  double scfac = b.scale_factor();
  AnglePatchIndPair::tol = tol;
  SegHitRecord::tol = tol;
  using std::ofstream;
  using std::endl;
  ostream logfile(0);
  Logstream logstr(logfile, 1);
  OnErrorFlushLog _unnamed_(logstr);
  
  // initialize the random number generator.
  
  random_real(1);
  int di = b.embedded_dim();
  small_initialize_static_data(di);
  Point::InitStaticData(di);
#ifdef DEBUGGING
  Local_to_PatchMath_CPP::dump_everything = false;
  Local_to_PatchMath_CPP::debug_str = &logstr.str();
#endif
  
  logstr.str() << "tol = " << tol << " scfac = " << scfac << '\n';
  PatchTable patchtable(b, logstr, scfac, tol);
  
  // Make a mapping from original brep's patchind index
  // to patchtable's patch index.

  map<PatchInBrepAddress, PatchTable::Index> lookup_patchind; 
  {
    for (PatchTable::Index patchind = 0; 
         patchind < patchtable.numpatch(); 
         ++patchind) {
      Brep::Face_Spec owner = patchtable.owner(patchind);
      if (patchtable.gdim(patchind) != owner.fdim()) continue;
      Brep::PatchIndex patchind1 = patchtable.orig_index(patchind);
      lookup_patchind[PatchInBrepAddress(owner,patchind1)] = patchind;
    }
  }
  
  // Mark faces that must be doubled.

  Brep::Face_Labeling<My_bool> must_be_doubled(b, My_bool(false));
  Brep::Face_Labeling<Brep::Face_Spec> 
    parent_of_doubled_face(b, Brep::Face_Spec());
  Brep::Ancestor_Lookup anc_lookup(b);

  // used only in 3D

  vector<bool> is_ib_ancestor_of_vtx(b.level_size(di - 1));


  {
    int ftds = faces_to_double.size();
    for (int j = 0; j < ftds; ++j) {
      Brep::Face_Spec fspec = faces_to_double[j];
      if (fspec.fdim() != di - 1)
        throw_error("Only faces of dimension di-1 may be doubled");
      must_be_doubled(fspec) = true;
      Brep::Face_Spec parent_fspec(-1,-1);
      for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop(anc_lookup, fspec);
           ancloop.notdone();
           ++ancloop) {
        Brep::Face_Spec afspec = ancloop.ancestor_fspec();
        if (afspec.fdim() != fspec.fdim() + 1) continue;
        if (ancloop.occurrence_count() % 2 != 0)
        throw_error("Specified face is not an internal boundary of parent topological entity");
        if (parent_fspec.fdim() >= 0)
          throw_error("Specified face occurs as boundary of more than one parent topological entity");
        parent_fspec = afspec;
      }
      if (parent_fspec.fdim() < 0)
        throw_error("Specified face is not a boundary of any parent topological entity");
      parent_of_doubled_face(fspec) = parent_fspec;
    }
  }

  int patchtable_size = patchtable.numpatch();
  vector<bool> patch_is_doubled(patchtable_size);

  // Mark doubled patches.
  {
    for (PatchTable::Index patchind = 0; patchind < patchtable_size; ++patchind) {
      patch_is_doubled[patchind] =  (patchtable.gdim(patchind) == di - 1) &&
        must_be_doubled(patchtable.owner(patchind));
    }
  }

  vector<AnglePatchIndPair> bigsortlist;
  vector<int> pos_in_bigsortlist(patchtable_size);
  vector<int> len_in_bigsortlist(patchtable_size, 0);
          
  // Make a table that holds the patches around
  // each bezier curve (3D) or point (2D) in order of angle.

  {
    vector<AnglePatchIndPair> sortlist;
    for (PatchTable::Index patchind = 0; patchind < patchtable_size; ++patchind) {
      if (patchtable.gdim(patchind) != di - 2) continue;

      // For a patch of dimension di-2, find all its parents
      // and properly list them in the big sort list  data structure.
      
      
      Point hhtransform;
      Point paramcoord;
      Real beta;
      
      // Get the angle that the parent patch makes with respect
      // to this patch.

      if (di == 3) {
        Point direc;
        direc[0] = 1.0;
        paramcoord[0] = 0.5;
        Point tangent = 
          patchtable.patchmath(patchind).direc_deriv(paramcoord, direc);
        beta = Matrix::compute_hh_transform(&hhtransform[0], &tangent[0], 1, 3);
      }

      sortlist.resize(0, AnglePatchIndPair());
      for (PatchTable::Loop_over_parents parloop(patchtable, patchind);
           parloop.notdone();
           ++parloop) {
        PatchTable::Index ppatchind = parloop.parent_index();
        Point parent_tan = 
          patchtable.get_outward_tan_in_parent(paramcoord,
                                               patchind, 
                                               parloop);
        Point liftcoord = patchtable.lift_parameters(paramcoord,
                                                     patchind,
                                                     parloop.listind());
        Point parent_normal = patchtable.patchmath(ppatchind).
          normal(liftcoord);
        if (di == 3) {
          PatchTable::apply_3d_hh_and_shift(beta, hhtransform, parent_tan);
          PatchTable::apply_3d_hh_and_shift(beta, hhtransform, parent_normal);
        }
        Real tanang = atan2(parent_tan[1], parent_tan[0]);
        Real nrmang = atan2(parent_normal[1], parent_normal[0]);
        if (nrmang < tanang)
          nrmang += 2 * PI;
        bool is_forward =  fabs(nrmang - tanang - PI_OVER_2) < PI_OVER_2;

        Brep::Face_Spec front = patchtable.front_side(ppatchind);
        Brep::Face_Spec back = patchtable.back_side(ppatchind);

        if (!patch_is_doubled[ppatchind]) {
          
          sortlist.push_back(AnglePatchIndPair(tanang,
                                               ppatchind, 
                                               0, 
                                               is_forward,
                                               is_forward? front: back,
                                               is_forward? back : front));
        }
        else {
          
          // if this is a doubled patch, put two copies
          // of the patch in the sort list.  Assign epsilon so that
          // the copy with the clockwise normal comes second.
          
          Brep::Face_Spec nullspec(di,-1);
          sortlist.push_back(AnglePatchIndPair(tanang, 
                                               ppatchind, 
                                               0,
                                               is_forward,
                                               is_forward? front : nullspec,
                                               is_forward? nullspec : front));

          sortlist.push_back(AnglePatchIndPair(tanang, 
                                               ppatchind, 
                                               1,
                                               is_forward,
                                               is_forward? nullspec : back,
                                               is_forward? back : nullspec));
        }
      }

      // Sort the parent patches in order of angle.
      
      sort(sortlist.begin(), sortlist.end());
      int sortlistlen = sortlist.size();

      // Find a starting point for the sweep.
      // A starting point has the brep on its clockwise side.

      int sweepstart = -1;
      int both_covered_count = 0;
      {
        Brep::Face_Spec prev_region_above(-1,-1), first_region_below(-1,-1);
        for (int ii = 0; ii < sortlistlen; ++ii) {
          PatchTable::Index ppatchind = sortlist[ii].patchind;
          int wc = sortlist[ii].which_copy;
          bool isfor = sortlist[ii].is_forward;
          Brep::Face_Spec region_above = sortlist[ii].region_above;
          Brep::Face_Spec region_below = sortlist[ii].region_below;
          if (ii > 0 && region_below != prev_region_above)
            throw_error("Front/back inconsistency");
          if (region_below.faceind() < 0 && region_above.faceind() >= 0)
            sweepstart = ii;
          if (region_below.faceind() >= 0 && region_above.faceind() >= 0)
            ++both_covered_count;
          prev_region_above = region_above;
          if (ii == 0)
            first_region_below = region_below;
        }
        if (first_region_below != prev_region_above)
          throw_error("Front back inconsistency 2");
      }
      if (sweepstart < 0) {
        if (both_covered_count == sortlistlen) {
          sweepstart = 0;
        }
        else {
          throw_error("ridge does not touch brep region");
        }
      }
      int base = bigsortlist.size();
      bigsortlist.resize(base + sortlistlen, AnglePatchIndPair());
      pos_in_bigsortlist[patchind] = base;
      len_in_bigsortlist[patchind] = sortlistlen;
      {
        for (int ii = 0; ii < sortlistlen; ++ii) {
          int srcii = (ii + sweepstart) % sortlistlen;
          bigsortlist[base + ii] = sortlist[srcii];
        }
      }
    }
  }

  Brep_Under_Construction newb(di, di);
  NewBrepType newb_graph(di);

  {
    // Copy property-values from b to newb.
    Brep_Under_Construction::Propval_inserter pvi(newb);
    for (Brep::Loop_over_propvals_of_brep pvloop(b);
         pvloop.notdone();
         ++pvloop) {
      if (compare_nocase(pvloop.prop(), Brep::global_id_propname()) == 0) {
        string newbrep_id;
        ostringstream os;
        os << "(" << pvloop.val() << " with doubled faces ";
        for (int j = 0; j < faces_to_double.size(); ++j)
          os << faces_to_double[j] << " ";
        os << ")";  
        pvi.insert_propval(Brep::global_id_propname(), os.str());
      }
      else {
        pvi.insert_propval(pvloop.prop(), pvloop.val());
      }
    }
  }
  
  {
    // Copy control points from b to newb.
    Brep_Under_Construction::Control_point_inserter cpi(newb);
    vector<Real> control_point(di);
    for (int i = 0; i < b.num_control_points(); ++i) {
      for (int j = 0; j < di; ++j)
        control_point[j] = b.control_point(i,j);
      cpi.insert_control_point(control_point);
    }
  }

      
  // Main loop for brep: Copy faces, doubling as needed.
  
  {

    map<EdgeLookupKey, int> edgelookup;
    vector<pair<PatchTable::Index, int> > grandpar_list;
    vector<int> edge_global_copy_num;
    PatchTableIndexMap<int> copy0_pos_in_small_list(patchtable,-1);
    UnionFind grandpar_unionfind;

    for (int fdim = di; fdim >= 0; --fdim) {
      Brep_Under_Construction::Top_face_inserter tfi(newb, fdim);
      for (Brep::Face_Spec_Loop_Over_Faces_Of_Dim fspec(b, fdim);
           fspec.notdone();
           ++fspec) {


             logstr.str() << "\n-----------------\nprocessing face " << fspec << '\n';

        Brep::Face_Spec next_new_fspec(fdim, newb.level_size(fdim));
      
        // How many times to add the new face?
        // In the big "if" branch that follows, we consider all
        // cases on how many copies to add.  Each branch computes
        // a value for num_copies.  Each branch also
        // inserts the copies as children into their parent faces
        // in the newb_graph data structure.
        
        int num_copies;
        if (fdim == di) {
          num_copies = 1;
        }
        else if (fdim == di - 1) {
          
          // Determine if the face is marked for doubling.
          
          if (must_be_doubled(fspec)) {
            num_copies = 2;
            
            // Add the face twice to the parent.
            Brep::Face_Spec parent_fspec = parent_of_doubled_face(fspec);
            vector<Brep::FaceIndex>& parent_childlist = 
              newb_graph.get_face(parent_fspec).childlist;
            vector<int>& ori = 
              newb_graph.get_face(parent_fspec).child_ori;
            for (int j = 0; j < num_copies; ++j) {
              parent_childlist.push_back(next_new_fspec.faceind() + j);
              ori.push_back(j);
            }
          }
          else {
            num_copies = 1;
            // Add the face to all parents.
            for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop(anc_lookup, fspec);
                 ancloop.notdone();
                 ++ancloop) {
              Brep::Face_Spec afspec = ancloop.ancestor_fspec();
              if (afspec.fdim() != fdim + 1) continue;
              vector<Brep::FaceIndex>& parent_childlist = 
                newb_graph.get_face(afspec).childlist;
              vector<Brep::FaceIndex>& parent_ori = 
                newb_graph.get_face(afspec).child_ori;
              int a = ancloop.occurrence_count();
              while (a > 2)
                a -= 2;
              int ori[2];
              // Select a patch on the face.
              PatchTable::Index patchind = 
                lookup_patchind[PatchInBrepAddress(fspec, 0)];
              if (a == 1) {
                if (patchtable.front_side(patchind) == afspec)
                  ori[0] = 0;
                else if (patchtable.back_side(patchind) == afspec)
                  ori[0] = 1;
                else 
                  throw_error("Ancestor face spec does not correspond to front/back side in patchtable");
              }
              else {
                if (patchtable.front_side(patchind) != afspec ||
                  patchtable.back_side(patchind) != afspec)
                  throw_error("Ancestor face spec not on both sides of patch");
                ori[0] = 0;
                ori[1] = 1;
              }
              for (int j = 0; j < a; ++j) {
                parent_childlist.push_back(next_new_fspec.faceind());
                parent_ori.push_back(ori[j]);
              }
            }
          }
        }
        else if (fdim == di - 2) {

          // Get a sample patch from the topological face.
          
          PatchTable::Index patchind = 
            lookup_patchind[PatchInBrepAddress(fspec, 0)];
          if (patchtable.gdim(patchind) != di - 2)
            throw_error("Some error setting up lookup_patchind");
          
          // Loop over the ordered parents of this patch.
          // Start a new copy every time a doubled ib is crossed,
          // or empty space.

          num_copies = 0;
          typedef vector<AnglePatchIndPair> VecAngPP;
          VecAngPP::iterator sortlist = 
            &bigsortlist[pos_in_bigsortlist[patchind]];
          int len1 = len_in_bigsortlist[patchind];
          Brep::Face_Spec this_copy_start(-1,-1);
          for (int ii = 0; ii < len1; ++ii) {
            PatchTable::Index ppatchind = sortlist[ii].patchind;

            Brep::Face_Spec old_parent_fspec = patchtable.owner(ppatchind);
            int wc = sortlist[ii].which_copy;
            Brep::Face_Spec new_parent_fspec = 
              Brep::Face_Spec(fdim + 1, 
                              oldfspec_to_new_map(old_parent_fspec) + wc);

            if (ii == 0 || 
                (sortlist[ii].region_below.faceind() < 0 && 
                sortlist[ii].region_above.faceind() >= 0)) {
              this_copy_start = new_parent_fspec;
              ++num_copies;
            }
            Brep::Face_Spec new_fspec1 =
              Brep::Face_Spec(fdim, next_new_fspec.faceind() + num_copies - 1);

            if (ii == len1 - 1 ||
              (sortlist[ii].region_below.faceind() >= 0 &&
              sortlist[ii].region_above.faceind() < 0)) {
              Brep::Face_Spec this_copy_end = new_parent_fspec;
              edgelookup[EdgeLookupKey(fspec, this_copy_start, this_copy_end)] = num_copies - 1;
              edgelookup[EdgeLookupKey(fspec, this_copy_end, this_copy_start)] = num_copies - 1;
              logstr.str() << " inserted key(" << fspec << "," << this_copy_start << ","
                << this_copy_end << ") into map with data " << num_copies - 1 << '\n';
            }

            // Insert this copy as a child face of the current parent.
            newb_graph.get_face(new_parent_fspec).childlist.
              push_back(new_fspec1.faceind());
            newb_graph.get_face(new_parent_fspec).child_ori.
              push_back(2);
          }

          if (num_copies == 0) {
            num_copies = 1;
            int rcount = 0;
            
            // Add the face to all parents (should be one).

            for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop(anc_lookup, fspec);
                 ancloop.notdone();
                 ++ancloop) {
              Brep::Face_Spec afspec = ancloop.ancestor_fspec();
              if (afspec.fdim() == fdim + 1)
                throw_error("hanging ridge contained by surface");
              if (afspec.fdim() == di) {
                ++rcount;
                newb_graph.get_face(afspec).iblist.push_back(next_new_fspec);
              }
            }

            if (rcount != 1)
              throw_error("Nonbounding ridge must be i.b. of exactly one region");
          }
        }
        else { // get here if fdim == 0 and di == 3.
          // This is the most complicated case.

          // Get a patch (control point) for this vertex.
          
          PatchTable::Index patchind = 
            lookup_patchind[PatchInBrepAddress(fspec, 0)];
          if (patchtable.gdim(patchind) != 0)
            throw_error("Some error setting up lookup_patchind");

          logstr.str() << " patchind = " << patchind << '\n';


          // Mark all grandparents initially as ib grandparents.
          {
            for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop(anc_lookup, fspec);
                 ancloop.notdone();
                 ++ancloop) {
              Brep::Face_Spec afspec = ancloop.ancestor_fspec();
              if (afspec.fdim() == 2)
                is_ib_ancestor_of_vtx[afspec.faceind()] = true;
            }
          }


          // Put all the parents and grandparents of this patch in a table.

          grandpar_list.resize(0, pair<PatchTable::Index,int>());
          copy0_pos_in_small_list.reset(-1);

          {
            for (PatchTable::Loop_over_parents parloop(patchtable, patchind);
                 parloop.notdone();
                 ++parloop) {

              PatchTable::Index ppatchind = parloop.parent_index();
              Brep::Face_Spec eowner = patchtable.owner(ppatchind);
              if (copy0_pos_in_small_list[ppatchind] < 0) {
                copy0_pos_in_small_list[ppatchind] = grandpar_list.size();
                for (int ii = 0; ii < face_num_copies(eowner); ++ii) {
                  grandpar_list.push_back(pair<PatchTable::Index,int>
                                          (ppatchind, ii));
                }
                logstr.str() << "grandparlist edge entry owner = " << eowner << "  patchind = " 
                  << ppatchind << " #copies = " << face_num_copies(eowner) << '\n'
                  << " at position " << copy0_pos_in_small_list[ppatchind] << '\n';
              }
              bool parent_is_on_edge = eowner.fdim() == 1;

              for (PatchTable::Loop_over_parents grandparloop(patchtable, ppatchind);
                   grandparloop.notdone();
                   ++grandparloop) {
                PatchTable::Index gppatchind = grandparloop.parent_index();
                Brep::Face_Spec fowner = patchtable.owner(gppatchind);
                if (parent_is_on_edge)
                  is_ib_ancestor_of_vtx[fowner.faceind()] = false;
                if (copy0_pos_in_small_list[gppatchind] < 0) {
                  copy0_pos_in_small_list[gppatchind] = grandpar_list.size();
                  for (int ii = 0; ii < face_num_copies(fowner); ++ii) {
                    grandpar_list.push_back(pair<PatchTable::Index,int>
                                            (gppatchind, ii));
                  }
                  logstr.str() << "grandparlist patch entry owner = " << fowner 
                      << " patchind = " << gppatchind << " numcopies = " << face_num_copies(fowner)
                      << " is_ib_anc = " << is_ib_ancestor_of_vtx[fowner.faceind()] << "\n"
                      << " at position " << copy0_pos_in_small_list[gppatchind] << '\n';
                }
              }
            }
          }

          // Conduct union-find, first pass. 
          // Merge parent & grandparent neighbors.

          int grandpar_size = grandpar_list.size();
          
          edge_global_copy_num.resize(grandpar_size);
          {
            for (int jj = 0; jj < grandpar_size; ++jj)
              edge_global_copy_num[jj] = -1;
          }
          grandpar_unionfind.initialize(grandpar_size);
          {
            for (PatchTable::Loop_over_parents parloop(patchtable, patchind);
                 parloop.notdone();
                 ++parloop) {
              PatchTable::Index ppatchind = parloop.parent_index();
              Brep::Face_Spec old_eowner = patchtable.owner(ppatchind);
              typedef vector<AnglePatchIndPair> VAPIP;
              VAPIP::iterator sortlist = 
                &bigsortlist[pos_in_bigsortlist[ppatchind]];
              int len1 = len_in_bigsortlist[ppatchind];

              int edge_copy_num = -1;
              Brep::Face_Spec this_copy_start;
              for (int ii = 0; ii < len1; ++ii) {
                PatchTable::Index gppatchind = sortlist[ii].patchind;


                Brep::Face_Spec old_gparent_fspec = patchtable.owner(gppatchind);
                int gwc = sortlist[ii].which_copy;
                Brep::Face_Spec new_gparent_fspec = Brep::Face_Spec(fdim + 2,
                  oldfspec_to_new_map(old_gparent_fspec) + gwc);

                if (ii == 0 || 
                    (sortlist[ii].region_below.faceind() < 0 && 
                    sortlist[ii].region_above.faceind() >= 0)) {
                  ++edge_copy_num;
                  this_copy_start = new_gparent_fspec;
                }
                  
                if (old_eowner.fdim() == 1 &&
                  (ii == len1 - 1 ||
                  (sortlist[ii].region_below.faceind() >= 0 &&
                  sortlist[ii].region_above.faceind() < 0))) {
                  Brep::Face_Spec this_copy_end = new_gparent_fspec;
                  EdgeLookupKey key(old_eowner, this_copy_start, this_copy_end);
                  if (edgelookup.find(key) == edgelookup.end()) 
                    throw_error("Missing key in edgelookup table");
                  int edgepos1 = copy0_pos_in_small_list[ppatchind] + edge_copy_num;
                  edge_global_copy_num[edgepos1] = edgelookup[key];
                  logstr.str() << " labeling position " << edgepos1 << " corresponding to patchind "
                    << patchind << " copy " << edge_copy_num << " oldfspec " << old_eowner
                    << " with label " << edgelookup[key] << '\n';
                }

                if (sortlist[ii].region_above.faceind() < 0) continue;
                int side = sortlist[ii].which_copy;
                int gppos = copy0_pos_in_small_list[gppatchind] + side;
                int nextii = (ii + 1) % len1;
                PatchTable::Index nextgppatchind = sortlist[nextii].patchind;
                int nextside = sortlist[nextii].which_copy;
                int nextgppos =  copy0_pos_in_small_list[nextgppatchind] + nextside;
                int edgepos = copy0_pos_in_small_list[ppatchind] + edge_copy_num;
                grandpar_unionfind.merge(edgepos, gppos);
                grandpar_unionfind.merge(edgepos, nextgppos);
                logstr.str() << " merging gplist entries at positions " << edgepos << " "
                  << gppos << " " << nextgppos << " corresponding to "
                  << " edge patchind = " << ppatchind << " copy " << edge_copy_num
                  << " and patches " << gppatchind << " side " << side
                  << " and " << nextgppatchind << " side " << nextside << '\n';

              }
            }
          }
          // Next pass: shoot rays to merge labels.

          {
            Real loose_tol = tol * 100.0;
            vector<SegHitRecord> hitlist;
            bool degen = true;
            int trycount = 0;
            vector<bool> ray_shot_from_label(grandpar_size, false);
            while (degen && trycount++ < 400) {
              degen = false;
              Point rootpoint;
              for (int j1 = 0; j1 < di; ++j1) {
                rootpoint[j1] = 2 * random_real() - 1.0;
              }
              int root_labelpos = -1;
              for (int pos0 = 0; pos0 < grandpar_size; ++pos0) {
                ray_shot_from_label[pos0] = false;
              }
              for (int pos1 = 0; pos1 < grandpar_size; ++pos1) {
                int this_lab = grandpar_unionfind.get_label(pos1);
                if (ray_shot_from_label[this_lab]) continue;
                PatchTable::Index anc_patchind = grandpar_list[pos1].first;
                if (patchtable.gdim(anc_patchind) == 1 &&
                    len_in_bigsortlist[anc_patchind] > 0) continue;

                
                // Set up a ray shoot.
                int which_copy = grandpar_list[pos1].second;
                Point target = 
                  patchtable.vtxnbd_select_patchcone_randpoint(patchind, anc_patchind);
                Point segtan = Point::subtract(rootpoint, target);

                
                logstr.str() << "shooting ray from pos " << pos1 << " whose label is"
                  << this_lab << " patchind = " << anc_patchind << " copy " << which_copy
                  << "\n startpoint = " << target << " from root" << rootpoint << '\n';
                // See what is hit by the segment
                // from rootpoint to target.
                hitlist.resize(0, SegHitRecord());
                for (int pos2 = 0; pos2 < grandpar_size; ++pos2) {
                  PatchTable::Index anc_patchind2 = grandpar_list[pos2].first;
                  if (patchtable.gdim(anc_patchind2) != 2) continue;
                  int which_copy2 = grandpar_list[pos2].second;
                  pair<PatchTable::ConeIntersectCode, Real> rval = 
                    patchtable.vtxnbd_patchcone_meets_seg(patchind, anc_patchind2,
                                                          rootpoint, target,
                                                          segtan, loose_tol);

                  logstr.str() << "    for pos " << pos2 << " patchind = " << anc_patchind2 
                    << " copy = " << which_copy2 << " result = " << rval.first <<  "," 
                    << rval.second << '\n';

                  if (rval.first == PatchTable::DEGEN || 
                      rval.first == PatchTable::HITS_EDGE) {
                    degen = true;
                    break;
                  }
                  if (rval.first == PatchTable::MISSES_PATCH)  continue;
                  if (rval.second > 1.0 + loose_tol) continue;
                  bool interior_in_front = patch_is_doubled[anc_patchind2]?
                    (which_copy2 == 0) : 
                    (patchtable.front_side(anc_patchind2).faceind() >= 0);
                  bool interior_in_back = patch_is_doubled[anc_patchind2]?
                    (which_copy2 == 1) : 
                    (patchtable.back_side(anc_patchind2).faceind() >= 0);
                  bool interior_on_far_side = 
                    (rval.first == PatchTable::HITS_PATCH_BACK)?
                    interior_in_front : interior_in_back;
                  bool interior_on_near_side = 
                    (rval.first == PatchTable::HITS_PATCH_BACK)?
                    interior_in_back : interior_in_front;
                  logstr.str() << " interior_in_front = " << interior_in_front << " in back = "
                    << interior_in_back << " far side " << interior_on_far_side << " near side = "
                    << interior_on_near_side << '\n';
                  hitlist.push_back(SegHitRecord(rval.second, pos2,
                                                 interior_on_far_side,
                                                 interior_on_near_side));
                }
                if (degen)
                  break;
                if (patchtable.gdim(anc_patchind) == 1)
                  hitlist.push_back(SegHitRecord(1.0, pos1, true, true));
                if (hitlist.size() == 0)
                  throw_error("Unexpected hitlist size = 0");
                sort(hitlist.begin(), hitlist.end());

                // Merge labels along segment.

                bool prev;
                for (int hitind = 0; hitind < hitlist.size(); ++hitind) {
                  int pos2 = hitlist[hitind].pos;
                  int pos2lab = grandpar_unionfind.get_label(pos2);
                  ray_shot_from_label[pos2lab] = true;
                  logstr.str() << "  for this rayshoot processing hit on pos " << pos2 
                    << " label " << pos2lab << '\n';
                  if (hitlist[hitind].interior_on_near_side) {
                    if (hitind == 0) {
                      if (root_labelpos == -2) {
                        throw_error("Inconsistent view from root point 1");
                      }
                      if (root_labelpos == -1) {
                        logstr.str() << "   (made the root)\n";
                        root_labelpos = pos2;
                      } 
                      else {
                        logstr.str() << "   (merge with root) \n";
                        grandpar_unionfind.merge(root_labelpos, pos2);
                      }
                    }
                    else { // hitind > 0
                      if (!prev) {
                        throw_error("Inconsistent front-back 1");
                      }
                      logstr.str() << " merge with pos " << hitlist[hitind-1].pos << '\n';
                      grandpar_unionfind.merge(hitlist[hitind - 1].pos, pos2);
                    }
                  }
                  else { // nothing in back
                    if (hitind == 0) {
                      if (root_labelpos >= 0)
                        throw_error("Inconsistent view from root point 2");
                      root_labelpos = -2;
                    }
                    else { // hitind > 0
                      if (prev) {
                        throw_error("Inconsistent front-back 2");
                      }
                    }
                  }
                  prev = hitlist[hitind].interior_on_far_side;
                }
              }
            }
            if (degen)
              throw_error("Unrecoverable degeneracy in shoot ray");
            for (int jj = 0; jj < grandpar_size; ++jj)
              if (!ray_shot_from_label[grandpar_unionfind.get_label(jj)])
                throw_error("Label exists with no ray shot");
          }

          num_copies = grandpar_unionfind.pack_labels();
          logstr.str() << " num_copies = " << num_copies << '\n';
          
          // Fourth pass: loop over edges, and insert
          // new endpoints into newb_graph data structure.

          {
            for (int pos = 0; pos < grandpar_size; ++pos) {
              PatchTable::Index anc_patchind = grandpar_list[pos].first;
              int gdim = patchtable.gdim(anc_patchind);
              Brep::Face_Spec owner = patchtable.owner(anc_patchind);
              if (owner.fdim() != gdim) continue;
              if (owner.fdim() == 2 && 
                  !is_ib_ancestor_of_vtx[owner.faceind()]) continue;
              int which_copy = grandpar_list[pos].second;
              int which_copy_global = 
                (owner.fdim() == 2)? which_copy : edge_global_copy_num[pos];
              if (which_copy_global < 0)
                which_copy_global = 0;
              int which_vtx_copy = grandpar_unionfind.get_label(pos);
              Brep::Face_Spec newowner(gdim, 
                                       oldfspec_to_new_map(owner) + which_copy_global);
              Brep::Face_Spec new_vtx_fspec(0, next_new_fspec.faceind() + which_vtx_copy);
              NewBrepFaceInfo& ancface = 
                newb_graph.get_face(newowner);
              if (gdim == 2) {
                if (ancface.iblist.size() == 0 ||
                    ancface.iblist.back() != new_vtx_fspec) {
                  ancface.iblist.push_back(new_vtx_fspec);
                }
              }
              else {
                ancface.childlist.push_back(new_vtx_fspec.faceind());
                ancface.child_ori.push_back(2);
              }
            }
          }
                

          if (num_copies == 0) {
            // Get here if the vertex does not lie any edges or surfaces,
            // and instead is an internal boundary.
            num_copies = 1;
            int rcount = 0;
            for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop(anc_lookup,fspec);
                 ancloop.notdone();
                 ++ancloop) {
              Brep::Face_Spec afspec = ancloop.ancestor_fspec();
              if (afspec.fdim() == 0) continue;
              if (afspec.fdim() <= 2) 
                throw_error("Num copies = 0 but vertex has edge/face ancestor?");
              newb_graph.get_face(afspec).iblist.push_back(next_new_fspec);
              ++rcount;
            }
            if (rcount != 1)
              throw_error("Vertex with no neighbors must be ib of exactly one region");
          }
        }

        // Now add the faces.
        
        oldfspec_to_new_map(fspec) = next_new_fspec.faceind();
        face_num_copies(fspec) = num_copies;

        logstr.str() << " face " << fspec << " mapped to " << next_new_fspec << " ... "
          << Brep::Face_Spec(next_new_fspec.fdim(), next_new_fspec.faceind() + num_copies - 1)
          << '\n';


        for (int which_copy = 0; which_copy < num_copies; ++which_copy) {
          {
            ostringstream newname;
            newname << b.face_name(fspec);
            if (num_copies > 1) {
              newname << "_$copy" << which_copy;
            }
            tfi.insert_top_face(newname.str());
          }
          newb_graph.add_face(fdim);


          // Insert topological entity propvals.
          Brep::Face_Spec 
            this_copy_fspec(fdim, 
                            next_new_fspec.faceind() + which_copy);

          Brep_Under_Construction::Top_face_propval_inserter 
            tfpvi(newb, this_copy_fspec);
          for (Brep::Loop_over_propvals_of_face pvloop2(b, fspec);
               pvloop2.notdone();
               ++pvloop2) {
            tfpvi.insert_propval(pvloop2.prop(), pvloop2.val());
          }

          // Skip children and ib's for now.

          // Insert patches.
          
          Brep_Under_Construction::Top_face_patch_inserter
            tfpi(newb, this_copy_fspec);
          vector<Brep::ControlPointIndex> controlpts;

          for (Brep::Loop_over_patches_of_face ptloop(b, fspec);
               ptloop.notdone();
               ++ptloop) {
            PatchType ptype = ptloop.ptype();
            int deg1 = ptloop.degree1();
            int ncp, deg2;
            
            PatchTable::Index patchind = 
              lookup_patchind[PatchInBrepAddress(fspec, ptloop.index())];
            if (patchtable.gdim(patchind) != fdim)
              throw_error("Some error setting up lookup_patchind");
            bool is_forward = patchtable.is_forward(patchind);
            patch_is_flipped[PatchInBrepAddress(fspec, ptloop.index())] = 
              !is_forward;

            if (fdim == 0) {
              controlpts.resize(1);
              controlpts[0] = ptloop.control_point(0);
            }
            else if (fdim == 1) {
              ncp = deg1  + 1;
              controlpts.resize(ncp);
              if (is_forward) {
                for (int j = 0; j < ncp; ++j)
                  controlpts[j] = ptloop.control_point(j);
              }
              else {
                for (int j = 0; j < ncp; ++j)
                  controlpts[j] = ptloop.control_point(deg1 - j);
              }
            }
            else  { // if (gdim == 2) 
              if (ptype == BEZIER_TRIANGLE) {
                ncp = (deg1 + 1) * (deg1 + 2) / 2;
                controlpts.resize(ncp);
                if (is_forward) {
                  for (int j = 0; j < ncp; ++j)
                    controlpts[j] = ptloop.control_point(j);
                }
                else {
                  for (int rownum = 0; rownum <= deg1; ++rownum) {
                    int rowbase = rownum * (rownum + 1) / 2;
                    for (int pos = 0; pos <= rownum; ++pos) {
                      int j = pos + rowbase;
                      int oldpos = rownum - pos + rowbase;
                      controlpts[j] = ptloop.control_point(oldpos);
                    }
                  }
                }
              } 
              else { // ptype == BEZIER_QUAD
                deg2 = ptloop.degree2();
                ncp = (deg1 + 1) * (deg2 + 1);
                controlpts.resize(ncp);
                if (is_forward) {
                  for (int j = 0; j < ncp; ++j)
                    controlpts[j] = ptloop.control_point(j);
                }
                else {
                  for (int rownum = 0; rownum <= deg2; ++rownum) {
                    int rowbase = rownum * (deg1 + 1);
                    for (int pos = 0; pos <= deg1; ++pos) {
                      int j = pos + rowbase;
                      int oldpos = deg1 - pos + rowbase;
                      controlpts[j] = ptloop.control_point(oldpos);
                    }
                  }
                }
              }
            }
            tfpi.insert_patch(deg1, deg2, ptype, controlpts);
          }
        }
      }
    }
  }

  // Go back to add the children and ib's.

  {
    for (int fdim = 0; fdim <= di; ++fdim) {
      for (Brep::Face_Spec_Loop_Over_Faces_Of_Dim fspec(newb, fdim);
           fspec.notdone();
           ++fspec) {
        Brep_Under_Construction::Top_face_bound_inserter tfbi(newb, fspec);
        vector<Brep::FaceIndex>& chlist = newb_graph.get_face(fspec).childlist;
        vector<int>& ori = newb_graph.get_face(fspec).child_ori;
        int sz = chlist.size();
        if (ori.size() != sz)
          throw_error("Child list and ori different sizes");
        {
          for (int j = 0; j < sz; ++j)
            tfbi.insert_boundary(Brep::Face_Spec(fdim - 1, chlist[j]), ori[j]);
        }
        Brep_Under_Construction::Top_face_ib_inserter tfii(newb, fspec);
        vector<Brep::Face_Spec>& iblist = newb_graph.get_face(fspec).iblist;
        int sz2 = iblist.size();
        {
          for (int j = 0; j < sz2; ++j)
            tfii.insert_ib(iblist[j]);
        }
      }
    }
  }
  return newb;
}

// This routine double a simplicial complex, assuming the
// brep has already been doubled by the previous routine.
// Other arguments were computed by previous routine.

QMG::SimpComplex_Under_Construction
QMG::MG::double_simpcomp(const Brep& b,
                         const Brep& newb,
                         const SimpComplex& sc,
                         const Brep::Face_Labeling<Brep::FaceIndex>& oldfspec_to_new_map,
                         const Brep::Face_Labeling<int>& num_copies,
                         const map<PatchInBrepAddress,bool>& patch_is_flipped,
                         double tol
                         ) {

  int di = b.embedded_dim();
  double scfac = b.scale_factor();
  if (sc.embedded_dim() != di)
    throw_error("Simplicial complex dimension does not match brep dimension");
  if (sc.gdim() != di)
    throw_error("Only full dimensional simplicial complex may be doubled");

  string brep_id1 = b.lookup_brep_propval(Brep::global_id_propname());
  string brep_id2 = sc.lookup_prop_val(Brep::global_id_propname());
  if (compare_nocase(brep_id1,brep_id2) != 0)
    throw_error("Brep and simpcomplex have different global id's");

  ostream log2(0);
#ifdef DEBUGGING
  Local_to_PatchMath_CPP::dump_everything = false;
  Local_to_PatchMath_CPP::debug_str = &log2;
#endif

  FacetMapKey::di = di;
  random_real(1);
  small_initialize_static_data(di);
  Point::InitStaticData(di);
  int numnodes = sc.num_nodes();

  // find the maximum number of copies.

  int maxnumcopy = 0;
  {
    for (Brep::Face_Spec_Loop_Over_Faces fspec(b); fspec.notdone(); ++fspec) {
      if (num_copies(fspec) > maxnumcopy)  {
        maxnumcopy = num_copies(fspec);
      }
    }
  }

  log2 << "maxnumpcopy = " << maxnumcopy << '\n';



  // Find the highest vertex global number.

  SimpComplex::VertexGlobalIndex next_vnum = 0;
  {
    for (SimpComplex::VertexOrdinalIndex vnumo = 0; vnumo < numnodes; ++vnumo) {
      SimpComplex::VertexGlobalIndex vnum = 
        sc.ordinal_to_global(vnumo);
      if (vnum > next_vnum)
        next_vnum = vnum;
    }
  }
  ++next_vnum;

  log2 << "next_vnum = " << next_vnum << '\n';


  // Make a hash table with vertex lowest dimensional owners.

  typedef map<SimpComplex::VertexGlobalIndex, Brep::Face_Spec> LowestDimOwnerMapType;
  LowestDimOwnerMapType lowest_dim_owner;

  // Initialize that map.
  {
    for (int fdim = 0; fdim < di; ++fdim) {
      int lsize = sc.brep_levelsize(fdim);
      for (Brep::FaceIndex faceind = 0; faceind < lsize; ++faceind) {
        Brep::Face_Spec fspec(fdim, faceind);
        int nnf = sc.num_node_on_face(fspec);
        for (int seqno = 0; seqno < nnf; ++seqno) {
          SimpComplex::VertexGlobalIndex vnum = sc.node_on_face(fspec, seqno);
          // Use insert rather than square brackets -- want to ensure that
          // only the first insertion of a given vertex remains.
          lowest_dim_owner.
            insert(pair<SimpComplex::VertexGlobalIndex, Brep::Face_Spec>(vnum,fspec));
        }
      }
    }
  }
 
  typedef map<FacetMapKey, FacetMapRec2> FacetMapType2;
  FacetMapType2 facetmap2;

  typedef map<FacetMapKey, Brep::Face_Spec> LowestDimOwnerMapType2;
  LowestDimOwnerMapType2 lowest_dim_owner2;

  
  
  // In this loop we hash all simplex faces lying on brep faces in facetmap2.
  // We also hash their subfaces in lowest_dim_owner2;


  {
    FacetMapKey key;
    FacetMapKey key2;
    FacetMapRec2 rec;
    for (int fdim = 0; fdim < di; ++fdim) {
      int lsize = sc.brep_levelsize(fdim);
      for (Brep::FaceIndex faceind = 0; faceind < lsize; ++faceind) {
        Brep::Face_Spec fspec(fdim, faceind);
        int nmf = sc.num_meshface_on_face(fspec);
        rec.fspec = fspec;
        for (int seqno = 0; seqno < nmf; ++seqno) {
          log2 << "facetmap2 processing " << fspec << " seqno " << seqno << '\n';
          rec.seqno = seqno;
          for (int j = 0; j < fdim + 1; ++j) {
            key.verts[j] = sc.node_of_meshface_on_face(fspec, seqno, j);
          }
          for (int j2 = fdim + 1; j2 < di; ++j2) {
            key.verts[j2] = next_vnum;
          }
          key.sort();
          facetmap2[key] = rec;

          // Now enter all subfaces into lowest_dim_owner2.

          for (unsigned int mask = 1; mask < (1 << (fdim + 1)); ++mask) {
            int sfdim = 0;
            for (int j = 0; j < fdim + 1; ++j) {
              if (mask & (1 << j))
                key2.verts[sfdim++] = key.verts[j];
            }
            --sfdim;
            for (int j2 = sfdim + 1; j2 < di; ++j2) {
              key2.verts[j2] = next_vnum;
            }
            key2.sort();
            // Use insert so that lowest dimension (earliest insertion) wins.
            pair<LowestDimOwnerMapType2::iterator, bool> rval = 
              lowest_dim_owner2.insert(pair<FacetMapKey,Brep::Face_Spec>(key2,fspec));
            log2 << "mask " << mask << " inserted key " << key2 << " owner " << fspec
              << ((rval.second)? " successful\n" : " unsuccessful\n");
          }
        }
      }
    }
  }


  // In this loop we hash all full-dimensional simplices incident
  // upon brep faces, and their facets.
  // We also hash simplex faces of all dimensions that lie on
  // a low dimensional doubled brep face.


  typedef multimap<FacetMapKey, FacetMapRec1> FacetMapType1;
  FacetMapType1 facetmap1;

  typedef map<FspecSimpMaskKey,int> FacetMapType3;
  FacetMapType3 facetmap3;
          

  {
    FacetMapRec1 rec;
    FspecSimpMaskKey key2;
    FacetMapKey key;
    vector<SimpComplex::VertexGlobalIndex> verts(di + 1);
    vector<Brep::Face_Spec> faceowners(1 << (di + 1));
    int lsize = sc.brep_levelsize(di);
    for (Brep::FaceIndex faceind = 0; faceind < lsize; ++faceind) {
      Brep::Face_Spec fspec(di, faceind);
      int nmf = sc.num_meshface_on_face(fspec);
      for (int seqno = 0; seqno < nmf; ++seqno) {
        // Loop over faces; find if there is a low-dimensional
        // owner.

        log2 << "facetmap1/3 loop processing " << fspec << " seq " << seqno << '\n';

        key2.simpfaceind = faceind;
        key2.simpseqno = seqno; 
        {
          for (int j = 0; j < di + 1; ++j) {
            SimpComplex::VertexGlobalIndex vnum = 
              sc.node_of_meshface_on_face(fspec, seqno, j);
            verts[j] = vnum;
          }
        }
        bool has_doubled_face = false;

        // Loop over faces of the simplex, except the simplex itself.

        faceowners[(1 << (di + 1)) - 1] = fspec;
        for (unsigned int vmask = (1 << (di + 1)) - 2; vmask >= 1; --vmask) {
          int sfdim = 0;
          int skipdim;
          
          // Enter this simplex into facetmap3, if the
          // simplex face we are looping over lies on a low-dimensional
          // brep face.
          {
            for (int j = 0; j < di + 1; ++j) {
              if (vmask & (1 << j)) {
                key.verts[sfdim++] = verts[j];
              }
              else
                skipdim = j;
            }
          }
          --sfdim;
          {
            for (int j = sfdim + 1; j < di; ++j) {
              key.verts[j] = next_vnum;
            }
          }
          key.sort();
          LowestDimOwnerMapType2::iterator it = lowest_dim_owner2.find(key);
          if (it != lowest_dim_owner2.end()) {
            faceowners[vmask] = it -> second;
          }
          else {
            faceowners[vmask] = faceowners[vmask | (1 << skipdim)];
          }
          log2 << "  faceowners[" << vmask << "] = " << faceowners[vmask] << '\n';
          if (num_copies(faceowners[vmask]) > 1) {
            has_doubled_face = true;
            key2.low_dim_fspec = faceowners[vmask];
            key2.mask = vmask;
            facetmap3[key2] = 1;
          }
        }
        
        // If the simplex has a doubled face, then hash all
        // its facets.
        

        if (!has_doubled_face) continue;

        for (int fa = 0; fa < di + 1; ++fa) {
          int k = 0;
          for (int j = 0; j < di + 1; ++j) {
            if (j != fa) {
              key.verts[k++] = verts[j];
            }
          }
          key.sort();
          rec.owner_faceind = faceind;
          rec.owner_seqno = seqno;
          rec.which_facet = fa;
          facetmap1.insert(pair<FacetMapKey, FacetMapRec1>(key,rec));
        }
      }
    }
  }

  // Constructor for the new mesh.
  

  vector<int> levsize1(di + 1);
  {
    for (int fdim = 0; fdim <= di; ++fdim)
      levsize1[fdim] = newb.level_size(fdim);
  }
  
  SimpComplex_Under_Construction newsc(di, di, levsize1);


  // Insert the geo global ID if available.

  {
    string brep_id = newb.lookup_brep_propval(Brep::global_id_propname());    
    if (brep_id.length() > 0) {
      newsc.add_prop_val(Brep::global_id_propname(), brep_id);
    }
  }
  
  // Insert the nodes.

  {
    vector<Real> realcoord(di);
    for (SimpComplex::VertexOrdinalIndex vnumo = 0; vnumo < numnodes; ++vnumo) {
      SimpComplex::VertexGlobalIndex vnum = 
        sc.ordinal_to_global(vnumo);
      LowestDimOwnerMapType::iterator it = lowest_dim_owner.find(vnum);
      int ncopy = 1;
      if (it != lowest_dim_owner.end()) {
        // If found get the owner and see if it is duplicated.
        Brep::Face_Spec owner = it -> second;
        ncopy = num_copies(owner);
      }

      for (int j = 0; j < di; ++j)
        realcoord[j] = sc.real_coord_o(vnumo, j);

      // Insert the copies.

      for (int k = 0; k < ncopy; ++k) {
        newsc.add_vertex(realcoord, vnum + k * next_vnum);
        log2 << " creating node " << vnum + k * next_vnum << " orig " << vnum << '\n';
      }
    }
  }

  // Make a map from new fspecs to old by inverting oldfspec_to_new and 
  // num_copies.

  Brep::Face_Labeling<Brep::FaceIndex> newfspec_to_old_map(newb, -1);
  {
    // Initialize newfspec_to_old_map
    for (Brep::Face_Spec_Loop_Over_Faces fspec(b);
         fspec.notdone();
         ++fspec) {
      int nc = num_copies(fspec);
      Brep::FaceIndex base = oldfspec_to_new_map(fspec);
      for (int ll = 0; ll < nc; ++ll) {
        newfspec_to_old_map(Brep::Face_Spec(fspec.fdim(),base+ll)) = 
          fspec.faceind();
      }
    }
  }

  int passcount = 0;
  int passcount1 = 0;
  vector<FspecSimpMaskKey> searchstack;
  vector<Brep::Face_Spec> faceowners(1 << (di + 1));
  vector<My_bool> is_reachable(1 << (di + 1));
  vector<int> is_a_vertex(1 << (di + 1));
  {
    for (unsigned int mask1 = 0; mask1 < (1 << (di + 1)); ++mask1) {
      int bitcount = 0;
      int whichbit;
      is_a_vertex[mask1] = -1;
      {
        for (int j = 0; j < di + 1; ++j) {
          if (mask1 & (1 << j))  {
            ++bitcount;
            whichbit = j;
          }
        }
      }
      if (bitcount == 1) is_a_vertex[mask1] = whichbit;
    }
  }

  vector<unsigned int> mask_stack;
  vector<SimpComplex::VertexGlobalIndex> sverts(di + 1);
  typedef map<SimpVertKey, int> VertexCopyNumMap;
  VertexCopyNumMap which_vertex_copy_in_simp;
  vector<Brep::Face_Spec> desc_searchstack;
  vector<Brep::Face_Spec> stack_active_faces;
  Brep::Face_Labeling<int> is_an_ib(b);
  Brep::Face_Labeling<int> active_on_pass(b);
  Brep::Face_Labeling<int> whichcopy_on_pass(b);
  Brep::Face_Labeling<int> temp_count(b);
  Brep::Ancestor_Lookup anc_lookup(newb);


  // This is the main loop to insert vertex-boundary-incidences
  // and simplex faces in the new mesh.

  for (int fdim = 0; fdim <= di; ++fdim) {
    vector<Real> param_coord(fdim);
    vector<SimpComplex::VertexGlobalIndex> verts(fdim + 1);
    int lsize = sc.brep_levelsize(fdim);

    for (int faceind = 0; faceind < lsize; ++faceind) {
      Brep::Face_Spec oldfspec(fdim, faceind);

      log2 << "\n----------\nMain processing loop for " << oldfspec << '\n';

      // Create a key for doing a simplex walk, and find an iterator.
  
      FspecSimpMaskKey key2;
      key2.low_dim_fspec = oldfspec;
      key2.simpfaceind = 0;
      key2.simpseqno = 0;
      key2.mask = 0;


      FacetMapType3::iterator startsearch = 
        facetmap3.lower_bound(key2);

      key2.low_dim_fspec = Brep::Face_Spec(fdim, faceind + 1);
      FacetMapType3::iterator endsearch =
        facetmap3.lower_bound(key2);
      Brep::FaceIndex newfaceindb = oldfspec_to_new_map(oldfspec);

      // Loop over copies; create new simplex faces for
      // each copy.

      int numcopy = num_copies(oldfspec);

      log2 << "numcopy = " << numcopy << '\n';

      // If there is more than one copy, must identify which are internal boundaries
      // for the simplex walk.

      if (numcopy > 1) {
        for (; startsearch != endsearch; ++startsearch) {
          if (startsearch -> second > 1) continue;
          log2 << "--- processing copy \n";
          
          // If there is more than one copy,
          // then we have to do a walk over simplices to figure
          // out which copy this is.

          typedef map<FspecSimpMaskKey, int> FacetMapType4;
          FacetMapType4 placed_on_stack;
          ++passcount;

          FspecSimpMaskKey startkey = startsearch -> first;
          searchstack.resize(0, FspecSimpMaskKey());
          searchstack.push_back(startkey);
          placed_on_stack[startkey] = 1;
          stack_active_faces.resize(0, Brep::Face_Spec());
          while (searchstack.size() > 0) {
            FspecSimpMaskKey searchkey = searchstack.back();
            searchstack.pop_back();
            facetmap3[searchkey] = 2;
            Brep::Face_Spec simplexowner(di, searchkey.simpfaceind);
            int simplexseqno = searchkey.simpseqno;

            
            // Pull out the vertices of this simplex.

            log2 << " processing in main search loop " << simplexowner << " seq "
              << simplexseqno << " mask " << searchkey.mask << " maskowner "
              << searchkey.low_dim_fspec << " verts ";

            {
              for (int j = 0; j < di + 1; ++j) {
                sverts[j] = sc.node_of_meshface_on_face(simplexowner, simplexseqno, j);
                log2 << sverts[j] << ' ';
              }
            }
            log2 << '\n';



            // Get all the owners of the faces of this simplex.                             
            
            // Loop over faces of the simplex, except the simplex itself.
            
            faceowners[(1 << (di + 1)) - 1] = Brep::Face_Spec(di, searchkey.simpfaceind);
            for (unsigned int vmask = (1 << (di + 1)) - 2; vmask >= 1; --vmask) {
              int sfdim = 0;
              int skipdim;
              
              FacetMapKey key;
              {
                for (int j = 0; j < di + 1; ++j) {
                  if (vmask & (1 << j)) {
                    key.verts[sfdim++] = sverts[j];
                  }
                  else
                    skipdim = j;
                }
              }
              --sfdim;
              {
                for (int j = sfdim + 1; j < di; ++j) {
                  key.verts[j] = next_vnum;
                }
              }
              key.sort();
              LowestDimOwnerMapType2::iterator it = lowest_dim_owner2.find(key);
              if (it != lowest_dim_owner2.end()) {
                faceowners[vmask] = it -> second;
              }
              else {
                faceowners[vmask] = faceowners[vmask | (1 << skipdim)];
              }
            }

            if (faceowners[searchkey.mask] != oldfspec) 
              throw_error("Data structure inconsistency");

            // Now do a graph search to mark all faces that
            // are on the low-dim face that are reachable
            // from searchkey.mask.

            {
              for (unsigned int mask1 = 0; mask1 < 1 << (di + 1); ++mask1)
                is_reachable[mask1] = false;
            }
            mask_stack.resize(0);
            mask_stack.push_back(searchkey.mask);
            FspecSimpMaskKey key2 = searchkey;
            while (mask_stack.size() > 0) {
              unsigned int newmask = mask_stack.back();
              mask_stack.pop_back();
              if (newmask == 0 || newmask == (1 << (di + 1)) - 1) continue;
              if (is_reachable[newmask]) continue;
              if (faceowners[newmask] != oldfspec) continue;
              is_reachable[newmask] = true;
              key2.mask = newmask;
              placed_on_stack[key2] = 1;
              facetmap3[key2] = 2;
              for (int j = 0; j < di + 1; ++j) {
                unsigned int onebit = 1 << j;
                unsigned int newmask2 = (newmask & onebit) ? 
                  newmask - onebit : newmask + onebit;
                mask_stack.push_back(newmask2);
              }
            }

            // Loop over subfaces.  For each subface that has as
            // a subface marked in the preceding graph search, 
            // check if the subface is a simplex, and if so,
            // add its owner to the list of active faces.

            
            //  In addition, for facets and possibly walk to the neighbor if allowable.
            // its neighbor if the neighbor is allowable.  A neighbor is allowable
            // if the facet has as a subface a mask marked
            // in the preceding graph search, and if the facet
            // does not cross through a doubled ib.

            {
              for (unsigned int mask1 = 1; mask1 < (1 << (di + 1)) - 1; ++mask1) {
                // CHeck if a subface of mask1 is marked as lying on oldfspec.

                unsigned int mask2 = 1;
                for (; mask2 < (1 << (di + 1)) - 1; ++mask2) {
                  if (is_reachable[mask2] && (mask2 & mask1) == mask2) {
                    break;
                  }
                }
                if (mask2 == (1 << (di + 1)) - 1) continue;
                
                // Subface mask2 is a subface of mask1 and 
                // is marked. 
                // Make a key called subfacekey for this facet.

                FacetMapKey subfacekey;
                int sfdim = 0;
                int fa;
                {
                  for (int j = 0; j < di + 1; ++j) {
                    if (mask1 & (1 << j)) {
                      subfacekey.verts[sfdim++] = sverts[j];
                    }
                    else {
                      fa = j;
                    }
                  }
                }
                --sfdim;
                {
                  for (int j = sfdim + 1; j < di; ++j)
                    subfacekey.verts[j] = next_vnum;
                }

                subfacekey.sort();

                // Check if the subface lies on a doubled ib.  If
                // so, make it active.

                FacetMapType2::iterator it5 = facetmap2.find(subfacekey);
                bool is_on_doubled_face = false;
                if (it5 != facetmap2.end() && num_copies(it5 -> second.fspec) > 1) {
                  Brep::Face_Spec owner = it5 -> second.fspec;
                  is_on_doubled_face = true;
                  if (active_on_pass(owner) < passcount) {
                    active_on_pass(owner) = passcount;
                    stack_active_faces.push_back(owner);
                    if (sfdim == di - 1) {
                      int which_side = determine_side(b, owner, sc, simplexowner,
                        simplexseqno, fa, patch_is_flipped, scfac, tol);
                      whichcopy_on_pass(owner) = which_side;
                    }
                  }
                }
                if (sfdim < di - 1 || is_on_doubled_face) continue;
                // Get here if the face is a legal facet and does not lie on
                // a doubled ib.  Find the next simplex over.
                
                pair<FacetMapType1::iterator, FacetMapType1::iterator> rval = 
                  facetmap1.equal_range(subfacekey);
                if (rval.first == rval.second)
                  throw_error("Facetmap inconsistency 1 - missing key");
                FacetMapType1::iterator foundptr = facetmap1.end();
                FacetMapType1::iterator otherptr = facetmap1.end();
                for (FacetMapType1::iterator it3 = rval.first; 
                it3 != rval.second; 
                ++it3) {
                  if (it3 -> second.owner_faceind == simplexowner.faceind() &&
                    it3 -> second.owner_seqno == simplexseqno) {
                    if (foundptr != facetmap1.end())
                      throw_error("Facetmap inconsistency 2 - same simplex appears twice");
                    foundptr = it3;
                  }
                  else {
                    if (otherptr != facetmap1.end())
                      throw_error("Facetmap inconsistency 3 - other simplex appears twice");
                    otherptr = it3;
                  }
                }
                
                if (foundptr == facetmap1.end())
                  throw_error("Factmap inconsistency 5 -- didn't find simplex in facetmap");
                if (otherptr != facetmap1.end()) {
                  // Put the simplex pointed to by otherptr on the
                  // stack.  We have to compute a mask for that 
                  // other simplex which is messy.  We take mask2
                  // and renumber the bits corresponding to the bits
                  // in the other simplex.
                  
                  unsigned int othermask = 0;
                  Brep::Face_Spec otherowner(di, otherptr -> second.owner_faceind);
                  int otherseqno = otherptr -> second.owner_seqno;
                  for (int j = 0; j < di + 1; ++j) {
                    if ((mask2 & (1 << j)) == 0) continue;
                    SimpComplex::VertexGlobalIndex vnum = sverts[j];
                    // Find vnum in the neighbor.
                    int otherj = 0;
                    for (; otherj < di + 1; ++otherj) {
                      if (sc.node_of_meshface_on_face(otherowner, otherseqno, otherj) == vnum) {
                        break;
                      }
                    }
                    if (otherj == di + 1) {
                      throw_error("Inconsistent data structure");
                    }
                    othermask |= (1 << otherj);
                  }
                  FspecSimpMaskKey otherkey;
                  otherkey.simpfaceind = otherowner.faceind();
                  otherkey.simpseqno = otherseqno;
                  otherkey.low_dim_fspec = oldfspec;
                  otherkey.mask = othermask;
                  pair<FacetMapType4::iterator, bool> rval = 
                    placed_on_stack.insert(pair<FspecSimpMaskKey, int>(otherkey, 1));
                  if (rval.second) {
                    searchstack.push_back(otherkey);
                    log2 << "  pushing " << otherowner << " seq "
                      << otherseqno << " mask " << othermask << " maskowner "
                      << oldfspec << '\n';
                  }
                }
              }
            }
          }

          
          // When we get here, we have searched neighboring simplices
          // and also marked active_on_pass for full-dimensional
          // faces.
          
          
          // Get here: the search is finished.  Next we have to
          // figure out which copy of each brep face was encountered
          // on the preceding search.  This was already taken care
          // of for faces of dimension di - 1, so we have to do it
          // only for faces of lower dimension.   

          log2 << " search finished; now labeling brep faces\n";
          {
            
            for (int j = 0; j < stack_active_faces.size(); ++j) {
              Brep::Face_Spec oldfspec1 = stack_active_faces[j];
              log2 << " oldfspec1 = " << oldfspec1 << '\n';
              if (oldfspec1.fdim() == di - 1) {
                log2 << "   copy " << whichcopy_on_pass(oldfspec1) << '\n';
              }
              if (oldfspec1.fdim() >= di - 1) continue;
              Brep::FaceIndex newfspecb = oldfspec_to_new_map(oldfspec1);
              int numc = num_copies(oldfspec1);
              log2 << "  numc = " << numc << '\n';
              int fdim1 = oldfspec1.fdim();
              bool found1 = false;
              int which_copy_found;
              for (int which_copy = 0; which_copy < numc; ++which_copy) {
                Brep::Face_Spec newfspec(fdim1, newfspecb + which_copy);
                log2 << " trying copy " << which_copy << " with new fspec " << newfspec << '\n';
                bool found2 = true;
                for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop(anc_lookup, newfspec);
                ancloop.notdone();
                ++ancloop) {
                  Brep::Face_Spec new_afspec = ancloop.ancestor_fspec();
                  if (new_afspec.fdim() != di - 1) continue;
                  Brep::Face_Spec old_afspec(di - 1,  newfspec_to_old_map(new_afspec));
                  if (active_on_pass(old_afspec) < passcount) {
                    if (num_copies(old_afspec) > 1) {
                      log2 << " dismissing -- ancestor " << old_afspec 
                        << " not active on pass but mult copies\n";
                      found2 = false;
                      break;
                    }
                    else {
                      continue;
                    }
                  }
                  
                  Brep::FaceIndex new_afspecb = oldfspec_to_new_map(old_afspec);
                  int a_which_copy = new_afspec.faceind() - new_afspecb;
                  if (a_which_copy < 0 || a_which_copy > 1) 
                    throw_error("Unexpected range problem");
                  if (active_on_pass(old_afspec) < passcount || 
                    whichcopy_on_pass(old_afspec) != a_which_copy) {
                    log2 << " dismissing; ancestor " << old_afspec 
                      << " has copy " << whichcopy_on_pass(old_afspec) << " seen on "
                      << " pass but anc of this copy is " << a_which_copy << '\n';
                    found2 = false;
                    break;
                  }
                }
                if (found2) {
                  if (found1) {
                    throw_error("Ambiguous choice of which_copy?");
                  }
                  found1 = true;
                  which_copy_found = which_copy;
                }
              }
              if (!found1) {
                throw_error("No allowable choice of which_copy?");
              }
              whichcopy_on_pass(oldfspec1) = which_copy_found;
            }
          }

          int whichcopy = whichcopy_on_pass(oldfspec);

          log2 << " identified whichcopy of " << oldfspec << " = " << whichcopy << '\n';
          log2 << "Now labeling low dimensional simplices\n";

          // Copies are now identified.  Now loop over simplices
          // to enter vertex info into which_vertex_copy_in_simp.

          for (FacetMapType4::iterator it = placed_on_stack.begin();
          it != placed_on_stack.end();
          ++it) {
            FspecSimpMaskKey key1 = it -> first;
            log2 << "pulled from stack simplex " << Brep::Face_Spec(di, key1.simpfaceind) << " seq "
              << key1.simpseqno << " mask " << key1.mask << " owner " << key1.low_dim_fspec << "\n";
            if (is_a_vertex[key1.mask] == -1) continue;
            int whichvtx = is_a_vertex[key1.mask];
            Brep::Face_Spec simplexowner(di, key1.simpfaceind);

            // Loop over faces of the simplex, and find
            // the ones that are simplices of sc.

            // Pull out the vertices of this simplex.

            {
              for (int j = 0; j < di + 1; ++j) {
                sverts[j] = sc.node_of_meshface_on_face(simplexowner, key1.simpseqno, j);
                log2 << sverts[j] << ' ';
              }
            }
            log2 << '\n';

            for (unsigned int mask1 = 1; mask1 < 1 << (di + 1); ++mask1) {
              // Make sure this face owns the vertex.
              if ((mask1 & key1.mask) != key1.mask) continue;
              bool is_a_simplex = false;
              Brep::Face_Spec new_owner;
              int seqno;
              log2 << " for mask " << mask1;
              // If the face is full dimensional, it is automatically
              // a simplex
              if (mask1 == (1 << (di + 1)) - 1) {
                is_a_simplex = true;
                // No renumbering of regions since none are doubled.
                new_owner = Brep::Face_Spec(di, key1.simpfaceind);
                seqno = key1.simpseqno;
              }
              else {
                // Make a key for this subface.
                FacetMapKey key2;
                int k = 0;
                for (int j = 0; j < di + 1; ++j) {
                  if (mask1 & (1 << j)) 
                    key2.verts[k++] = sverts[j];
                }
                while (k < di) 
                  key2.verts[k++] = next_vnum;
                key2.sort();
                log2 << " key " << key2;
                FacetMapType2::iterator it2 = facetmap2.find(key2);
                if (it2 == facetmap2.end()) {
                  is_a_simplex = false;
                }
                else {
                  is_a_simplex = true;
                  Brep::Face_Spec owner = it2 -> second.fspec;
                  seqno = it2 -> second.seqno;
                  // Now we must figure out the new owner.
                  int which_copy1 = 0;
                  if (num_copies(owner) > 1) {
                    if (active_on_pass(owner) < passcount)
                      throw_error("Owner should have been active on pass");
                    which_copy1 = whichcopy_on_pass(owner);
                  }
                  new_owner = Brep::Face_Spec(owner.fdim(), 
                    oldfspec_to_new_map(owner) + which_copy1);
                }
              }
              log2 << " is a simplex = " << is_a_simplex << ((is_a_simplex)? "" : "\n");

              if (!is_a_simplex) continue;

              log2 << " new owner = " << new_owner << " seqno = " << seqno 
                << " vertex = " << sverts[whichvtx]
                << '\n';

              SimpVertKey key3;
              key3.fspec = new_owner;
              key3.seqno = seqno;
              key3.vertex = sverts[whichvtx];
              which_vertex_copy_in_simp[key3] = whichcopy;
            }
          }
        }
      }

      // Now create the new simplex faces.

      for (int whichcopy = 0; whichcopy < numcopy; ++whichcopy) {

        Brep::Face_Spec newfspec(fdim, oldfspec_to_new_map(oldfspec) + whichcopy);

        // First insert the vertices.  This requires us to know
        // which copies of each descendent face are owned by
        // newfspec.  So make a map desc_list that holds all
        // descendants and which copy is included.


        typedef map<FspecCopyNum, int> DescListType;
        DescListType desc_list;

        desc_searchstack.resize(0,Brep::Face_Spec());
        desc_searchstack.push_back(newfspec);
        while (desc_searchstack.size() > 0) {
          Brep::Face_Spec newfspec1 = desc_searchstack.back();
          desc_searchstack.pop_back();
          Brep::FaceIndex oldfaceind1 = newfspec_to_old_map(newfspec1);
          Brep::Face_Spec oldfspec1(newfspec1.fdim(), oldfaceind1);
          Brep::FaceIndex nfaceind = oldfspec_to_new_map(oldfspec1);
          int whichcopy1 = newfspec1.faceind() - nfaceind;
          FspecCopyNum fspecc;
          fspecc.fspec = oldfspec1;
          fspecc.copynum = whichcopy1;
          desc_list[fspecc] = 1;
          for (Brep::Face_Spec_Loop_Over_Face_Subfaces newsubfspec(newb, newfspec1);
          newsubfspec.notdone();
          ++newsubfspec) {
            desc_searchstack.push_back(newsubfspec);
          }
        }

        // Now insert the nodes for this face.

        if (fdim < di) {
          int nnf = sc.num_node_on_face(oldfspec);
          for (int seqno = 0; seqno < nnf; ++seqno) {
            SimpComplex::VertexGlobalIndex vnum = sc.node_on_face(oldfspec,seqno);
            Brep::PatchIndex patchind = sc.patchind_on_face(oldfspec,seqno);
            {
              for (int j = 0; j < fdim; ++j) {
                param_coord[j] = sc.param_coord_on_face(oldfspec,seqno,j);
              }
            }
            typedef map<PatchInBrepAddress,bool> MPBIB;
            MPBIB::const_iterator it4 = 
              patch_is_flipped.find(PatchInBrepAddress(oldfspec,patchind));
            if (it4 == patch_is_flipped.end())
              throw_error("patch missing from patch_is_flipped");
            if (it4 -> second) {
              if (fdim == 1) {
                param_coord[0] = 1.0 - param_coord[0];
              }
              else if (fdim == 2 && b.patch_type(oldfspec,patchind) == BEZIER_TRIANGLE) {
                param_coord[0] = 1.0 - param_coord[0] - param_coord[1];
              }
              else if (fdim == 2 && b.patch_type(oldfspec,patchind) == BEZIER_QUAD) {
                param_coord[0] = 1.0 - param_coord[0];
              }
            }


            // Now insert copies of this node in the new sc.
            Brep::Face_Spec ldowner = lowest_dim_owner[vnum];
            FspecCopyNum key1;
            key1.fspec = ldowner;
            key1.copynum = 0;
            DescListType::iterator beginsearch = desc_list.lower_bound(key1);
            key1.copynum = maxnumcopy;
            DescListType::iterator endsearch = desc_list.upper_bound(key1);
            if (beginsearch == endsearch)
              throw_error("no copies of subface?");
            for (DescListType::iterator it1 = beginsearch; it1 != endsearch; ++it1) {
              int whichcopy1 = it1 -> first.copynum;
              newsc.add_vertex_bdryinc(vnum + whichcopy1 * next_vnum, newfspec,
                patchind, param_coord);
            }
          }
        }

        // Now add the simplices.

        if (fdim > 0) {
          int nmf = sc.num_meshface_on_face(oldfspec);
          SimpVertKey key3;
          key3.fspec = newfspec;
          for (int seqno = 0; seqno < nmf; ++seqno) {
            {
              for (int j = 0; j < fdim + 1; ++j) {
                SimpComplex::VertexGlobalIndex vnum = 
                  sc.node_of_meshface_on_face(oldfspec, seqno, j);
                Brep::Face_Spec oldowner = lowest_dim_owner[vnum];
                if (num_copies(oldowner) > 1) {
                  key3.seqno = seqno;
                  key3.vertex = vnum;
                  VertexCopyNumMap::iterator it =
                    which_vertex_copy_in_simp.find(key3);
                  if (it == which_vertex_copy_in_simp.end())
                    throw_error("Missing key in which_vertex_copy");
                  vnum += (it -> second) * next_vnum;
                }
                verts[j] = vnum;
              }
            }
            newsc.add_simplex_face(verts, newfspec);
          }
        }
      }
    }
  }
  return newsc;
}
