// ------------------------------------------------------------------
// meshgen.cpp
//
// This file contains the main routine for mesh generation.
// ------------------------------------------------------------------
// Author: Stephen A. Vavasis
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------

#include <iomanip>

#ifdef _MSC_VER
#if _MSC_VER == 1200
#include "qmatvec.h"
#endif
#endif

#include "qboxstack.h"
#include "qlog.h"
#include "qsizectl.h"
#include "qerr.h"
#include "qseparation.h"
#include "qmg_gui.h"
#include "qlifter.h"


// declare extern functions.


QMG::ostream* debug_ostream;

namespace QMG {
  namespace MG {

    using namespace QMG;


    // This is the function exported from this file.

    SimpComplex_Under_Construction meshgen(const Brep& g,
      const string& default_size_control,
      const string& default_curv_control,
      const string& user_data,
      const vector<Real>& expa,
      const vector<Real>& asp_degrade,
      ostream& logfile,
      int verbosity1,
      QMG::FrontEnd::Interpreter& interp,
      bool gui_active,
      double tol,
      int ori);

    // See align.cpp for these.

    extern void alignInitStaticData(int di); 


    
    extern void align_orbit(ActiveBoxVec& alignvec,
      int closefacedim,
      Real maxwarpdist,
      ActiveBoxVec& idlevec,
      ActiveBoxVec& smallboxvec,
      ActiveBoxVec& small_idle_boxvec,
      const Brep::Face_Spec& alignfspec,
      BoxFaceDag& dag,
      SimpComplex_Under_Construction& sc,
      vector<Brep::Face_Spec>& vtx_owner,
      const vector<Real>& asp_degrade,
      Logstream& logstr,
      Meshgen_gui& gui,
      const PatchTable& patchtable, 
      IncidenceTable& inc_table,
      ActiveBoxVec::Create_Subbox_Workspace& wkspa);


    extern void align_orbit_lastphase(ActiveBoxVec& alignvec,
      ActiveBoxVec& smallboxvec,
      const Brep::Face_Spec& alignfspec,
      BoxFaceDag& dag,
      SimpComplex_Under_Construction& sc,
      vector<Brep::Face_Spec>& vtx_owner,
      Logstream& logstr,
      Meshgen_gui& gui);


    // see qboxstack.cpp
    extern void boxstack_init_static_data(int di); 

    // see qsmall.cpp
    extern void small_initialize_static_data(int di);
  }
}


// Variables and defs with file scope.


namespace {
  using namespace QMG;
  using namespace QMG::MG;


  // ------------------------------------------------------------------
  // Boxvecs_by_level
  // An array of ActiveBoxVec's, indexed by level.

  class Boxvecs_by_level {
  private:
    vector<ActiveBoxVec*> vecs_;
    Real expa_;
    void enlarge_(int lev) {
      if (lev >= vecs_.size()) {
        vecs_.resize(lev + 1, 0);
      }
    }
    Boxvecs_by_level(const Boxvecs_by_level&) { }
    void operator=(const Boxvecs_by_level&) { }
  public:
    ActiveBoxVec* operator[](int lev) {
      enlarge_(lev);
      return vecs_[lev];
    }
    ActiveBoxVec& force_creation(int lev) {
      enlarge_(lev);
      if (!vecs_[lev]) {
        vecs_[lev] = new ActiveBoxVec(expa_, lev);
      }
      return *(vecs_[lev]);
    }
    void reset_expa(Real expa) {
      for (int i = 0; i < vecs_.size(); ++i) {
        if (vecs_[i])
          vecs_[i] -> reset_expa(expa);
      }
      expa_ = expa;
    }
    void erase(int lev) {
      enlarge_(lev);
      if (vecs_[lev])
        delete vecs_[lev];
      vecs_[lev] = 0;
    }
    int size() const {return vecs_.size();}
    Boxvecs_by_level(Real expa) : expa_(expa) { }
    ~Boxvecs_by_level() {
      for (int lev = vecs_.size() - 1; lev >= 0; --lev) {
        if (vecs_[lev]) delete vecs_[lev];
      }
    }

    void dump_size(ostream& os) {
      for (int lev = 0; lev < vecs_.size(); ++lev) {
        os << "Level " << lev;
        if (vecs_[lev]) {
          os << " size " << vecs_[lev] -> numboxes() << '\n';
        }
        else {
          os << " not created\n";
        }
      }
    }
  };


  // ------------------------------------------------------------------
  // This local routine adds a simplex to the simplicial complex.
  // It computes the face owners and lifts parameters


  void add_simplex(SimpComplex_Under_Construction& sc,
    const vector<SimpComplex::VertexOrdinalIndex>& vertices,
    int surf_ori,
    int ori,
    const vector<Brep::Face_Spec>& vtx_owner,
    const Brep_Parameter_Lifter& lifter,
    // workspaces:
    vector<SimpComplex::VertexGlobalIndex>& vlist1,
    vector<Real>& param1,
    vector<Point>& params,
    vector<Brep::PatchIndex>& patchinds
    ) {
    int di = sc.embedded_dim();

    SimpComplex::VertexOrdinalIndex vnumf_o = vertices[di];
    Brep::Face_Spec lastowner = vtx_owner[vnumf_o];
    SimpComplex::VertexGlobalIndex vnumf = sc.ordinal_to_global(vnumf_o);
    if (lastowner.fdim() < di) {
      patchinds[di] = sc.retrieve_vertex_bdryinc_patchind(vnumf, lastowner);
      params[di] = sc.retrieve_vertex_bdryinc_param(vnumf,lastowner);
    }
    vlist1.resize(1);
    vlist1[0] = vnumf;
    for (int j1 = di - 1; j1 >= 0; --j1) {
      SimpComplex::VertexOrdinalIndex vnum_o = vertices[j1];
      SimpComplex::VertexGlobalIndex vnum = sc.ordinal_to_global(vnum_o);
      Brep::Face_Spec newowner = vtx_owner[vnum_o];
      int newfdim = newowner.fdim();
#ifdef DEBUGGING
      if (newfdim < lastowner.fdim() ||
          (newfdim == lastowner.fdim() && newowner != lastowner)) {
        throw_error("problem with vertex owner sequence");
      }
#endif
      if (newfdim < di) {
        patchinds[j1] = sc.retrieve_vertex_bdryinc_patchind(vnum, newowner);
        params[j1] = sc.retrieve_vertex_bdryinc_param(vnum,newowner);
        if (newowner != lastowner) {
          for (int j2 = j1 + 1; j2 < di + 1; ++j2) {
            SimpComplex::VertexOrdinalIndex vnum_o2 = vertices[j2];
            SimpComplex::VertexGlobalIndex vnum2 = sc.ordinal_to_global(vnum_o2); 
            pair<Point, Brep::PatchIndex> rval = lifter.lift_parameters(lastowner,
              patchinds[j2], params[j2], newowner);
            param1.resize(newfdim);
            for (int j3 = 0; j3 < newfdim; ++j3)
              param1[j3] = rval.first[j3];
            sc.add_vertex_bdryinc(vnum2, newowner, rval.second, param1);
            params[j2] = rval.first;
            patchinds[j2] = rval.second;
          }
        }
      }
      lastowner = newowner;
      vlist1.push_back(vnum);
      if (newfdim == di - j1) {
        bool swap1 = false;
        if (newfdim == di && ori) 
          swap1 = true;
        if (newfdim == di - 1) {
          if (surf_ori > 1) {
            throw_error("Missing surf ori");
          }
          if (di % 2 == 0)
            surf_ori = 1 - surf_ori;
          if ((surf_ori && ori) || (!surf_ori && !ori) )
            swap1 = true;
        }
        if (swap1) {
          SimpComplex::VertexGlobalIndex tmp = vlist1[0];
          vlist1[0] = vlist1[1];
          vlist1[1] = tmp;
        }
        sc.add_simplex_face(vlist1, newowner);
        if (swap1) {
          SimpComplex::VertexGlobalIndex tmp = vlist1[0];
          vlist1[0] = vlist1[1];
          vlist1[1] = tmp;
        }
      }
    }
  }

  
  // ------------------------------------------------------------------
  // The next few functions were used for debugging.  They compute
  // the total weight in all ActiveBoxVec's of a certain point.

  pair<bool, Base3>
    coord_is_inside_box(const ActiveBox& b,
                        const IntCoord& coord) {
    bool is_inside = true;
    Base3 c_rindex = 0;
    IntCoord lb = b.boxaddress().lowerleft();
    IntCoord ub = lb;
    int di = b.embedded_dim();
    int width = b.iwidth();

    for (int jj = 0; jj < di; ++jj) {
      if (!b.flatdim(jj))
        ub.increment(jj, width);
    }
    int reldim = 0;
    for (int j = 0; j < di; ++j) {
      if (coord[j] < lb[j]) {
        is_inside = false;
        break;
      }
      else if (coord[j] > ub[j]) {
        is_inside = false;
        break;
      }
      else if (!b.flatdim(j)) {
        int newdig;
        if (coord[j] == lb[j])
          newdig = 0;
        else if (coord[j] == ub[j])
          newdig = 1;
        else
          newdig = 2;
        c_rindex.set_digit(reldim++,newdig);
      }
    }
    return pair<bool, Base3>(is_inside, c_rindex);
  }
  
  
  Weight_t sum_weight_stack(const ActiveBoxVec& vec,
                            const IntCoord& coord, 
                            int iw, 
                            ActiveBoxVec::Create_Subbox_Workspace& wkspa) {
    Weight_t sum = 0;
    int di = vec.embedded_dim();
    int width = vec.iwidth();
    BoxAddress bxa;
    bxa.lowerleft()[0] = coord[0];
    bxa.lowerleft()[1] = coord[1];
    if (di == 3) {
      bxa.lowerleft()[2] = coord[2];
      bxa.flatdim() = 7;
    } 
    else {
      bxa.flatdim() = 3;
    }
    for (ActiveBoxVec::Loop_over_boxes loop(vec); loop.notdone(); ++loop) {
      ActiveBox b = loop.getbox();
      pair<bool, Base3> rval = 
        coord_is_inside_box(b, coord);
      bool is_inside = rval.first;
      if (is_inside) {
        Base3 c_rindex = rval.second;
        *debug_ostream << "***contained by ";
        //      b.full_dump2(*debug_ostream, patchtable, inc_table);
        if (width > 6)
          b.full_dump2(*debug_ostream, wkspa.patchtable, wkspa.inc_table);
        else
          *debug_ostream << b;
        *debug_ostream << " face ";
        c_rindex.output(*debug_ostream, b.dim());
        *debug_ostream << " with weight " << b.weight(c_rindex);
        ActiveBoxVec tmpvec(.01,iw);
        ActiveBoxVec::create_subbox(ActiveBox_As_Parent(b), bxa, iw,
          di, 0, 0, 0, tmpvec, tmpvec, wkspa);
        if (tmpvec.numboxes() > 0) {
          *debug_ostream << " actually inside\n";
          sum += b.weight(c_rindex);
        }
        else {
          *debug_ostream << " not inside\n";
        }
      }
    }
    return sum;
  }
  
  Weight_t sum_weight_bvl(Boxvecs_by_level& sep,
                          const IntCoord& coord,
                          int iw,
                          ActiveBoxVec::Create_Subbox_Workspace& wkspa) {
    Weight_t sum = 0;
    for (int level = 0; level < sep.size(); ++level) {
      ActiveBoxVec* vecp = sep[level];
      if (vecp)
        sum += sum_weight_stack(*vecp, coord, iw, wkspa);
    }
    return sum;
  }

  Weight_t sum_weight_orbit(Brep_Orbits& all_orbits,
                            const Brep::Face_Spec& fspec,
                            const IntCoord& coord,
                            int iw,
                            ActiveBoxVec::Create_Subbox_Workspace& wkspa){
    int level = 0;
    ActiveBoxVec* vecp = all_orbits.find_next_nonempty_level(fspec, level);
    Weight_t sum = 0;
    while (vecp) {
      level = vecp -> iwidth();
      sum += sum_weight_stack(*vecp, coord, iw, wkspa);
      vecp = all_orbits.find_next_nonempty_level(fspec, level+1);
    }
    return sum;
  } 

  void detect_weight_problem_at_point(Boxvecs_by_level& sep0, 
                                      Boxvecs_by_level& sep1, 
                                      Brep_Orbits& all_orbits,
                                      const Brep& g,
                                      const IntCoord& coord,
                                      int testiw,
                                      ActiveBoxVec::Create_Subbox_Workspace& wkspa,
                                      bool throw_if_error) {
    

    *debug_ostream << "Coord = ";
    for (int i1 = 0; i1 < g.embedded_dim(); ++i1) {
      *debug_ostream << std::hex << coord[i1] << " " << std::dec;
    }
    *debug_ostream << "\nreal coord = ";
    for (int i2 = 0; i2 < g.embedded_dim(); ++i2) {
      *debug_ostream << std::setprecision(17) 
        << coord[i2] * BoxAddress::multiplier() / (1 << MAXLEV)
        + BoxAddress::global_coordinate_origin(i2) << " ";
    }
    *debug_ostream << '\n';
  


    Weight_t sum = 0;
    *debug_ostream << " weight test sep 0\n";
    sum += sum_weight_bvl(sep0, coord, testiw, wkspa);
    *debug_ostream << " weight test sep 1\n";
    sum += sum_weight_bvl(sep1, coord, testiw, wkspa);
    for (Brep::Face_Spec_Loop_Over_Faces fspec(g); fspec.notdone(); ++fspec) {
      *debug_ostream << " weight test orbit " << fspec << "\n";
      sum += sum_weight_orbit(all_orbits, fspec, coord, testiw, wkspa);
    }
    *debug_ostream << " weight sum in detect = " << sum << '\n';
    if (sum  < .99999 || sum > 1.00001) {
      *debug_ostream << " illegal sum = " << sum << std::endl;
      if (throw_if_error)
        throw_error("Weight error detected");
    }
  }


  // ------------------------------------------------------------------
  // This class exists to cause an error message to be written to the log,
  // and then for the log to flushed whenever an error is thrown.  

  class OnErrorFlushLog : public QMG::Error_Task {
  private:
    Logstream& log_;
    // Override base case function
    virtual void on_error_do(const string& errmsg) { 
      log_.str() << errmsg << '\n';
      QMG::Error_Message::put_error_messages_into_stream(log_.str());
      log_.str() << std::flush;
    }
  public:
    OnErrorFlushLog(Logstream& logs) : log_(logs) { }
    ~OnErrorFlushLog() { }
  };

  // ------------------------------------------------------------------
  // This class creates a button labeled "dismiss" when an error
  // occurs during mesh generation.

  class OnErrorCreateDismissButton : public QMG::Error_Task {
  private:
    Meshgen_gui& gui_;
    virtual void on_error_do(const string& errmsg) {
      gui_.indicate_error_has_occurred(errmsg);
    }
  public:
    OnErrorCreateDismissButton(Meshgen_gui& gui) : gui_(gui) { }
    ~OnErrorCreateDismissButton() { }
  };
}

#ifdef DEBUGGING
namespace QMG {
  namespace Local_to_PatchMath_CPP {
    using namespace QMG;
    extern bool dump_everything;
    extern ostream* debug_str;
  }
}
#endif

#ifdef NSF_DEMO
namespace QMG {
  namespace Nsf_Demo {
    using namespace QMG;
    using namespace QMG::MG;
    extern void indicate_phase_start(const string& phasename);
  }
}
#endif
          

// ------------------------------------------------------------------
// meshgen is the main mesh generation routine.


QMG::SimpComplex_Under_Construction
QMG::MG::meshgen(const Brep& g,
                 const string& default_size_control,
                 const string& default_curv_control,
                 const string& user_data,
                 const vector<Real>& expa,
                 const vector<Real>& asp_degrade,
                 ostream& logfile,
                 int verbosity1,
                 QMG::FrontEnd::Interpreter& interp,
                 bool gui_active,
                 double tol,
                 int ori) {




  using std::endl;

  // initialize the log

  Logstream logstr(logfile, verbosity1);

  // Initialize the object to cause errors to go to the log.

  OnErrorFlushLog unnamed1_(logstr);
  debug_ostream = &logstr.str();

  // initialize the random number generator.

  random_real(1);


  int di = g.embedded_dim();

  // Check the ranges of parameters.

  if (g.gdim() != di) 
    throw_error("Intrinsic and embedded dims of input brep must be equal.");

  if (di < MINDIM || di > MAXDIM) 
    throw_error("Dimension of brep is out of range.");

  if (g.level_size(di) == 0) 
    throw_error("Brep has no chambers.");

  if (expa.size() != di + 1)
    throw_error("Wrong size for expa");
  if (asp_degrade.size() != di + 1)
    throw_error("Wrong size for asp_degrade");
  {
    for (int i = 0; i < di + 1; ++i) {
      if (expa[i] < 0 || expa[i] > 1) 
        throw_error("expansion factor is out of range (must be in [0,1])");
      if (asp_degrade[i] < 0 || asp_degrade[i] > 1) 
        throw_error("asp_degrade factor is out of range (must be in [0,1])");
    }
  }

  if (ori != 1 && ori != 2)
    throw_error("invalid value for 'orientiation' argument");


  // Print the calling arguments in the log.

  if (logstr.verbosity() >= 1) {
    logstr.str() << "QMG 2.0.2 MESH GENERATION LOG (verbosity = " 
      << logstr.verbosity () << ") started on ";
    logstr.logdate();
    logstr.str() << endl << "Input brep:\n" << g << endl;
    logstr.str() << "Default size function = '" << default_size_control << "'" << endl;
    logstr.str() << "Default curvature function = '" << default_curv_control << "'" << endl;
    logstr.str() << "User data = " << user_data << endl;
    logstr.str() << "expa = ";
    {
      for (int j = 0; j < asp_degrade.size(); ++j)
        logstr.str() << expa[j] << " ";
    }
    logstr.str() << "asp degrade = ";
    {
      for (int j = 0; j < asp_degrade.size(); ++j)
        logstr.str() << asp_degrade[j] << " ";
    }
    logstr.str() << endl;
    logstr.str() << "verbosity = " << verbosity1 << endl;
    logstr.str() << "gui_active = " << gui_active << endl;
    logstr.str() << "tol = " << tol << endl;
  }



  //Initialize static data.

  small_initialize_static_data(di);
  boxstack_init_static_data(di); 
  alignInitStaticData(di);
  Point::InitStaticData(di);
#ifdef DEBUGGING
  Local_to_PatchMath_CPP::dump_everything = false;
  Local_to_PatchMath_CPP::debug_str = &logstr.str();
#endif


 /*
 
IntCoord coord;
coord[0] = 0x4000000U;
coord[1] = 0xc000000U;
int testiw = 4;
*/

  // Create the mesh generation gui panel.

  Meshgen_gui gui(gui_active, interp, g, logstr);

  // Set up a class to insert a button labeled "Dismiss"
  // if an error occurs.

  OnErrorCreateDismissButton unnamed2_(gui);


  gui.operation(string("Preprocessing Brep"));
  double scfac = g.scale_factor();

  vector<int> level_size0(di + 1);
  for (int fdim = 0; fdim <= di; ++fdim)
    level_size0[fdim] = g.level_size(fdim);

  SimpComplex_Under_Construction sc(di, di, level_size0);

  
  string brep_id = g.lookup_brep_propval(Brep::global_id_propname());

  if (brep_id.length() == 0) {
    if (logstr.verbosity() >= 1) {
      logstr.str() << "global id propval not found\n";
    }
  }
  else {
    if (logstr.verbosity() >= 1) {
      logstr.str() << "global id propval value = " << brep_id << '\n';
    }
    sc.add_prop_val(Brep::global_id_propname(), brep_id);
  }

  vector<Brep::Face_Spec> vtx_owner;

  BoxFaceDag dag(di);

  gui.associate_sc_and_dag(sc, dag);

 
  gui.operation("Quadtree generation");

  if (logstr.verbosity() >= 1) 
    logstr.str() << "\n\n-------------- Quadtree generation --------------------" << endl;
  
  // Open brace to make separation stacks and orbits vanish after this code block.

  {
  
    // Make the patch table.  Initialize the incidence table.
    PatchTable patchtable(g, logstr, scfac, tol);    
    IncidenceTable inc_table;


    // Gather size control info.
    
    SizeControl size_control(g, 
      patchtable,
      "sizecontrol",
      "gm_sizecontrol",
      default_size_control,
      user_data,
      interp,
      logstr);
    
    SizeControl curvature_control(g,
      patchtable,
      "curvecontrol",
      "gm_curvecontrol",
      default_curv_control,
      user_data,
      interp,
      logstr);
    
    // Generate the top box and push it on a stack.
    
    Boxvecs_by_level sep_stack0(expa[0]);
    Boxvecs_by_level sep_stack1(expa[1]);
    Brep_Orbits all_orbits(g,expa);


    // Initialize the separation operator.
    
    Separation_Functor separation(g, patchtable, inc_table, size_control,
      curvature_control, all_orbits, scfac, tol, logstr, gui);

    // Push top box on stack -- top box contains the whole object.
    
    
    ActiveBoxVec::Create_Subbox_Workspace crt_wkspa(patchtable,
      inc_table, logstr, gui, scfac, tol);

    sep_stack0.force_creation(0).push_top_box(g, inc_table, crt_wkspa);

    int max_lev_so_far = 0;
    

    for (int phase = 0; phase <= di; ++phase) {
      if (logstr.verbosity() >= 1) {
        logstr.str() << "\n\n------------------ Begin phase " << phase << " ";
        logstr.logdate();
        logstr.str() << endl;
      }
      // For even phases the input box list is sep_stack0 and
      // the output is sep_stack1.  This is reversed for odd phases.
      
      Boxvecs_by_level& input_boxvec_by_level =  (phase & 1) ?  sep_stack1 : sep_stack0;
      Boxvecs_by_level& idle_boxvec_by_level = (phase & 1) ? sep_stack0 : sep_stack1;

      if (phase < di)
        idle_boxvec_by_level.reset_expa(expa[phase + 1]);


     
      if (logstr.verbosity() >= 2) {
        logstr.str() << " input boxvec sizes: \n";
        input_boxvec_by_level.dump_size(logstr.str());
        logstr.str() << " idle boxec sizes: \n";
        idle_boxvec_by_level.dump_size(logstr.str());
      }


      {
        ostringstream ost;
        ost << "Phase " << phase << " separation";
        gui.operation2(ost.str());
#ifdef NSF_DEMO
        QMG::Nsf_Demo::indicate_phase_start(ost.str());
#endif
      }

      vector<ActiveBoxVec::Index> first_live_box;
      int level = input_boxvec_by_level.size();

      
      // Now do the separation, always finding the deepest box first.
    
      for (;;) {
        ++level;
        while (level >= 0 &&
          (input_boxvec_by_level[level] == 0 ||
          input_boxvec_by_level[level] -> numboxes() == 0))
          --level;

        if (level < 0)
          break;

        ActiveBoxVec& input_boxvec = *input_boxvec_by_level[level];

        if (first_live_box.size() <= level)
          first_live_box.resize(level + 1, 0);

        ActiveBoxVec::Index boxind = first_live_box[level];
        ActiveBoxVec::Index endindex = input_boxvec.endindex();
        
        if (level > max_lev_so_far)
          max_lev_so_far = level;
        while (boxind < endindex && input_boxvec.iskilled(boxind))
          ++boxind;

        if (boxind >= endindex)
          throw_error("Ran off the end of boxvec");
        first_live_box[level] = boxind;

        ActiveBoxVec& small_boxvec = input_boxvec_by_level.force_creation(level+1);
        ActiveBoxVec& idle_boxvec = idle_boxvec_by_level.force_creation(level);
        ActiveBoxVec& small_idle_boxvec = idle_boxvec_by_level.force_creation(level + 1);

        NonConstActiveBox b = input_boxvec.getbox_nonconst(boxind);
                
//pair<bool, Base3> rval =  coord_is_inside_box( b, coord);
//int oldverb = logstr.verbosity();
//if (rval.first) {
//  logstr.change_verbosity(8);
//detect_weight_problem_at_point(sep_stack0, sep_stack1, all_orbits, g, coord, testiw, crt_wkspa, true);
//}
 
       if (b.completely_interior()) {
          separation.completely_interior(b, small_boxvec);
        }
        else {
          separation.not_interior(b, phase, input_boxvec, 
            small_boxvec, idle_boxvec, small_idle_boxvec);
        }
        input_boxvec.killbox(boxind, logstr);

//if (rval.first) {
//  logstr.change_verbosity(oldverb);
//  detect_weight_problem_at_point(sep_stack0, sep_stack1, all_orbits, g, coord, testiw, crt_wkspa, true);
//}



        gui.check_update();
      } 
      
      // Alignment stage face by face.
      
      for (Brep::Face_Spec_Loop_Over_Faces_Of_Dim alignfspec(g,phase);
      alignfspec.notdone();
      ++alignfspec) { 
        if (logstr.verbosity() >= 1) {
          logstr.str() << "\n------ Begin alignment of " << g.face_name(alignfspec) 
            << " = "<< alignfspec << endl;
          
          if (logstr.verbosity() >= 4) {
            logstr.str() << "Debugging orbit of " << alignfspec.fdim() << ':' 
              << alignfspec.faceind() << '\n';
            all_orbits.debug(patchtable, inc_table, dag,alignfspec);
            logstr.str() << "Dumping orbit of " << alignfspec.fdim() << ':' 
              << alignfspec.faceind() << '\n';
            all_orbits.dump(logstr.str(), alignfspec);
            
          }
        }

#ifdef NSF_DEMO
        {
          ostringstream os;
          os << "Alignment of " << g.face_name(alignfspec);
          QMG::Nsf_Demo::indicate_phase_start(os.str());
        }
#endif

        int level = 0;
        ActiveBoxVec* input_boxvecp = all_orbits.find_next_nonempty_level(alignfspec,level);
        while (input_boxvecp) {
          ActiveBoxVec& input_boxvec = *input_boxvecp;
          level = input_boxvec.iwidth();
          if (input_boxvec.numboxes() > 0) {
            if (level > max_lev_so_far) {
              if (phase == di) {
                if (logstr.verbosity() >= 1) {
                  logstr.str() << "Alignment oversplit error.  Example hanging box:\n";
                  ActiveBoxVec::Index boxind = 0;  
                  ActiveBoxVec::Index endindex = input_boxvec.endindex();
        
                  while (boxind < endindex && input_boxvec.iskilled(boxind))
                    ++boxind;
                  
                  if (boxind >= endindex)
                    throw_error("Ran off the end of boxvec");
                  ActiveBox b = input_boxvec.getbox(boxind);
                  b.full_dump(logstr.str());
                }
                throw_error("Alignment oversplit (probably indicates ray intersection failure in QMG)");
              }
              max_lev_so_far = level;
            }
            {
              ostringstream ost;
              ost << "Phase " << phase << " alignment face = " << g.face_name(alignfspec)
                << " level = " << level;
              gui.operation2(ost.str());
            } 
            
            if (level >= MAXLEV)
              throw_error("Alignment oversplit");

            ActiveBoxVec& idle_boxvec = idle_boxvec_by_level.force_creation(level);
            ActiveBoxVec& smallboxvec = all_orbits.retrieve_orbit(alignfspec, level+1);
            ActiveBoxVec& small_idle_boxvec = idle_boxvec_by_level.force_creation(level + 1);
            if (logstr.verbosity() >= 1) {
              logstr.str() << "============ At start of alignment level " << level << " for " <<
                g.face_name(alignfspec) << " input boxvec size = "
                << input_boxvec.numboxes() << " small boxvec size = "
                << smallboxvec.numboxes() << " idle_boxvec size = "
                << idle_boxvec.numboxes() << " small idle boxvec size = "
                << small_idle_boxvec.numboxes() << '\n';
              logstr.str() << " input box vec expa = "
                << input_boxvec.expa() << " smallboxvec expa = "
                << smallboxvec.expa() << " idleboxvec expa = "
                << idle_boxvec.expa() << " small idle boxvec expa = "
                << small_idle_boxvec.expa() << '\n';
            }

            int maxclosefacedim;
            maxclosefacedim = di - phase;
            Real maxwarpdist = expa[phase];
            
            // Loop over possible dimensions for the close face of the box.
            
            for (int closefacedim = 0; closefacedim <= maxclosefacedim; closefacedim++) {
              if (logstr.verbosity() >= 1) {
                logstr.str() << "----------------- Phase " << phase << " alignment face = "
                  << g.face_name(alignfspec) << " level = " << level 
                  << " orbit size = " << input_boxvec.numboxes() << " closefacedim = "
                  << closefacedim << endl;
              }       
              


//detect_weight_problem_at_point(sep_stack0, sep_stack1, all_orbits, g, coord, testiw, crt_wkspa, true);

              if (phase < di) {
                align_orbit(input_boxvec, closefacedim, maxwarpdist, idle_boxvec, smallboxvec,
                  small_idle_boxvec, alignfspec, dag, sc, vtx_owner, asp_degrade, 
                  logstr, gui, patchtable, inc_table, crt_wkspa);
              }
              else {
                align_orbit_lastphase(input_boxvec, smallboxvec, alignfspec, dag, sc,
                  vtx_owner, logstr, gui);
              }
//detect_weight_problem_at_point(sep_stack0, sep_stack1, all_orbits, g, coord, testiw, crt_wkspa, true);

            }            
            
            // Put boxes with no close face in the idle stack.
            
            for (ActiveBoxVec::Loop_over_boxes loop(input_boxvec);
            loop.notdone();
            ++loop) {
              throw_error("boxes remain in orbit after alignment");
              // ActiveBox b = loop.getbox();
              // ActiveBox_As_Parent ab(b);
              // idle_boxvec.push_box_shrink_ex(ab, crt_wkspa);
              // input_boxvec.killbox(loop.index(), logstr);
            }
            
//detect_weight_problem_at_point(sep_stack0, sep_stack1, all_orbits, g, coord, testiw, crt_wkspa, true);

            if (logstr.verbosity() >= 1) {
              logstr.str() << " at level end input box vec size = "
                << input_boxvec.numboxes() << " smallboxvec size = "
                << smallboxvec.numboxes() << " idleboxvec size = "
                << idle_boxvec.numboxes() << " small idle boxvec size = "
                << small_idle_boxvec.numboxes() << '\n';
            }
          }
          all_orbits.erase(alignfspec,level);          
          input_boxvecp = all_orbits.find_next_nonempty_level(alignfspec,level + 1);
        }
        if (logstr.verbosity() >= 2) {
          logstr.str() << "At alignment end idle vec size:";
          idle_boxvec_by_level.dump_size(logstr.str());
        }
      }
    }
  }

  // Triangulate the quadtree.

  gui.operation("Triangulation");
  if (logstr.verbosity() >= 1) {
    logstr.str() << "\n\n------------- Begin triangulation ";
    logstr.logdate();
    logstr.str() << endl;
  }
  int link_size = dag.size();

  // workspaces for add_simplex.
  vector<SimpComplex::VertexOrdinalIndex> volist(di + 1);
  vector<SimpComplex::VertexGlobalIndex> vlist1(di + 1);
  vector<Real> param1(di);
  vector<Point> params(di + 1);
  vector<Brep::PatchIndex> patchinds(di + 1);
  Brep_Parameter_Lifter lifter(g);
  int surf_ori;
  int pmt;

  int ori1 = ori - 1;
  if ((di + 1) % 2)
    ori1 = 1 - ori1;


  for (BoxFaceDag::Index leafind = 0; leafind < link_size; ++leafind) {
    gui.check_update();
    if (dag.is_leaf(leafind)) {
      dag.trace_path_to_root(leafind, volist, pmt, surf_ori);
#ifdef DEBUGGING
      if (volist.size() != di + 1) {
        throw_error("Link stack wrong size.");
      }
#endif
      if (logstr.verbosity() >= 2) {
        logstr.str() << "for leaf bfdi#" << leafind << "# generating simplex ";
        for (int ll = 0; ll < di + 1; ++ll) {
          logstr.str() << "ord#" << volist[ll] << "# ";
        }
        logstr.str() << "\n";
      }
      int ori2 = (pmt < 0) ? (1 - ori1) : ori1;
      add_simplex(sc, volist, surf_ori, ori2, vtx_owner, lifter,
        vlist1, param1, params, patchinds);
    }
  }
  sc.remove_unused_vertices();
  if (logstr.verbosity() >= 1) {
    logstr.str() << "****************************************\nTriangulation Complete ";
    logstr.logdate();
  }
  gui.update();
  return sc;
}
