// ------------------------------------------------------------------
// checktri.cpp
//
// This file contains the routine for checking that a triangulation
// of a brep is correct.
// ------------------------------------------------------------------
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------

#ifdef _MSC_VER
#if _MSC_VER == 1200
#include "qmatvec.h"
#endif
#endif

#include "qbrep.h"
#include "qsimpcomp.h"
#include "qmatvec.h"


// Function exported by this file.

namespace QMG {
  Real checktri(const Brep& brep,
    const SimpComplex& simpcomp,
    ostream& os,
    int checko,
    double tol);
}



namespace {
  using namespace QMG;

  const int CHECKTRI_MAXDIM = 3;
  
  // ------------------------------------------------------------------
  // struct Facet_Rec
  // A record of a facet of a simplex.  Used as a map data.  Keep track
  // of how many times a facet occurs.

  struct Facet_Rec {
    int pos_occurrence_count;
    int neg_occurrence_count;
    int occurs_on_bdry;
    int seqno_occ[2];
    Facet_Rec() : pos_occurrence_count(0),
      neg_occurrence_count(0),
      occurs_on_bdry(0) { }
  };
  
  // ------------------------------------------------------------------
  // struct Facet_Key
  // A record of a facet of a simplex.  Used as a map key. The key is
  // the list of vertices of the facet, sorted into increasing order.


  struct Facet_Key {
    static int facetdim;
    SimpComplex::VertexGlobalIndex vert[CHECKTRI_MAXDIM];
    bool operator<(const Facet_Key& other) const;
    bool operator==(const Facet_Key& other) const;
  };

  bool Facet_Key::operator<(const Facet_Key& other) const {
    for (int j = 0; j < facetdim + 1; ++j) {
      if (vert[j] < other.vert[j])
        return true;
      if (vert[j] > other.vert[j])
        return false;
    }
    return false;
  }

  int Facet_Key::facetdim;

  bool Facet_Key::operator==(const Facet_Key& other) const {
    for (int j = 0; j < facetdim + 1; ++j) {
      if (vert[j] != other.vert[j])
        return false;
    }
    return true;
  }

  class Dumpsimp {
  private:
    const SimpComplex& s;
    const Brep& g;
    Brep::Face_Spec fspec;
    int seqno;
  public:
    Dumpsimp(const SimpComplex& s1, 
      const Brep& g1,       
      const Brep::Face_Spec& f1,
      int seq1) :
    s(s1), g(g1), fspec(f1), seqno(seq1) {}
    friend QMG::ostream& operator<<(QMG::ostream& os, const Dumpsimp& ds);
  };

  QMG::ostream& operator<<(QMG::ostream& os, const Dumpsimp& ds) {
    os << "simplex #" << ds.seqno << " of topological entity "
      << ds.g.face_name(ds.fspec) << " (";
    ::operator<<(os, ds.fspec); 
    os << ") which has vertices ";
    for (int j = 0; j < ds.fspec.fdim() + 1; ++j)
      os << ds.s.node_of_meshface_on_face(ds.fspec, ds.seqno, j) << " ";
    os << '\n';
    return os;
  }
}

  
// ------------------------------------------------------------------
// checktri
// Argument g is a brep and argument s is a putative triangulation
// for g.  This routine checks whether the simplicial complex is
// correct.  Problems detected are printed on outstream.  
// The return value is the worst aspect ratio among all simplices.
// It is negative if there is an error.
// Set checko=1 if every elt should have forward orientation,
// 2 if every elt should have backward orientation, 0 not to check orientation.
// Note: orientation not checked for low-dimensional complexes.
// ------------------------------------------------------------------

QMG::Real 
QMG::checktri(const Brep& g,
              const SimpComplex& s,
              ostream& outstream,
              int checko,
              double tol) {
  
  int di = g.embedded_dim();
  if (di != s.embedded_dim()) {
    throw_error("Brep and simpcomplex have different embedded dims");
  }
  
  // initialize the random number generator.

  random_real(1);
  Point::InitStaticData(di);

  string brep_id1 = g.lookup_brep_propval(Brep::global_id_propname());
  string brep_id2 = s.lookup_prop_val(Brep::global_id_propname());
  bool match = compare_nocase(brep_id1, brep_id2) == 0;
  if (!match)
    throw_error("Brep and simpcomplex have different global id's");
  int sdim = s.gdim();

  if (sdim > CHECKTRI_MAXDIM) {
    throw_error("Dimension out of range for checktri");
  }
  if (sdim > g.gdim()) {
    throw_error("Intrinsic dimension of simp complex too high for brep");
  }
  double scfac = g.scale_factor();
  double scaled_tol = scfac * tol;
  
  Real minaltglobal = BIG_REAL;
  Real maxsideglobal = 0.0;
  Real maxaspglobal = 0.0;
  int worsttriseq = -999;
  Brep::Face_Spec worsttrifspec;
  bool imperfectflag = false;
  
  Matrix simplexmatrix(sdim, di);
  Matrix simplexvertex(sdim + 1, di);

  // hash table on facets; counts occurrences of facets on the
  // positive and negative sides.

  typedef map<Facet_Key, Facet_Rec> Facet_Map_Type;
  Facet_Map_Type facet_occurrence_map;
  map<SimpComplex::VertexGlobalIndex, int> checkoff;
  map<SimpComplex::VertexGlobalIndex, int> onface;
  int numvtx = s.num_nodes();
  int numelt1 = 0;
  vector<int> sortord(sdim + 1);
  Real hhtrans[4];

  for (int fdim = 0; fdim <= sdim; ++fdim) {
    if (g.level_size(fdim) != s.brep_levelsize(fdim))
      throw_error("Mismatch between brep and mesh (different level sizes)");
    Facet_Key::facetdim = fdim - 1;
    sortord.resize(fdim + 1);
    for (Brep::FaceIndex faceind = 0; faceind < g.level_size(fdim); ++faceind) {
      Brep::Face_Spec fspec(fdim, faceind);
      if (fdim < di) {
        int nnf = s.num_node_on_face(fspec);
        if (nnf == 0) {
          outstream << "?? Topological face " << g.face_name(fspec) << " (" << fspec 
            << ") does not have any mesh nodes\n";
          imperfectflag = true;
        }
        onface.clear();
      
        for (int seqno = 0; seqno < nnf; ++seqno) {
          SimpComplex::VertexGlobalIndex vnum = s.node_on_face(fspec,seqno);
          onface[vnum] = 1;
          SimpComplex::VertexOrdinalIndex vnumo = s.global_to_ordinal_nothrow(vnum);
          if (vnumo < 0) {
            outstream << "?? Global node #" << vnum << " listed as lying on face "
              << g.face_name(fspec) << "(" << fspec << ") but no node\n"
              << "  with that number occurs in the coordinate list.\n";
            imperfectflag = true;
            continue;
          }
          Point real1, param1;
          for (int j = 0; j < di; ++j)
            real1[j] = s.real_coord_o(vnumo, j);
          for (int j2 = 0; j2 < fdim; ++j2)
            param1[j2] = s.param_coord_on_face(fspec, seqno,j2);
          Brep::PatchIndex patchind = s.patchind_on_face(fspec, seqno);

          Point real2 = g.patchmath(fspec, patchind).get_real_point(param1);
          Real l2dist = Point::l2dist(real1,real2);
          if (l2dist > scfac * tol) {
            outstream << "?? Vertex " << vnum << " real coordinates too far away "
              << " from actual point on " << g.face_name(fspec) << " (" 
              << fspec << ")\n"
              << "  ( dist = " << l2dist << ")\n";
            imperfectflag = true;
          }
        }
      }
      if (fdim == 0) continue;

      int nmf = s.num_meshface_on_face(fspec);
      facet_occurrence_map.clear();
     
      {
        Facet_Key key;  
        for (int seqno = 0; seqno < nmf; ++seqno) {
          if (fdim < di) {
            for (int jj = 0; jj < fdim + 1; ++jj) {
              SimpComplex::VertexGlobalIndex vtx = 
                s.node_of_meshface_on_face(fspec, seqno,jj);
              if (!onface[vtx]) {
                outstream << "?? Vertex " << vtx << " occurs in " 
                  << Dumpsimp(s,g,fspec,seqno) << " but is\n"
                  << " not listed as a vertex on that topological entity.\n";
                imperfectflag = true;
              }
            }
          }
          bool notfound = false;
          for (int jj = 0; jj < fdim + 1; ++jj) {
            SimpComplex::VertexGlobalIndex vtx = 
              s.node_of_meshface_on_face(fspec, seqno, jj);
            SimpComplex::VertexOrdinalIndex vtxo = 
              s.global_to_ordinal_nothrow(vtx);
            if (vtxo < 0) {
              outstream << "?? Vertex " << vtx << " occurs in " 
                << Dumpsimp(s,g,fspec,seqno) << " but has no coordinates in table.\n";
              imperfectflag = true;
              notfound = true;
            }
            else {
              for (int j = 0; j < di; ++j)
                simplexvertex(jj,j) = s.real_coord_o(vtxo,j);
            }
          }
          if (notfound) continue;

          if (fdim == sdim) {
            ++numelt1;
            for (int jj = 0; jj < fdim + 1; ++jj) {
              SimpComplex::VertexGlobalIndex vtx = 
                s.node_of_meshface_on_face(fspec, seqno,jj);
              checkoff[vtx] = 1;
            }
          }      
          // Make sure the simplex has the correct orientation.
          
          
          if (checko > 0 && fdim == di) {
            for (int j = 1; j < di + 1; ++j) {
              for (int k = 0; k < di; ++k) {
                simplexmatrix(j - 1, k) = simplexvertex(j,k) - simplexvertex(0,k);
              }
            }
            Real det = simplexmatrix.determinant();
            if ((det <= 0 && checko == 1) || (det >= 0 && checko == 2)) {
              outstream << "?? " << Dumpsimp(s,g,fspec,seqno) 
                <<  " orientation incorrect\n";
              imperfectflag = true;
            }
          }
          
          Real minalt = BIG_REAL;

          // Now find the max side length.
          
          Real maxside = 0.0;
          if (sdim == fdim) {
            for (int i = 0; i < sdim; ++i) {
              for (int j = i + 1; j < sdim + 1; ++j) {
                Real side = 0.0;
                for (int d = 0; d < di; ++d) {
                  Real t = simplexvertex(i, d) - simplexvertex(j,d);
                  side += t*t;
                }
                side = sqrt(side);
                if (side > maxside)
                  maxside = side;
              }
            }
          }
          
          for (int fa = 0; fa < fdim + 1; ++fa) {
            int kk = 0;
            SimpComplex::VertexGlobalIndex lastv;
            for (int jj = 0; jj < fdim + 1; ++jj) {
              if (jj != fa) {
                sortord[kk] = jj;
                key.vert[kk] = s.node_of_meshface_on_face(fspec, seqno, jj);
                ++kk;
              }
              else {
                sortord[fdim] = jj;
                lastv = s.node_of_meshface_on_face(fspec,seqno,jj);
              }
            }
            // Do an insertion sort on the key.     
            int swapcount = 0;
            {
              for (int j = 1; j < fdim; j++) {
                int k1 = j;
                while (k1 > 0 && key.vert[k1] < key.vert[k1-1]) {
                  SimpComplex::VertexGlobalIndex tmp = key.vert[k1];
                  key.vert[k1] = key.vert[k1-1];
                  key.vert[k1-1] = tmp;
                  int tmp2 = sortord[k1];
                  sortord[k1] = sortord[k1 - 1];
                  sortord[k1 - 1] = tmp2;
                  --k1;
                  ++swapcount;
                }
              }
            }
            
            
            // Check the altitude if this simplex is full-dimensional.
            
            // Make a matrix with the data for this simplex facet.
            
            if (fdim == sdim) {
              {
                for (int i = 0; i < sdim; i++) {
                  for (int j = 0; j < di; j++) 
                    simplexmatrix(i,j) = 
                    simplexvertex(sortord[i + 1],j) - simplexvertex(sortord[0],j);
                }
              }
              
              // QR factor the matrix to get the altitude & determinant.
              
              Real det = 1.0; 
              {
                for (int i = 0; i < sdim; i++) {
                  Real beta = Matrix::compute_hh_transform(&hhtrans[i], &simplexmatrix(i,i), 
                    1, di - i);
                  for (int j = i; j < sdim; ++j) {
                    Real ip = 0.0;
                    {
                      for (int k = i; k < di; ++k)
                        ip += simplexmatrix(j,k) * hhtrans[k];
                    }
                    ip *= beta;
                    {
                      for (int k = i; k < di; ++k)
                        simplexmatrix(j,k) -= ip * hhtrans[k];
                    }
                  }
                  det *= simplexmatrix(i,i);
                }
              }    
              
              // Compute the altitude to this facet.
              
              Real alt = fabs(simplexmatrix(sdim - 1,sdim - 1));
              if (alt < minalt)
                minalt = alt;
            }
            
            // hash the facet
            
            Facet_Rec table_entry = facet_occurrence_map[key];     
            if ((swapcount + fa) % 2)
              ++(table_entry.pos_occurrence_count);
            else
              ++(table_entry.neg_occurrence_count);
            int tot_occ = table_entry.pos_occurrence_count + 
              table_entry.neg_occurrence_count;
            if (tot_occ < 3) {
              table_entry.seqno_occ[tot_occ - 1] = seqno;
            }
            facet_occurrence_map[key] = table_entry;
          } // end of facet loop

          if (fdim == sdim) {
            if (maxside > maxsideglobal)
              maxsideglobal = maxside;
            if (minalt < minaltglobal)
              minaltglobal = minalt;        
            if (sdim > 0) {
              Real asp = maxside/minalt;
              if (asp > maxaspglobal) {
                maxaspglobal = asp;
                worsttriseq = seqno;
                worsttrifspec = fspec;
              }
            }
          }
          // Done with this element
        }
      }
      // Now go over the facets to make sure that each facet occurs the proper
      // number of times.

      // First, loop over the facets that belong to lower-dimensional
      // topological entities.

      for (Brep::Face_Spec_Loop_Over_Face_Subfaces subfspec(g, fspec);
      subfspec.notdone();
      ++subfspec) {
        if (subfspec.fdim() != fdim - 1) continue;
        int nmf2 = s.num_meshface_on_face(subfspec);
        Facet_Key key;
        for (int seqno2 = 0; seqno2 < nmf2; ++seqno2) {
          for (int jj = 0; jj < fdim; ++jj)
            key.vert[jj] = s.node_of_meshface_on_face(subfspec, seqno2, jj);
          // Insertion sort on key.          
          {
            for (int j = 1; j < fdim; ++j) {
              int k1 = j;
              while (k1 > 0 && key.vert[k1] < key.vert[k1-1]) {
                SimpComplex::VertexGlobalIndex tmp = key.vert[k1];
                key.vert[k1] = key.vert[k1-1];
                key.vert[k1-1] = tmp;
                --k1;
              }
            }
          }
          Facet_Rec table_entry = facet_occurrence_map[key];
          ++table_entry.occurs_on_bdry;
          facet_occurrence_map[key] = table_entry;
        }
      }

      // Next pass: make sure counts for boundaries add up correctly.

      for (Facet_Map_Type::const_iterator it = facet_occurrence_map.begin();
      it != facet_occurrence_map.end();
      ++it) {
        Facet_Key key = it -> first;
        Facet_Rec table_entry = it -> second;
        int pos1 = table_entry.pos_occurrence_count;
        int neg1 = table_entry.neg_occurrence_count;
        int tot1 = pos1 + neg1;
        int bdry1 = table_entry.occurs_on_bdry;
        const char* msg = "";
        if (bdry1 > 0 && tot1 == 0) {
          msg = "mesh entity occur on topological boundary but not as boundary of mesh";
        }
        else if (bdry1 > 0 && ((tot1 + bdry1) % 2))  {
          msg = "mesh entity occurs on topological boundary with wrong parity";
        }
        else if (bdry1 == 0 && tot1 % 2) {
          msg = "internal mesh entity occurs an odd number of times (hole in mesh)";
        }
        if (fdim >= di - 1 && (pos1 > 1 || neg1 > 1)) 
          msg = "mesh entity contained twice on the same side";
        if (strlen(msg) > 0) {
          outstream << "?? Problem with mesh entity bounded by nodes ";
          for (int j = 0; j < fdim; ++j)
            outstream << key.vert[j] << " ";
          outstream << " " << msg 
           << "\n This mesh entity is contained by the following higher dimensional\n"
           << " mesh entities:\n";
          int nume1 = (tot1 > 2)? 2 : tot1;
          for (int j2 = 0; j2 < nume1; ++j2)
            outstream << Dumpsimp(s,g,fspec,table_entry.seqno_occ[j2]) << '\n';
          if (tot1 > 2)
            outstream << " and others.\n";
          imperfectflag = true;
        }
      }
    }
  }

  // Make sure every vertex is checked off.

  {
    for (SimpComplex::VertexOrdinalIndex vnumo = 0; 
    vnumo < numvtx;
    ++vnumo) {
      SimpComplex::VertexGlobalIndex vnum = s.ordinal_to_global(vnumo);
      if (checkoff[vnum] == 0) {
        outstream << "?? Vertex " << vnum << " not used in any simplex\n";
        imperfectflag = true;
      }
    }
  }

  int numelt = s.num_elements();
  if (numelt1 != numelt) {
    outstream << "?? Mismatch: stated number of elements = " << numelt
      << " versus actual number = " << numelt1 << '\n';
    imperfectflag = true;
  }
      

  if (worsttriseq >= 0) {
    outstream << "Maximum aspect ratio =        " << maxaspglobal 
      << " achieved in\n" << Dumpsimp(s,g,worsttrifspec,worsttriseq);
    outstream << "Maximum global side length =  " << maxsideglobal << "\n";
    outstream << "Minimum global altitude =     " << minaltglobal << "\n";
  }
  outstream << "Number of nodes = " << s.num_nodes() 
	    << " number of elements = " << s.num_elements() << '\n';
  if (imperfectflag) {
   outstream << "!! One or more flaws were detected in the mesh\n";
   maxaspglobal = -1.0;
  }
  return maxaspglobal;
}
