// ------------------------------------------------------------------
// gm_list2obj.cpp
//
// This file contains the routine for converting a brep or simpcomplex
// to a Tcl/Tk list.
// ------------------------------------------------------------------
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------


extern "C" {
#include "tcl.h"
}

#include "qbrep_constr.h"
#include "qsimpcomp.h"
#include "qerr.h"
namespace QMG {


  // ------------------------------------------------------------------
  // see objectstream.cpp

  extern void init_faceselect_order(vector<unsigned int>& order, int gdim);
  extern string get_element_typename(int gdim, int ord);
  namespace FrontEnd {

    // See qmg_init.cpp
    extern Tcl_Obj* new_simpcomp_obj(Cptc_MshDesc* mesh);
    extern Tcl_Obj* new_brep_obj(Cptc_GeomDesc* geo);
  }
}


namespace {

  using namespace QMG;

  void throw_tcl_error(Tcl_Interp* interp) {
    throw_error(string(interp -> result));
  }
  // ------------------------------------------------------------------
  // get_tcl_string, get_tcl_int, get_tcl_double, get_tcl_list_length
  // The following local routines are for extracting various data
  // types from a Tcl list.

  string get_tcl_string(Tcl_Interp* interp,
    Tcl_Obj* list,
    int listindex) {

    Tcl_Obj* subobj;
    int returncode = Tcl_ListObjIndex(interp, list, listindex, &subobj);
    if (returncode != TCL_OK)
      throw_tcl_error(interp);
    int length1;
    char* data = Tcl_GetStringFromObj(subobj, &length1);
    return string(data, length1);
  }

  int get_tcl_int(Tcl_Interp* interp,
    Tcl_Obj* list,
    int listindex) {

    Tcl_Obj* subobj;
    int returncode = Tcl_ListObjIndex(interp, list, listindex, &subobj);
    if (returncode != TCL_OK)
      throw_tcl_error(interp);
    int val;
    returncode = Tcl_GetIntFromObj(interp, subobj,  &val);
    if (returncode != TCL_OK)
      throw_tcl_error(interp);
    return val;
  }


  double get_tcl_double(Tcl_Interp* interp,
    Tcl_Obj* list,
    int listindex) {

    Tcl_Obj* subobj;
    int returncode = Tcl_ListObjIndex(interp, list, listindex, &subobj);
    if (returncode != TCL_OK)
      throw_tcl_error(interp);
    double val;
    returncode = Tcl_GetDoubleFromObj(interp, subobj,  &val);
    if (returncode != TCL_OK)
      throw_tcl_error(interp);
    return val;
  }

  Tcl_Obj* get_tcl_obj(Tcl_Interp* interp,
    Tcl_Obj* list,
    int listindex) {

    Tcl_Obj* subobj;
    int returncode = Tcl_ListObjIndex(interp, list, listindex, &subobj);
    if (returncode != TCL_OK)
      throw_tcl_error(interp);
    return subobj;
  }

  int get_tcl_list_length(Tcl_Interp* interp,
    Tcl_Obj* list) {

    int length1;
    int returncode = Tcl_ListObjLength(interp, list, &length1);
    if (returncode != TCL_OK)
      throw_tcl_error(interp);
    return length1;
  }

  // This is a function object used for a map whose key is a string.
  // We want to compare strings disregarding case.

  class Nocase {
  public:
    bool operator() (const string& x, const string& y) const {
      return compare_nocase(x,y) < 0;
    }
  };

  // ------------------------------------------------------------------
  // Class Face_lookup_map_type
  // This is a map from strings to fspecs.
  // It generates errors upon duplication, and so on.  It is used
  // for keeping track of face names when a brep is read in.



  class Face_lookup_map_type {
  private:
    typedef map <string, Brep::Face_Spec, Nocase> Actual_Map_type;
    Actual_Map_type actual_map_;
  public:
    Face_lookup_map_type() { }
    ~Face_lookup_map_type() { }
    void insert(const string& facename, const Brep::Face_Spec& fspec) {
      const char* tmp = facename.c_str();
      string tmp2 = string(tmp);
      pair<Actual_Map_type::iterator, bool> rval = 
        actual_map_.insert(pair<string,Brep::Face_Spec>(tmp2, fspec));

      if (!rval.second) {
        string errmsg = string("Duplicate face name ") + facename
          + " occurs in brep";
        throw_error(errmsg);
      }
      return;
    }

    Brep::Face_Spec find(const string& facename) const {
      Actual_Map_type::const_iterator it =
        actual_map_.find(facename);
      if (it == actual_map_.end()) {
        string errmsg = string("Face ") + facename  + 
          " listed as a subface but is not found";
        throw_error(errmsg);
      }
      return it -> second;
    }
  };

  // ------------------------------------------------------------------
  // This routine converts a list to a brep.

  Brep_Under_Construction get_brep_from_list(Tcl_Interp* interp,
    Tcl_Obj* olist) {

    Error_Message errorinfo;
    errorinfo.set_string("Error detected while processing header information");

    
    int di = get_tcl_int(interp, olist, 2);
    if (di < 2 || di > 3)
      throw_error("Embedded dimension out of range");
    int gdim = get_tcl_int(interp, olist, 1);
    if (gdim > di || gdim < 0)
      throw_error("Intrinsic dimension out of range");
    int listlen = get_tcl_list_length(interp, olist);
    if (listlen != 6 + gdim) 
      throw_error("Wrong number of entries in list (should be 6+intrinsic dim)");


    Brep_Under_Construction g(gdim, di);

    // convert brep property values
    {    
      errorinfo.set_string("Error detected while processing brep's property-value pairs");

      Tcl_Obj* pvlist = get_tcl_obj(interp, olist, 3);
      int pvlen = get_tcl_list_length(interp, pvlist);
      if (pvlen % 2)
        throw_error("Property-value list must have an even number of entries");
      Brep_Under_Construction::Propval_inserter pvi(g);
      for (int i = 0; i < pvlen; i += 2) {
        string prop = get_tcl_string(interp, pvlist, i);
        string val = get_tcl_string(interp, pvlist, i+1);
        pvi.insert_propval(prop,val);
      }
    }

    // convert control points

    {
      errorinfo.set_string("Error detected while processing brep's control points");

      Tcl_Obj* cplist = get_tcl_obj(interp, olist, 4);
      int cplen = get_tcl_list_length(interp, cplist);
      if (cplen % di)
        throw_error("Control point list length must be divisible by embedded dim");
      Brep_Under_Construction::Control_point_inserter cpi(g);
      vector<Real> coord(di);
      for (int i = 0; i < cplen; i += di) {
        for (int j = 0; j < di; ++j)
          coord[j] = get_tcl_double(interp, cplist, i + j);
        cpi.insert_control_point(coord);
      }
    }

    // convert faces


    Face_lookup_map_type facemap;

    for (int dim = 0; dim <= gdim; ++dim) {
      {
        ostringstream os;
        os << "Error detected while processing topological faces of dimension " << dim;
        errorinfo.set_string(os.str());
      }
      Tcl_Obj* dimlist = get_tcl_obj(interp, olist, 5 + dim);
      int dimlen = get_tcl_list_length(interp, dimlist);
      if (dimlen % 5)
        throw_error("List for topological faces must be divisible by 5");
      Brep_Under_Construction::Top_face_inserter tfi(g, dim);
      for (int faceind = 0; faceind < dimlen / 5; ++faceind) {

        // convert face name

        string facename = get_tcl_string(interp, dimlist, faceind * 5);
        Brep::Face_Spec fspec = tfi.insert_top_face(facename);
        facemap.insert(facename, fspec);
        Error_Message error_info2;
        {
          ostringstream os;
          os << "Error detected while processing property-value pairs of face " << facename;
          error_info2.set_string(os.str());
          Tcl_Obj* pvlist = get_tcl_obj(interp, dimlist, faceind * 5 + 1);
          int pvlen = get_tcl_list_length(interp, pvlist);
          if (pvlen % 2)
            throw_error("Property-value list length must be even");
          Brep_Under_Construction::Top_face_propval_inserter tfpi(g, fspec);
          for (int i = 0; i < pvlen; i += 2) {
            string prop = get_tcl_string(interp, pvlist, i);
            string val = get_tcl_string(interp, pvlist, i + 1);
            tfpi.insert_propval(prop,val);
          }
        }

        // convert child faces

        {
          ostringstream os;
          os << "Error detected while processing bounding subfaces of face " << facename;
          error_info2.set_string(os.str());
          Tcl_Obj* chlist = get_tcl_obj(interp, dimlist, faceind * 5 + 2);
          int chlen = get_tcl_list_length(interp, chlist);
          Brep_Under_Construction::Top_face_bound_inserter tfbi(g, fspec);
          for (int i = 0; i < chlen; ++i) {
            string subfacename = get_tcl_string(interp, chlist, i);
            int ori = 2;
            if (subfacename[0] == '+') {
              ori = 0;
              subfacename = subfacename.substr(1, subfacename.length() - 1);
            }
            else if (subfacename[0] == '-') {
              ori = 1;
              subfacename = subfacename.substr(1, subfacename.length() - 1);
            }
        
            Brep::Face_Spec subfspec = facemap.find(subfacename);
            tfbi.insert_boundary(subfspec, ori);
          }
        }
        // Convert internal boundaries.

        {
          ostringstream os;
          os << "Error detected while processing internal boundary subfaces of face " << facename;
          error_info2.set_string(os.str());
          Tcl_Obj* iblist = get_tcl_obj(interp, dimlist, faceind * 5 + 3);
          int iblen = get_tcl_list_length(interp, iblist);
          Brep_Under_Construction::Top_face_ib_inserter tfii(g, fspec);
          for (int i = 0; i < iblen; ++i) {
            string subfacename = get_tcl_string(interp, iblist, i);
            Brep::Face_Spec subfspec = facemap.find(subfacename);
            tfii.insert_ib(subfspec);
          }
        }

        // convert patches.

        {
          ostringstream os;
          os << "Error detected while processing patches of face " << facename;
          error_info2.set_string(os.str());
          vector<Brep::ControlPointIndex> control_points;
          Tcl_Obj* patchlist = get_tcl_obj(interp, dimlist, faceind * 5 + 4);
          int numpatch = get_tcl_list_length(interp, patchlist);
          Brep_Under_Construction::Top_face_patch_inserter tfpi(g, fspec);
          for (int i = 0; i < numpatch; ++i) {
            Tcl_Obj* thispatch = get_tcl_obj(interp, patchlist, i);
            int plen = get_tcl_list_length(interp, thispatch);
            if (plen < 2)
              throw_error("Patch specification must have at least two entries");
            string ptype = get_tcl_string(interp, thispatch, 0);
            PatchType ptype1;
            int degree1;
            int degree2;
            if (dim == 0) {
              if (compare_nocase(ptype, "vertex") != 0)
                throw_error(string("For vertices, patch type must be 'vertex' not '")
                + ptype + "'");
              if (plen != 2)
                throw_error("Vertex patch item must have format (vertex <cpind>)");
              control_points.resize(1);
              control_points[0] = get_tcl_int(interp, thispatch, 1);
              degree1 = 1;
            }
            else if (dim == 1) {
              if (compare_nocase(ptype, "bezier_curve") != 0)
                throw_error(string("For topological edges, patch type must be"
                " 'bezier_curve' not '") + ptype + "'");
              degree1 = get_tcl_int(interp, thispatch, 1);
              int ncp = degree1 + 1;
              if (plen != ncp + 2)                  
                throw_error("Wrong number of control points in bezier curve");
              control_points.resize(ncp);
              for (int j = 0; j < ncp; ++j) 
                control_points[j] = get_tcl_int(interp, thispatch, 2 + j);
            }
            else if (dim == 2 && di == 3) {
              if (compare_nocase(ptype, "bezier_triangle") == 0) {
                ptype1 = BEZIER_TRIANGLE;
                degree1 = get_tcl_int(interp, thispatch, 1);
                int ncp = (degree1 + 1) * (degree1 + 2) / 2;
                if (plen != 2 + ncp)
                  throw_error("Wrong number of control points in bezier triangle patch");
                control_points.resize(ncp);
                for (int j = 0; j < ncp; ++j) 
                  control_points[j] = get_tcl_int(interp, thispatch, 2 + j);
              }
              else if (compare_nocase(ptype, "bezier_quad") == 0) {
                ptype1 = BEZIER_QUAD;
                degree1 = get_tcl_int(interp, thispatch, 1);
                degree2 = get_tcl_int(interp, thispatch, 2);
                int ncp = (degree1 + 1) * (degree2 + 1);
                if (plen != 3 + ncp)
                  throw_error("Wrong number of control points in bezier quadrilateral patch");
                control_points.resize(ncp);
                for (int j = 0; j < ncp; ++j)
                  control_points[j] = get_tcl_int(interp, thispatch, 3 + j);
              }
              else  
                throw_error(string("For topological surfaces, patch type must be"
                "'bezier_triangle' or 'bezier_quad' not '") + ptype + "'");
            }
            tfpi.insert_patch(degree1, degree2, ptype1, control_points);
          }
        }
      }
    }
    return g;
  }
  // ------------------------------------------------------------------
  // This routine converts a simplicial complex to a list.

  SimpComplex_Under_Construction get_simpcomp_from_list(Tcl_Interp* interp,
    Tcl_Obj* olist) {
  
    Error_Message errorinfo;
    errorinfo.set_string("Error detected while processing header information");

    int listlen = get_tcl_list_length(interp, olist);
    if (listlen < 5)
      throw_error("List is too short for a simplicial complex");
    int di = get_tcl_int(interp, olist, 2);
    if (di < 2 || di > 3)
      throw_error("Embedded dimension out of range");
    int gdim = get_tcl_int(interp, olist, 1);
    if (gdim > di || gdim < 1)
      throw_error("Intrinsic dimension out of range");
    if (listlen != 6 + gdim)
      throw_error("Wrong length in list for simplicial complex");
    vector<int> levsize(gdim + 1, 0);
    SimpComplex_Under_Construction sc(gdim, di, levsize);
    // Get propval pairs.

    {
      Tcl_Obj* pvlist = get_tcl_obj(interp, olist, 3);
      int numpv = get_tcl_list_length(interp, pvlist);
      if (numpv % 2)
        throw_error("Property value list must have an even number of entries");
      numpv /= 2;
      for (int j = 0; j < numpv; ++j) {
        string prop = get_tcl_string(interp, pvlist, 2 * j);
        string val = get_tcl_string(interp, pvlist, 2 * j + 1);
        sc.add_prop_val(prop,val);
      }
    }

    // Convert list of nodes.

    {
      Tcl_Obj* nodelist = get_tcl_obj(interp, olist, 4);
      int numnode = get_tcl_list_length(interp, nodelist);
      if (numnode % (1 + di))
        throw_error("Length of list of nodes must be divisible by 1+emb.dim");
      numnode /= 1 + di;
      vector<Real> coord(di);
      errorinfo.set_string("Error detected while processing node list");
      for (SimpComplex::VertexOrdinalIndex vnumo = 0; vnumo < numnode; ++vnumo) {
        SimpComplex::VertexGlobalIndex vnum = get_tcl_int(interp, nodelist, (di + 1) * vnumo);
        for (int j = 0; j < di; ++j)
          coord[j] = get_tcl_double(interp, nodelist, (di + 1) * vnumo + j + 1);
        sc.add_vertex(coord, vnum);
      }
    }

    // Convert list of elements.
    {
      vector<Real> param_coord;
      vector<SimpComplex::VertexGlobalIndex> vlist;
      for (int fdim = 0; fdim <= gdim; ++fdim) {
        param_coord.resize(fdim);
        vlist.resize(fdim + 1);
        Tcl_Obj* levlist = get_tcl_obj(interp, olist, 5 + fdim);
        int numface = get_tcl_list_length(interp, levlist);
        errorinfo.set_string("Error occurred while parsing level list");
        if (numface % 2)
          throw_error("Must have an even number of lists associated with a brep level");
        numface /= 2;
        for (int faceind = 0; faceind < numface; ++faceind) {
          Brep::Face_Spec fspec(fdim, faceind);
          sc.add_fspec(fspec);
          {
            ostringstream os;
            os << "Error occurred while processing nodes lying on top. entity " << fspec;
            errorinfo.set_string(os.str());
          }
          Tcl_Obj* nodeinc = get_tcl_obj(interp, levlist, 2 * faceind);
          int nlsize = (fdim == 0)? 1 : 2 + fdim;
          int numnode = get_tcl_list_length(interp, nodeinc);
          if (numnode % nlsize)
            throw_error("Node incidence list wrong size");
          numnode /= nlsize;
          for (int j = 0; j < numnode; ++j) {
            SimpComplex::VertexGlobalIndex vnum = get_tcl_int(interp, nodeinc, nlsize * j);
            Brep::PatchIndex patchind = (fdim > 0)? get_tcl_int(interp, nodeinc, nlsize * j + 1)
              : 0;
            for (int k = 0; k < fdim; ++k)
              param_coord[k] = get_tcl_double(interp, nodeinc, nlsize * j + 2 + k);
            sc.add_vertex_bdryinc(vnum, fspec, patchind, param_coord);
          }
          Tcl_Obj* meshface = get_tcl_obj(interp, levlist, 2 * faceind + 1);
          int numface = get_tcl_list_length(interp, meshface);
          if (numface % (fdim + 1))
            throw_error("Mesh incidence list wrong length");
          numface /= (fdim + 1);
          for (int l = 0; l < numface; ++l) {
            for (int k = 0; k < fdim + 1; ++k)
              vlist[k] = get_tcl_int(interp, meshface, (fdim + 1) * l + k);
            sc.add_simplex_face(vlist, fspec);
          }
        }
      }
    }
    return sc;
  }
}

              

// This is the wrapper.  It's not the standard wrapper because it uses Tcl-specific
// constructs (not available in Matlab).


extern "C" {
  int gm_list2obj(ClientData clientdata,
                    Tcl_Interp* interp,
                    int objc,
                    Tcl_Obj* const objv[]) {

    using namespace QMG;
    
    
    Error_Message::init_error_message_list();
    Error_Task::init_error_task_list();
    if (objc != 2) {
      Tcl_WrongNumArgs(interp, 1, objv, "list");
      return TCL_ERROR;
    }
    
    int listlen;
    int returncode = Tcl_ListObjLength(interp, objv[1], &listlen);
    if (returncode != TCL_OK || listlen < 5) {
      interp -> result = "First argument must be a list with at least 5 entries";
      return returncode;
    }
    
    Tcl_Obj* subobj;
    returncode = Tcl_ListObjIndex(interp, objv[1], 0, &subobj);
    if (returncode != TCL_OK) {
      return returncode;
    }
    
    
    int strlen1;
    char* typecode0 = Tcl_GetStringFromObj(subobj, &strlen1);
    string typecode(typecode0, strlen1);
    if (compare_nocase(typecode, string(Brep::io_header_code())) == 0) {
      try {
        Brep_Under_Construction brep = get_brep_from_list(interp, objv[1]);
        Cptc_GeomDesc* geo = brep.release();
        Tcl_Obj* newobj = QMG::FrontEnd::new_brep_obj(geo);
        Tcl_SetObjResult(interp, newobj);
        return TCL_OK;
      }
      catch (QMG::Error e) {
        Tcl_AppendResult(interp, e.get_error_messages(), 0);
        Tcl_AppendResult(interp, "gm_list2obj terminated because of exception.", 0);
        return TCL_ERROR;
      }
      catch (...) {
        Tcl_AppendResult(interp, "gm_list2obj terminated because of unknown exception.", 0);
        return TCL_ERROR;
      }   
    }
    else if (compare_nocase(typecode, string(SimpComplex::io_header_code())) == 0) {
      try {
        SimpComplex_Under_Construction sc = get_simpcomp_from_list(interp, objv[1]);
        Cptc_MshDesc* msh = sc.release();
        Tcl_Obj* newobj = QMG::FrontEnd::new_simpcomp_obj(msh);
        Tcl_SetObjResult(interp, newobj);
        return TCL_OK;
      }
      catch (QMG::Error e) {
        Tcl_AppendResult(interp, e.get_error_messages(), 0);
        Tcl_AppendResult(interp, "gm_list2obj terminated because of exception.", 0);
        return TCL_ERROR;
      }
      catch (...) {
        Tcl_AppendResult(interp, "gm_list2obj terminated because of unknown exception.", 0);
        return TCL_ERROR;
      }   
    }
    else {
      interp -> result = "First item in list does not match brep or simpcomplex type code";
      return TCL_ERROR;
    }
  }
}





