// ------------------------------------------------------------------
// qsimpcomp.cpp
//
// This file contains member functions of classes SimpComp and 
// SimpComp_Under_Construction.  See simpcomp.h for more information.
// ------------------------------------------------------------------
// Author: Stephen A. Vavasis
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------

#include "qsimpcomp.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
  void cptc_free_mesh(Cptc_MshDesc* mesh);
}



namespace {
  using namespace QMG;
  void* malloc1(size_t sz) {
    char tmp[100];
    void* p = malloc(sz);
    if (p == 0 && sz > 0) {
      sprintf(tmp, "Out of memory on malloc of size %u", sz);
      throw_error(tmp);
    }
    return p;
  }
}



#ifdef _MSC_VER
static void unusedfunc() {
  using namespace QMG;
  map<int, int> q;
  q.insert(pair<int,int>(2,3));
  if (q.begin() != q.end())
    throw_error("");
}
#endif


int 
QMG::SimpComplex::brep_levelsize(int level) const {
#ifdef RANGECHECK
  if (level < 0 || level > mesh_ -> embedded_dimension)
    throw_error("levelsize arg out of range");
#endif
  if (level == 0)
    return mesh_ -> num_top_vertices;
  else if (level == 1)
    return mesh_ -> num_top_edges;
  else if (level == 2)
    return mesh_ -> num_top_surfaces;
  else
    return mesh_ -> num_top_regions;
}

QMG::string
QMG::SimpComplex::lookup_prop_val(const string& prop) const {
  for (int i = 0; i < mesh_ -> num_prop_val; ++i) {
    if (compare_nocase(prop, string(mesh_ -> props[i])) == 0)
      return string(mesh_ -> vals[i]);
  }
  return string("");
}


QMG::Real 
QMG::SimpComplex::real_coord_o(VertexOrdinalIndex vnumo, int jj) const {
#ifdef RANGECHECK
  if (jj < 0 || jj >= mesh_ -> embedded_dimension)
    throw_error("Range check");
  return mesh_ -> nodes[process_ordinal_vnum_(vnumo)].coord[jj];
#else
  return mesh_ -> nodes[vnumo].coord[jj];
#endif
}

void 
QMG::SimpComplex::check_fspec_(const Brep::Face_Spec& fspec) const {
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();
  if (fdim < 0 || fdim > mesh_ -> intrinsic_dimension || 
    fdim >= mesh_ -> embedded_dimension)
    throw_error("Fspec dim out of range");
  if (faceind < 0)
    throw_error("Face ind out of range");
  if (fdim == 0) {
    if (faceind >= mesh_ -> num_top_vertices)
      throw_error("Face ind out of range");
  }
  else if (fdim == 1) {
    if (faceind >= mesh_ -> num_top_edges)
      throw_error("Face ind out of range");
  }
  else {
    if (faceind >= mesh_ -> num_top_surfaces)
      throw_error("Face ind out of range");
  }
}

void
QMG::SimpComplex::check_seqno_(const Brep::Face_Spec& fspec, int seqno) const {
  if (seqno < 0)
    throw_error("seqno out of range");
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();
  if (fdim == 0) {
    if (seqno >= static_cast<int>(mesh_ -> vtxinfo[faceind].vtx >= 0))
      throw_error("seqno out of range");
  } 
  else if (fdim == 1) {
    if (seqno >= mesh_ -> edgeinfo[faceind].num_mnode)
      throw_error("seqno out of range");
  } 
  else { 
    if (seqno >= mesh_ -> surfinfo[faceind].num_mnode)
      throw_error("seqno out of range");
  } 
}


int 
QMG::SimpComplex::num_node_on_face(const Brep::Face_Spec& fspec) const {
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();
#ifdef RANGECHECK
  check_fspec2_(fspec);
#endif
  if (fdim == mesh_ -> embedded_dimension)
    return 0;
  else if (fdim == 0)
    return static_cast<int>(mesh_ -> vtxinfo[faceind].vtx >= 0);
  else if (fdim == 1)
    return mesh_ -> edgeinfo[faceind].num_mnode;
  else if (fdim == 2)
    return mesh_ -> surfinfo[faceind].num_mnode;
  return 0;
}

QMG::SimpComplex::VertexGlobalIndex 
QMG::SimpComplex::node_on_face(const Brep::Face_Spec& fspec, int seqno) const {
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();
#ifdef RANGECHECK
  check_fspec_(fspec);
  check_seqno_(fspec, seqno);
#endif
  if (fdim == 0) {
    return mesh_ -> vtxinfo[faceind].vtx;
  }
  else if (fdim == 1) {
    return mesh_ -> edgeinfo[faceind].incs[seqno].id;
  }
  else {
    return mesh_ -> surfinfo[faceind].incs[seqno].id;
  }
}


QMG::Brep::PatchIndex 
QMG::SimpComplex::patchind_on_face(const Brep::Face_Spec& fspec, int seqno) const {
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();
#ifdef RANGECHECK
  check_fspec_(fspec);
  check_seqno_(fspec, seqno);
#endif
  if (fdim == 0) {
    return 0;
  }
  else if (fdim == 1) {
    return mesh_ -> edgeinfo[faceind].incs[seqno].curve_index;
  }
  else {
    return mesh_ -> surfinfo[faceind].incs[seqno].patch_index;
  }
}

QMG::Real 
QMG::SimpComplex::param_coord_on_face(const Brep::Face_Spec& fspec, int seqno, int jj) const {
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();
#ifdef RANGECHECK
  check_fspec_(fspec);
  check_seqno_(fspec, seqno);
  if (jj < 0 || jj >= fdim)
    throw_error("param coord index out of range");
#endif
  if (fdim == 1) {
    return mesh_ -> edgeinfo[faceind].incs[seqno].param;
  }
  else {
    return mesh_ -> surfinfo[faceind].incs[seqno].param[jj];
  }
}

void 
QMG::SimpComplex::check_fspec2_(const Brep::Face_Spec& fspec) const {
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();
  if (fdim < 0 || fdim > mesh_ -> intrinsic_dimension)
    throw_error("Fspec dim out of range");
  if (faceind < 0)
    throw_error("Face ind out of range");
  if (fdim == 0) {
    if (faceind >= mesh_ -> num_top_vertices)
      throw_error("Face ind out of range");
  }
  else if (fdim == 1) {
    if (faceind >= mesh_ -> num_top_edges)
      throw_error("Face ind out of range");
  }
  else if (fdim == 2) {
    if (faceind >= mesh_ -> num_top_surfaces)
      throw_error("Face ind out of range");
  }
  else {
    if (faceind >= mesh_ -> num_top_regions)
      throw_error("Face ind out of range");
  }
}



int 
QMG::SimpComplex::num_meshface_on_face(const Brep::Face_Spec& fspec) const {
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();
#ifdef RANGECHECK
  check_fspec2_(fspec);
#endif
  if (fdim == 0)    
    return static_cast<int>(mesh_ -> vtxinfo[faceind].vtx >= 0);
  else if (fdim == 1)
    return mesh_ -> edgeinfo[faceind].num_medge;
  else if (fdim == 2)
    return mesh_ -> surfinfo[faceind].num_mtrifacet;
  else
    return mesh_ -> regioninfo[faceind].num_tet;
}

QMG::SimpComplex::VertexGlobalIndex 
QMG::SimpComplex::node_of_meshface_on_face(const Brep::Face_Spec& fspec, 
    int seqno, 
    int jj) const {
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();
#ifdef RANGECHECK
  int nm1 = num_meshface_on_face(fspec);
  if (seqno < 0 || seqno >= nm1)
    throw_error("seqno out of range");
  if (jj < 0 || jj > fspec.fdim())
    throw_error("Vertex selector out of range");
#endif
  if (fdim == 0)
    return mesh_ -> vtxinfo[faceind].vtx;
  else if (fdim == 1) 
    return mesh_ -> edgeinfo[faceind].edgelist[seqno][jj];
  else if (fdim == 2)
    return mesh_ -> surfinfo[faceind].trifacetlist[seqno][jj];
  else {
    return mesh_ -> regioninfo[faceind].tets[seqno][jj];
   // map<int,int> q;
    //q.insert(pair<int,int>(2,3));
    //return q.begin() == q.end();
  }
}




void
QMG::SimpComplex_From_FrontEnd::make_global_to_ordinal_() const {
  if (global_to_ordinal_) throw_error("Make global to ordinal called twice");
  global_to_ordinal_ = new GlobalToOrdinalMapType;
  for (VertexOrdinalIndex vnumo = 0;
  vnumo < mesh_ -> num_nodes;
  ++vnumo) {
    VertexGlobalIndex vnum = mesh_ -> nodes[vnumo].id;
    if (vnum < 0)
      throw_error("global vertex id out of range");
    pair<GlobalToOrdinalMapType::iterator, bool> rval =
      global_to_ordinal_ -> 
      insert(pair<VertexGlobalIndex, VertexOrdinalIndex>(vnum,vnumo));
    if (!rval.second)
      throw_error("global vertex id occurs twice in input");
  }
}


QMG::SimpComplex_From_FrontEnd::~SimpComplex_From_FrontEnd() {
  delete global_to_ordinal_;
}



QMG::SimpComplex_From_FrontEnd::
SimpComplex_From_FrontEnd(SimpComplex_From_FrontEnd& other) :
SimpComplex(other.mesh_, other.global_to_ordinal_) {
  other.global_to_ordinal_ = 0;
}


QMG::SimpComplex_From_FrontEnd::
SimpComplex_From_FrontEnd(const SimpComplex_From_FrontEnd_Returnval& sucr) :
SimpComplex(sucr.sref_.mesh_, sucr.sref_.global_to_ordinal_) {
  sucr.sref_.global_to_ordinal_ = 0;
}




bool 
QMG::SimpComplex_Under_Construction::Face_Bdry_Inc_::
operator<(const Face_Bdry_Inc_& other) const {
  if (meshv[0] < other.meshv[0]) return true;
  if (meshv[0] > other.meshv[0]) return false;
  if (meshv[1] < other.meshv[1]) return true;
  if (meshv[1] > other.meshv[1]) return false;
  return meshv[2] < other.meshv[2];
}

bool
QMG::SimpComplex_Under_Construction::Vertex_Bdry_Inc_::
operator<(const Vertex_Bdry_Inc_& other) const {
  if (vnum < other.vnum) return true;
  if (vnum > other.vnum) return false;
  return fspec < other.fspec;
}



void 
QMG::SimpComplex_Under_Construction::clobber_() {
  mesh_ = 0;
  global_to_ordinal_ = 0;
  vertinclookup_ = 0;  
  faceinclookup_ = 0;
  max_node_edge_.resize(0);
  max_edge_edge_.resize(0);
  max_node_surf_.resize(0);
  max_tri_surf_.resize(0);
  max_tet_region_.resize(0);
}



QMG::SimpComplex::VertexOrdinalIndex
QMG::SimpComplex_Under_Construction::add_vertex(const vector<Real>& realcoord, 
 VertexGlobalIndex thisid) {
  VertexOrdinalIndex nn = mesh_ -> num_nodes;
  if (nn >= max_nodes_) {
    int newsize = 2 * max_nodes_ + 1;
    Cptc_Mnode* new_nodes = 
      static_cast<Cptc_Mnode*>(malloc1(sizeof(Cptc_Mnode) * newsize));
    for (int jj = 0; jj < nn; ++jj)
      new_nodes[jj] = mesh_ -> nodes[jj];
    free(mesh_ -> nodes);
    mesh_ -> nodes = new_nodes;
    max_nodes_ = newsize;
  }
  mesh_ -> nodes[nn].id = thisid;
#ifdef RANGECHECK
  if (realcoord.size() != mesh_ -> embedded_dimension)
    throw_error("realcoord wrong size in add_vertex");
#endif
  for (int j = 0; j < mesh_ -> embedded_dimension; ++j)
    mesh_ -> nodes[nn].coord[j] = realcoord[j];
  ++(mesh_ -> num_nodes);
  pair<GlobalToOrdinalMapType::iterator, bool> rval =
    global_to_ordinal_ -> 
    insert(pair<VertexGlobalIndex, VertexOrdinalIndex>(thisid,nn));
  if (!rval.second)
    throw_error("global vertex id inserted twice");
  return nn;
}


#ifndef BUG_IN_USING
using QMG::pair;
#endif

pair<QMG::SimpComplex::VertexOrdinalIndex, QMG::SimpComplex::VertexGlobalIndex> 
QMG::SimpComplex_Under_Construction::
add_vertex(const vector<Real>& realcoord) {
  VertexGlobalIndex thisid = next_vnum_;
  VertexOrdinalIndex vnum_o = add_vertex(realcoord, thisid);
  ++next_vnum_;
  return pair<VertexOrdinalIndex,VertexGlobalIndex>(vnum_o, thisid);
}



void
QMG::SimpComplex_Under_Construction::
add_fspec(const Brep::Face_Spec& fspec) {
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();
  if (fdim < 0 || fdim > mesh_ -> intrinsic_dimension)
    throw_error("Face dimension out of range");
  if (faceind != this -> brep_levelsize(fdim))
    throw_error("Attempt to add faces out of order");
  if (fdim == 0) {
    if (faceind >= max_lev_size_[0]) {
      int newsize = faceind * 2 + 2;
      Cptc_MeshTopVertexInfo* newinfo = 
        static_cast<Cptc_MeshTopVertexInfo*>(malloc(sizeof(Cptc_MeshTopVertexInfo) * newsize));
      for (int j = 0; j < faceind; ++j)
        newinfo[j] = mesh_ -> vtxinfo[j];
      max_lev_size_[0] = newsize;
      free(mesh_ -> vtxinfo);
      mesh_ -> vtxinfo = newinfo;
    }
    ++(mesh_ -> num_top_vertices);
    mesh_ -> vtxinfo[faceind].vtx = -1;
  }
  else if (fdim == 1) {
    if (faceind >= max_lev_size_[1]) {
      int newsize = faceind * 2 + 2;
      Cptc_MeshTopEdgeInfo* newinfo = 
        static_cast<Cptc_MeshTopEdgeInfo*>(malloc(sizeof(Cptc_MeshTopEdgeInfo) * newsize));
      for (int j = 0; j < faceind; ++j)
        newinfo[j] = mesh_ -> edgeinfo[j];
      max_lev_size_[1] = newsize;
      free(mesh_ -> edgeinfo);
      mesh_ -> edgeinfo = newinfo;
    }
    max_node_edge_.resize(faceind + 1, 1);
    max_edge_edge_.resize(faceind + 1, 1);
    mesh_ -> edgeinfo[faceind].num_mnode = 0;
    mesh_ -> edgeinfo[faceind].num_medge = 0;
    mesh_ -> edgeinfo[faceind].incs =
        static_cast<Cptc_NodeCurveInc*>(malloc1(sizeof(Cptc_NodeCurveInc)));
    typedef VertexGlobalIndex Twoint[2];
    mesh_ -> edgeinfo[faceind].edgelist = 
      static_cast<Twoint*>(malloc1(sizeof(Twoint)));
    ++(mesh_ -> num_top_edges);
  }
  else if (fdim == 2) {
    if (faceind >= max_lev_size_[2]) {
      int newsize = faceind * 2 + 2;
      Cptc_MeshTopSurfaceInfo* newinfo = 
        static_cast<Cptc_MeshTopSurfaceInfo*>(malloc(sizeof(Cptc_MeshTopSurfaceInfo) * newsize));
      for (int j = 0; j < faceind; ++j)
        newinfo[j] = mesh_ -> surfinfo[j];
      max_lev_size_[2] = newsize;
      free(mesh_ -> surfinfo);
      mesh_ -> surfinfo = newinfo;
    }
    max_node_surf_.resize(faceind + 1, 1);
    max_tri_surf_.resize(faceind + 1, 1);
    mesh_ -> surfinfo[faceind].incs =
      static_cast<Cptc_NodePatchInc*>(malloc1(sizeof(Cptc_NodePatchInc)));
    mesh_ -> surfinfo[faceind].num_mnode = 0;
    typedef VertexGlobalIndex Threeint[3];
    mesh_ -> surfinfo[faceind].num_mtrifacet = 0;
    mesh_ -> surfinfo[faceind].trifacetlist = 
      static_cast<Threeint*>(malloc1(sizeof(Threeint)));
    mesh_ -> surfinfo[faceind].num_mquadfacet = 0;
    mesh_ -> surfinfo[faceind].quadfacetlist = 0;
    ++(mesh_ -> num_top_surfaces);
  }
  else if (fdim == 3) {
    if (faceind >= max_lev_size_[3]) {
      int newsize = faceind * 2 + 2;
      Cptc_MeshTopRegionInfo* newinfo = 
        static_cast<Cptc_MeshTopRegionInfo*>(malloc(sizeof(Cptc_MeshTopRegionInfo) * newsize));
      for (int j = 0; j < faceind; ++j)
        newinfo[j] = mesh_ -> regioninfo[j];
      max_lev_size_[3] = newsize;
      free(mesh_ -> regioninfo);
      mesh_ -> regioninfo = newinfo;
    }    
    max_tet_region_.resize(faceind + 1, 1);
    typedef VertexGlobalIndex Fourint[4];
    mesh_ -> regioninfo[faceind].tets = 
      static_cast<Fourint*>(malloc1(sizeof(Fourint)));
    mesh_ -> regioninfo[faceind].num_tet = 0;
    mesh_ -> regioninfo[faceind].num_prism = 0;
    mesh_ -> regioninfo[faceind].num_pyramid = 0;
    mesh_ -> regioninfo[faceind].num_hex = 0;
    mesh_ -> regioninfo[faceind].prisms = 0;
    mesh_ -> regioninfo[faceind].pyramids = 0;
    mesh_ -> regioninfo[faceind].hexs = 0;
    ++(mesh_ -> num_top_regions);
  }
}


void 
QMG::SimpComplex_Under_Construction::
add_vertex_bdryinc(VertexGlobalIndex vnum,
                   const Brep::Face_Spec& fspec,
                   Brep::PatchIndex patchind,
                   const vector<Real>& param) {
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();
  if (global_to_ordinal_ -> find(vnum) == global_to_ordinal_ -> end())
    throw_error("Global vertex id not found");
#ifdef RANGECHECK
  if (fdim < 0 || fdim >= mesh_ -> embedded_dimension || 
    fdim > mesh_ -> intrinsic_dimension)
    throw_error("Face dim out of range");
  if (faceind < 0)
    throw_error("Face index out of range");
  if ( (fdim == 0 && faceind >= mesh_ -> num_top_vertices) 
    || (fdim == 1 && faceind >= mesh_ -> num_top_edges)
    || (fdim == 2 && faceind >= mesh_ -> num_top_surfaces))
    throw_error("Face index out of range");
  if (param.size() != fdim) 
    throw_error("Wrong length for param coords");
#endif
  if (patchind < 0)
    throw_error("Patch index out of range");
  Vertex_Bdry_Inc_ inc;
  Vertex_Pos_ pos;
  inc.vnum = vnum;
  inc.fspec = fspec;
  for (int k1 = 0; k1 < fdim; ++k1)
    pos.param[k1] = param[k1];
  pos.patchind = patchind;
  pair<Vert_Bdry_Lookup_::iterator,bool> rval = 
    vertinclookup_ -> insert(pair<Vertex_Bdry_Inc_, Vertex_Pos_>(inc,pos));
  if (!rval.second) {
    // if (patchind != rval.first -> second.patchind)
    //   throw_error("Adding vertex twice; patchind mismatch");
    // for (int k2 = 0; k2 < fdim; ++k2) {
    //  Real p2 = rval.first -> second.param[k2];
    //  if (param[k2] != p2)
    //    throw_error("Adding vertex twice; param mismatch");
    // }
    return;
  }
  if (fdim == 0) {
    VertexGlobalIndex oldvnum = mesh_ -> vtxinfo[faceind].vtx;
    if (oldvnum < 0) 
      mesh_ -> vtxinfo[faceind].vtx = vnum;
    else if (oldvnum != vnum)
      throw_error("Distinct vertex numbers assigned to topological vertex");
  }
  else if (fdim == 1) {
    int nn = mesh_ -> edgeinfo[faceind].num_mnode;
    int mx = max_node_edge_[faceind];
    if (nn >= mx) {
      int newsize = 2 * mx + 1;
      Cptc_NodeCurveInc* newincs = 
        static_cast<Cptc_NodeCurveInc*>(malloc1(sizeof(Cptc_NodeCurveInc) * newsize));
      for (int jj = 0; jj < nn; ++jj)
        newincs[jj] = mesh_ -> edgeinfo[faceind].incs[jj];
      free(mesh_ -> edgeinfo[faceind].incs);
      mesh_ -> edgeinfo[faceind].incs = newincs;
      max_node_edge_[faceind] = newsize;
    }
    mesh_ -> edgeinfo[faceind].incs[nn].id = vnum;
    mesh_ -> edgeinfo[faceind].incs[nn].curve_index = patchind;
    mesh_ -> edgeinfo[faceind].incs[nn].param = param[0];
    ++(mesh_ -> edgeinfo[faceind].num_mnode);
  }
  else { // fdim == 2
    int nn = mesh_ -> surfinfo[faceind].num_mnode;
    int mx = max_node_surf_[faceind];
    if (nn >= mx) {
      int newsize = 2 * mx + 1;
      Cptc_NodePatchInc* newincs = 
        static_cast<Cptc_NodePatchInc*>(malloc1(sizeof(Cptc_NodePatchInc) * newsize));
      for (int jj = 0; jj < nn; ++jj)
        newincs[jj] = mesh_ -> surfinfo[faceind].incs[jj];
      free(mesh_ -> surfinfo[faceind].incs);
      mesh_ -> surfinfo[faceind].incs = newincs;
      max_node_surf_[faceind] = newsize;
    }
    mesh_ -> surfinfo[faceind].incs[nn].id = vnum;
    mesh_ -> surfinfo[faceind].incs[nn].patch_index = patchind;
    mesh_ -> surfinfo[faceind].incs[nn].param[0] = param[0];
    mesh_ -> surfinfo[faceind].incs[nn].param[1] = param[1];
    ++(mesh_ -> surfinfo[faceind].num_mnode);
  }
}

QMG::Brep::PatchIndex 
QMG::SimpComplex_Under_Construction::
retrieve_vertex_bdryinc_patchind(VertexGlobalIndex vnum,
                                 const Brep::Face_Spec& fspec) const {
  Vertex_Bdry_Inc_ vp;
  vp.vnum = vnum;
  vp.fspec = fspec;
  Vert_Bdry_Lookup_::const_iterator it = 
    vertinclookup_ -> find(vp);
  if (it == vertinclookup_ -> end())
    throw_error("Vertex incidence not found");
  return it -> second.patchind;
}


QMG::Point
QMG::SimpComplex_Under_Construction::
retrieve_vertex_bdryinc_param(VertexGlobalIndex vnum,
                              const Brep::Face_Spec& fspec) const {
  Vertex_Bdry_Inc_ vp;
  vp.vnum = vnum;
  vp.fspec = fspec;
  Vert_Bdry_Lookup_::const_iterator it = 
    vertinclookup_ -> find(vp);
  if (it == vertinclookup_ -> end())
    throw_error("Vertex incidence not found");
  Point pt;
  pt[0] = it -> second.param[0];
  pt[1] = it -> second.param[1];
  return pt;
}




void
QMG::SimpComplex_Under_Construction::
add_simplex_face(const vector<VertexGlobalIndex>& verts,
                 const Brep::Face_Spec& fspec) { 
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();

#ifdef RANGECHECK
  if (fdim <= 0 || fdim > mesh_ -> intrinsic_dimension)
    throw_error("Face dim out of range");
  if (faceind < 0)
    throw_error("Face index out of range");
  if ( (fdim == 3 && faceind >= mesh_ -> num_top_regions) 
    || (fdim == 1 && faceind >= mesh_ -> num_top_edges)
    || (fdim == 2 && faceind >= mesh_ -> num_top_surfaces))
    throw_error("Face index out of range");
  if (verts.size() != fdim + 1) 
    throw_error("Wrong length for vertex list");
#endif
  for (int jj = 0; jj < fdim + 1; ++jj) {
    if (global_to_ordinal_ -> find(verts[jj]) == global_to_ordinal_ -> end())
      throw_error("Global vertex id not found");
  }
  if (fdim == 1) {
    Face_Bdry_Inc_ inc;
    if (verts[0] > verts[1]) {      
      inc.meshv[1] = verts[0];
      inc.meshv[0] = verts[1];
    }
    else {      
      inc.meshv[0] = verts[0];
      inc.meshv[1] = verts[1];
    }
    inc.meshv[2] = -1;
    pair<Face_Bdry_Lookup_::iterator, bool> rval =
      faceinclookup_ -> insert(pair<Face_Bdry_Inc_,Brep::Face_Spec>(inc,fspec));
    if (!rval.second) {
      if (rval.first -> second != fspec) {
        throw_error("Attempt to add simplex face with mismatched fspecs");
      }
      return;
    }    
    int ne = mesh_ -> edgeinfo[faceind].num_medge;
    int mx = max_edge_edge_[faceind];
    if (ne >= mx) {
      int newsize = 2 * mx + 1;
      typedef VertexGlobalIndex Twoint[2];
      Twoint* newedges = 
        static_cast<Twoint*>(malloc1(2 * sizeof(Twoint) * newsize));
      for (int jj = 0; jj < ne; ++jj) {
        newedges[jj][0] = mesh_ -> edgeinfo[faceind].edgelist[jj][0];
        newedges[jj][1] = mesh_ -> edgeinfo[faceind].edgelist[jj][1];
      }
      free(mesh_ -> edgeinfo[faceind].edgelist);
      mesh_ -> edgeinfo[faceind].edgelist = newedges;
      max_edge_edge_[faceind] = newsize;
    }
    mesh_ -> edgeinfo[faceind].edgelist[ne][0] = verts[0];
    mesh_ -> edgeinfo[faceind].edgelist[ne][1] = verts[1];
    ++(mesh_ -> edgeinfo[faceind].num_medge);
    if (mesh_ -> intrinsic_dimension == 1)
      ++(mesh_ -> num_elems);
  }
  else if (fdim == 2) {
    Face_Bdry_Inc_ inc;
    for (int j = 0; j < 3; ++j)
      inc.meshv[j] = verts[j];
    if (inc.meshv[0] > inc.meshv[1]) {
      VertexGlobalIndex tmp = inc.meshv[0];
      inc.meshv[0] = inc.meshv[1];
      inc.meshv[1] = tmp;
    }
    if (inc.meshv[1] > inc.meshv[2]) {
      VertexGlobalIndex tmp = inc.meshv[1];
      inc.meshv[1] = inc.meshv[2];
      inc.meshv[2] = tmp;
    }
    if (inc.meshv[0] > inc.meshv[1]) {
      VertexGlobalIndex tmp = inc.meshv[0];
      inc.meshv[0] = inc.meshv[1];
      inc.meshv[1] = tmp;
    }
    pair<Face_Bdry_Lookup_::iterator, bool> rval =
      faceinclookup_ -> insert(pair<Face_Bdry_Inc_,Brep::Face_Spec>(inc,fspec));
    if (!rval.second) {
      if (rval.first -> second != fspec) {
        throw_error("Attempt to add simplex face with mismatched fspecs");
      }
      return;
    }   
    int nt = mesh_ -> surfinfo[faceind].num_mtrifacet;
    int mx = max_tri_surf_[faceind];
    if (nt >= mx) {
      int newsize = 2 * mx + 1;
      typedef VertexGlobalIndex Threeint[3];
      Threeint* newtri = 
        static_cast<Threeint*>(malloc1(2 * sizeof(Threeint) * newsize));
      for (int jj = 0; jj < nt; ++jj) {
        for (int kk = 0; kk < 3; ++kk)
          newtri[jj][kk] = mesh_ -> surfinfo[faceind].trifacetlist[jj][kk];
      }
      free(mesh_ -> surfinfo[faceind].trifacetlist);
      mesh_ -> surfinfo[faceind].trifacetlist = newtri;
      max_tri_surf_[faceind] = newsize;
    }
    for (int kk2 = 0; kk2 < 3; ++kk2) 
      mesh_ -> surfinfo[faceind].trifacetlist[nt][kk2] = verts[kk2];
    ++(mesh_ -> surfinfo[faceind].num_mtrifacet);
    if (mesh_ -> intrinsic_dimension == 2)
      ++(mesh_ -> num_elems);
  }
  else { // fdim == 3  
    int nt = mesh_ -> regioninfo[faceind].num_tet;
    int mx = max_tet_region_[faceind];
    if (nt >= mx) {
      int newsize = 2 * mx + 1;
      typedef VertexGlobalIndex Fourint[4];
      Fourint* newtet = 
        static_cast<Fourint*>(malloc1(2 * sizeof(Fourint) * newsize));
      for (int jj = 0; jj < nt; ++jj) {
        for (int kk = 0; kk < 4; ++kk)
          newtet[jj][kk] = mesh_ -> regioninfo[faceind].tets[jj][kk];
      }
      free(mesh_ -> regioninfo[faceind].tets);
      mesh_ -> regioninfo[faceind].tets = newtet;
      max_tet_region_[faceind] = newsize;
    }
    for (int kk2 = 0; kk2 < 4; ++kk2) 
      mesh_ -> regioninfo[faceind].tets[nt][kk2] = verts[kk2];
    ++(mesh_ -> regioninfo[faceind].num_tet);
    ++(mesh_ -> num_elems);
  }
}


QMG::SimpComplex_Under_Construction::
SimpComplex_Under_Construction(int intrinsic_dim, 
                               int embedded_dim, 
                               const vector<int>& level_size):
SimpComplex(static_cast<Cptc_MshDesc*>(malloc1(sizeof(Cptc_MshDesc))), 
            new GlobalToOrdinalMapType)
{
  if (embedded_dim != 2 && embedded_dim != 3)
    throw_error("Illegal embedded dimension");
  if (intrinsic_dim < 1 || intrinsic_dim > embedded_dim)
    throw_error("Illegal intrinsic dimension");
  int brep_gdim = level_size.size() - 1;
  if (brep_gdim < intrinsic_dim || brep_gdim > embedded_dim)
    throw_error("illegal brep intrinsic dimension");
  max_lev_size_.resize(intrinsic_dim + 1);
  for (int ll = 0; ll <= intrinsic_dim; ++ll)
    max_lev_size_[ll] = level_size[ll];
  mesh_ -> intrinsic_dimension = intrinsic_dim;
  mesh_ -> embedded_dimension = embedded_dim;
  mesh_ -> num_nodes = 0;
  mesh_ -> num_elems = 0;
  mesh_ -> num_prop_val = 0;
  max_prop_val_ = 1;
  next_vnum_ = 0;
  mesh_ -> props = static_cast<char**>(malloc1(sizeof(char*)));
  mesh_ -> vals = static_cast<char**>(malloc1(sizeof(char*)));
  max_nodes_ = 1;
  mesh_ -> nodes = static_cast<Cptc_Mnode*>(malloc1(sizeof(Cptc_Mnode)));
  int ntv = level_size[0];
  if (ntv < 0)
    throw_error("num top vtx out of range");
  mesh_ -> num_top_vertices = ntv;
  mesh_ -> vtxinfo = 
    static_cast<Cptc_MeshTopVertexInfo*>(malloc1(sizeof(Cptc_MeshTopVertexInfo) * ntv));
  {
    for (int faceind = 0; faceind < ntv; ++faceind)
      mesh_ -> vtxinfo[faceind].vtx = -1;
  }
  int nte = level_size[1];
  if (nte < 0)
    throw_error("num top edge out of range");
  mesh_ -> num_top_edges = nte;
  max_node_edge_.resize(nte, 1);
  max_edge_edge_.resize(nte, 1);
  mesh_ -> edgeinfo =
    static_cast<Cptc_MeshTopEdgeInfo*>(malloc1(sizeof(Cptc_MeshTopEdgeInfo) * nte));
  {
    for (int faceind = 0; faceind < nte; ++faceind) {
      mesh_ -> edgeinfo[faceind].num_mnode = 0;
      mesh_ -> edgeinfo[faceind].num_medge = 0;
      mesh_ -> edgeinfo[faceind].incs =
        static_cast<Cptc_NodeCurveInc*>(malloc1(sizeof(Cptc_NodeCurveInc)));
      typedef VertexGlobalIndex Twoint[2];
      mesh_ -> edgeinfo[faceind].edgelist = 
        static_cast<Twoint*>(malloc1(sizeof(Twoint)));
    }
  }
  int nts = (brep_gdim < 2)? 0 : level_size[2];
  if (nts < 0)
    throw_error("num top surf out of range");
  mesh_ -> num_top_surfaces = nts;
  max_node_surf_.resize(nts, 1);
  max_tri_surf_.resize(nts, 1);
  mesh_ -> surfinfo =
    static_cast<Cptc_MeshTopSurfaceInfo*>(malloc1(sizeof(Cptc_MeshTopSurfaceInfo) * nts));
  {
    for (int faceind = 0; faceind < nts; ++faceind) {
      mesh_ -> surfinfo[faceind].num_mnode = 0;
      mesh_ -> surfinfo[faceind].num_mtrifacet = 0;
      mesh_ -> surfinfo[faceind].num_mquadfacet = 0;
      mesh_ -> surfinfo[faceind].incs =
        static_cast<Cptc_NodePatchInc*>(malloc1(sizeof(Cptc_NodePatchInc)));
      typedef VertexGlobalIndex Threeint[3];
      mesh_ -> surfinfo[faceind].trifacetlist = 
        static_cast<Threeint*>(malloc1(sizeof(Threeint)));
      mesh_ -> surfinfo[faceind].quadfacetlist = 0;
    }
  }
  
  int ntr = (brep_gdim < 3)? 0 : level_size[3];
  if (ntr < 0)
    throw_error("num top region out of range");

  mesh_ -> num_top_regions = ntr;
  max_tet_region_.resize(ntr, 1);
  mesh_ -> regioninfo =
    static_cast<Cptc_MeshTopRegionInfo*>(malloc1(sizeof(Cptc_MeshTopRegionInfo) * ntr));
  {
    for (int faceind = 0; faceind < ntr; ++faceind) {
      typedef VertexGlobalIndex Fourint[4];
      mesh_ -> regioninfo[faceind].num_tet = 0;
      mesh_ -> regioninfo[faceind].num_pyramid = 0;
      mesh_ -> regioninfo[faceind].num_prism = 0;
      mesh_ -> regioninfo[faceind].num_hex = 0;
      mesh_ -> regioninfo[faceind].tets = 
        static_cast<Fourint*>(malloc1(sizeof(Fourint)));
      mesh_ -> regioninfo[faceind].pyramids = 0;
      mesh_ -> regioninfo[faceind].prisms = 0;
      mesh_ -> regioninfo[faceind].hexs = 0;
      
    }
  }
  vertinclookup_ = new Vert_Bdry_Lookup_;
  faceinclookup_ = new Face_Bdry_Lookup_;
}

void
QMG::SimpComplex_Under_Construction::add_prop_val(const string& prop,
                                                  const string& val) {
  int npv = mesh_ -> num_prop_val;
  int mx = max_prop_val_;
  if (npv >= mx) {
    int newsize = 2 * mx + 1;
    char** newprop = static_cast<char**>(malloc1(sizeof(char*) * newsize));
    char** newval = static_cast<char**>(malloc1(sizeof(char*) * newsize));
    for (int jj = 0; jj < npv; ++jj) {
      newprop[jj] = mesh_ -> props[jj];
      newval[jj] = mesh_ -> vals[jj];
    }
    free(mesh_ -> props);
    free(mesh_ -> vals);
    mesh_ -> props = newprop;
    mesh_ -> vals = newval;
    max_prop_val_ = newsize;
  }
  
  int l1 = prop.length();
  mesh_-> props[npv] = static_cast<char*>(malloc1(sizeof(char) * (l1 + 1)));
  prop.copy(mesh_ -> props[npv], l1);
  mesh_ -> props[npv][l1] = 0;
  int l2 = val.length();
  mesh_ -> vals[npv] = static_cast<char*>(malloc(sizeof(char) * (l2 + 1)));
  val.copy(mesh_ -> vals[npv], l2);
  mesh_ -> vals[npv][l2] = 0;
  ++(mesh_ -> num_prop_val);
}



void
QMG::SimpComplex_Under_Construction::remove_unused_vertices() {
  map<VertexGlobalIndex, int> lookupnode;
  int gdim = mesh_ -> intrinsic_dimension;
  int levsize = this -> brep_levelsize(gdim);
  {
    for (Brep::FaceIndex faceind = 0; faceind < levsize; ++faceind) {
      Brep::Face_Spec fspec(gdim, faceind);
      int nf = this -> num_meshface_on_face(fspec);
      for (int k = 0; k < nf; ++k) {
        for (int l = 0; l < gdim + 1; ++l) {
          VertexGlobalIndex vnum = this -> node_of_meshface_on_face(fspec, k, l);
          lookupnode[vnum] = 1;
        }
      }
    }
  }
  VertexOrdinalIndex dest_vnumo = 0;
  int nn = mesh_ -> num_nodes;
  for (VertexOrdinalIndex source_vnumo = 0;
  source_vnumo < nn;
  ++source_vnumo) {
    VertexGlobalIndex vnum = ordinal_to_global(source_vnumo);
    if (lookupnode[vnum]) {
      mesh_ -> nodes[dest_vnumo].id = vnum;
      for (int jj = 0; jj < 3; ++jj)
        mesh_ -> nodes[dest_vnumo].coord[jj] = 
        mesh_ -> nodes[source_vnumo].coord[jj];
      (*global_to_ordinal_)[vnum] = dest_vnumo;
      ++dest_vnumo;
    }
  }
  mesh_ -> num_nodes = dest_vnumo;

  for (int fdim = 0; fdim < gdim; ++fdim) {
    int levsize = this -> brep_levelsize(fdim);
    for (Brep::FaceIndex faceind = 0; faceind < levsize; ++faceind) {
      Brep::Face_Spec fspec(fdim, faceind);
      int nnf = num_node_on_face(fspec);
      int destseqno = 0;
      for (int seqno = 0; seqno < nnf; ++seqno) {
        VertexGlobalIndex vnum = node_on_face(fspec, seqno);
        if (lookupnode[vnum]) {
          if (fdim == 1) {
            mesh_ -> edgeinfo[faceind].incs[destseqno] = 
              mesh_ -> edgeinfo[faceind].incs[seqno];
          }
          else if (fdim == 2) {
            mesh_ -> surfinfo[faceind].incs[destseqno] =
              mesh_ -> surfinfo[faceind].incs[seqno];
          }
          ++destseqno;
        }
      }
      if (fdim == 0) {
        if (destseqno == 0) {
          mesh_ -> vtxinfo[faceind].vtx = -1;
        }
      }
      else if (fdim == 1) {
        mesh_ -> edgeinfo[faceind].num_mnode = destseqno;
      }
      else if (fdim == 2) {
        mesh_ -> surfinfo[faceind].num_mnode = destseqno;
      }
      if (fdim == 0) continue;
      int nmf = num_meshface_on_face(fspec);
      int destseqno2 = 0;
      for (int seqno2 = 0; seqno2 < nmf; ++seqno2) {
        bool all_still_there = true;
        for (int jj = 0; jj < fdim + 1; ++jj) {
          VertexGlobalIndex vnum = node_of_meshface_on_face(fspec, seqno2, jj);
          if (!lookupnode[vnum]) {
            all_still_there = false;
            break;
          }
        }
        if (all_still_there) {
          if (fdim == 1) {
            mesh_ -> edgeinfo[faceind].edgelist[destseqno2][0] =
              mesh_ -> edgeinfo[faceind].edgelist[seqno2][0];
            mesh_ -> edgeinfo[faceind].edgelist[destseqno2][1] =
              mesh_ -> edgeinfo[faceind].edgelist[seqno2][1];
          }
          else if (fdim == 2) {
            for (int jj2 = 0; jj2 < 3; ++jj2) 
              mesh_ -> surfinfo[faceind].trifacetlist[destseqno2][jj2] =
              mesh_ -> surfinfo[faceind].trifacetlist[seqno2][jj2];
          }
          ++destseqno2;
        }
      }
      if (fdim == 1) 
        mesh_ -> edgeinfo[faceind].num_medge = destseqno2;
      else if (fdim == 2)
        mesh_ -> surfinfo[faceind].num_mtrifacet = destseqno2;
    }
  }
}


Cptc_MshDesc* 
QMG::SimpComplex_Under_Construction::release() {
  Cptc_MshDesc* p = mesh_;
  delete faceinclookup_;
  delete vertinclookup_;
  delete global_to_ordinal_;
  clobber_();
  return p;
}

QMG::SimpComplex_Under_Construction::~SimpComplex_Under_Construction() {
  if (mesh_)
    cptc_free_mesh(mesh_);
  delete faceinclookup_;
  delete vertinclookup_;
  delete global_to_ordinal_;
}

  
QMG::SimpComplex_Under_Construction::
SimpComplex_Under_Construction(SimpComplex_Under_Construction& other) :
SimpComplex(other.mesh_, other.global_to_ordinal_),
max_nodes_(other.max_nodes_),
max_lev_size_(other.max_lev_size_),
max_node_edge_(other.max_node_edge_),
max_edge_edge_(other.max_edge_edge_),
max_node_surf_(other.max_node_surf_),
max_tri_surf_(other.max_tri_surf_),
max_tet_region_(other.max_tet_region_),
max_prop_val_(other.max_prop_val_),
next_vnum_(other.next_vnum_),
vertinclookup_(other.vertinclookup_),
faceinclookup_(other.faceinclookup_) {
  other.clobber_();
}


QMG::SimpComplex_Under_Construction::
SimpComplex_Under_Construction(const SimpComplex_Under_Construction_Returnval& sucr) :
SimpComplex(sucr.sref_.mesh_, sucr.sref_.global_to_ordinal_),
max_nodes_(sucr.sref_.max_nodes_),
max_lev_size_(sucr.sref_.max_lev_size_),
max_node_edge_(sucr.sref_.max_node_edge_),
max_edge_edge_(sucr.sref_.max_edge_edge_),
max_node_surf_(sucr.sref_.max_node_surf_),
max_tri_surf_(sucr.sref_.max_tri_surf_),
max_tet_region_(sucr.sref_.max_tet_region_),
max_prop_val_(sucr.sref_.max_prop_val_),
next_vnum_(sucr.sref_.next_vnum_),
vertinclookup_(sucr.sref_.vertinclookup_),
faceinclookup_(sucr.sref_.faceinclookup_) {
  sucr.sref_.clobber_();
}


  
